#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "defines.h"
#include "less.h"
#include "kanji_map.h"
#include "unicode_map.h"

static SETCHARSET save_scs = 0;
int32_t *unicode0_map = NULL;
int32_t *unicode2_map = NULL;

int16_t iso8859_list[] = {
  LATIN1, LATIN2, LATIN3, LATIN4, CYRILLIC, ARABIC, GREEK,  HEBREW,
  LATIN5, LATIN6, THAI,   LATIN7, LATIN8,   LATIN9, LATIN10,
};

static void
set_unicode_char(int32_t ucode, int32_t c)
{
  int plane, code;

  if (ucode < 0)
    return;

  code  = ucode & 0xffff;
  plane = (ucode & 0x7fff0000) >> 16;

  if (plane == 0) {
    if (unicode0_map[code] != U_error)
      return;
    unicode0_map[code] = c;
  } else if (plane == 2) {
    if (unicode2_map[code] != U_error)
      return;
    unicode2_map[code] = c;
  }
}

void
make_unicode_map(SETCHARSET scs, int round_trip)
{
  int i, j;

  if (unicode0_map && save_scs == scs)
    return;

  /* initialize table */
  if (!unicode0_map)
    unicode0_map = (int32_t *) malloc(sizeof(int32_t) * 0x10000);
  if (!unicode2_map)
    unicode2_map = (int32_t *) malloc(sizeof(int32_t) * 0x10000);
  if (!unicode0_map || !unicode2_map) {
    fprintf(stderr, "out of memory\n");
    exit(1);
  }
  for (i = 0; i < 0x10000; ++ i) {
    unicode0_map[i] = -1;
    unicode2_map[i] = -1;
  }

  /* ASCII  */
  if (scs & SCSASCII)
    for (i = 0; i < 128; ++ i)
      set_unicode_char(ucode_latin[UMAP_ISO8859_1][i], UMAP(ASCII, i));

  /* EUC Japan Kanji set 1 */
  if (scs & SCSJISX0208_1983)
    for (i = 0; i < U_kanji; ++ i)
      if (kanji_map[i / 94][i % 94] & CHAR_JISX0208_1983)
	set_unicode_char(ucode_kanji1[i], UMAP(JISX0208KANJI, i));
  if (scs & SCSJISX0208_1990)
    for (i = 0; i < U_kanji; ++ i)
      if (kanji_map[i / 94][i % 94] & CHAR_JISX0208_1990)
	set_unicode_char(ucode_kanji1[i], UMAP(JISX0208_90KANJI, i));
  if (scs & SCSJISX0213_2000)
    for (i = 0; i < U_kanji; ++ i)
      if (kanji_map[i / 94][i % 94] & CHAR_JISX0213_2000_1)
	set_unicode_char(ucode_kanji1[i], UMAP(JISX0213KANJI1, i));
  if (scs & SCSJISX0213_2004)
    for (i = 0; i < U_kanji; ++ i)
      if (kanji_map[i / 94][i % 94] & CHAR_JISX0213_2004_1)
	set_unicode_char(ucode_kanji1[i], UMAP(JISX02132004KANJI1, i));

  /* EUC Japan Kanji set 2 */
  if (scs & SCSJISX0213_2ND)
    for (i = 0; i < U_kanji; ++ i)
      if (kanji_map[i / 94][i % 94] & CHAR_JISX0213_2000_2)
	set_unicode_char(ucode_kanji2[i], UMAP(JISX0213KANJI2, i));
  if (scs & SCSJISX0212_1990)
    for (i = 0; i < U_kanji; ++ i)
      if (kanji_map[i / 94][i % 94] & CHAR_JISX0212_1990)
	set_unicode_char(ucode_kanji2[i], UMAP(JISX0212KANJISUP, i));

  /* Microsoft Windows-31J aka CP932 */
  if (scs & SCSCP932EX)
    for (i = 0; i < U_kanji; ++ i)
	set_unicode_char(ucode_cp932[i], UMAP(CP932, i));

  /* JIS Roman */
  if (scs & SCSJISX0201_1976)
    for (i = 0; i < 128; ++ i)
      set_unicode_char(ucode_latin[UMAP_JISX0201][i], UMAP(JISX0201ROMAN, i));

  /* JIS Kana */
  if (scs & SCSJISX0201_1976)
    for (i = 128; i < U_latin; ++ i)
      set_unicode_char(ucode_latin[UMAP_JISX0201][i], UMAP(JISX0201KANA, i));

  if (!round_trip) {
    /* Correction for OLD/Erratical mapping*/
    if (scs & SCSJISX0208_1983) {
      set_unicode_char(0x2015, UMAP(JISX0208KANJI, 1*94+29-95));
                                                           /* glibc et al */
      set_unicode_char(0x2225, UMAP(JISX0208KANJI, 1*94+34-95)); /* CP932 */
      set_unicode_char(0x22ef, UMAP(JISX0208KANJI, 1*94+36-95)); /* MAC */
      set_unicode_char(0xff0d, UMAP(JISX0208KANJI, 1*94+61-95)); /* CP932 */
      set_unicode_char(0xff5e, UMAP(JISX0208KANJI, 1*94+33-95)); /* CP932 */
      set_unicode_char(0xffe0, UMAP(JISX0208KANJI, 1*94+81-95)); /* CP932 */
      set_unicode_char(0xffe1, UMAP(JISX0208KANJI, 1*94+82-95)); /* CP932 */
      set_unicode_char(0xffe2, UMAP(JISX0208KANJI, 2*94+44-95)); /* CP932 */
    }
    if (scs & (SCSJISX0213_2000 | SCSJISX0213_2004)) {
      set_unicode_char(0xffe4, UMAP(JISX0213KANJI1, 9*94+5-95));  /* CP932 */
      set_unicode_char(0x2299, UMAP(JISX0213KANJI1, 3*94+26-95));
                                                              /* old glibc */
    } 
  }

  /* ISO 8859 */
  if (scs & SCSOTHERISO)
    for (j = UMAP_ISO8859_1; j <= UMAP_ISO8859_16; ++ j) 
      for (i = 128; i < U_latin; ++ i)
	set_unicode_char(ucode_latin[j][i], UMAP(iso8859_list[j], i));

  if (!round_trip) {
    /* Emacs-23 JIS C 6226-1978 */
    if (scs & SCSJISC6226_1978) {
      set_unicode_char(0x3d4e,  UMAP(JISX0208_78KANJI, 20*94+34-95));
      set_unicode_char(0x25874, UMAP(JISX0208_78KANJI, 23*94+46-95));
      set_unicode_char(0x8346,  UMAP(JISX0208_78KANJI, 23*94+53-95));
      set_unicode_char(0x28ef6, UMAP(JISX0208_78KANJI, 23*94+68-95));
      set_unicode_char(0x663b,  UMAP(JISX0208_78KANJI, 25*94+23-95));
      set_unicode_char(0x6805,  UMAP(JISX0208_78KANJI, 26*94+84-95));
      set_unicode_char(0x688e,  UMAP(JISX0208_78KANJI, 59*94+84-95));
    }

    /* CP932 unify to JIS */
    if (scs & SCSJISX0208_1983) {
      set_unicode_char(0x2211, UMAP(JISX0208KANJI, 6*94+18-95));  /* unify */
      set_unicode_char(0x6085, UMAP(JISX0208KANJI, 17*94+57-95)); /* unify */
      set_unicode_char(0x6120, UMAP(JISX0208KANJI, 56*94+18-95)); /* unify */
      set_unicode_char(0x654e, UMAP(JISX0208KANJI, 22*94+21-95)); /* unify */
      set_unicode_char(0x663b, UMAP(JISX0208KANJI, 25*94+23-95)); /* unify */
      set_unicode_char(0x6df8, UMAP(JISX0208KANJI, 32*94+22-95)); /* unify */
      set_unicode_char(0x7682, UMAP(JISX0208KANJI, 66*94+5-95));  /* unify */
      set_unicode_char(0x9592, UMAP(JISX0208KANJI, 20*94+54-95)); /* unify */
      set_unicode_char(0x9751, UMAP(JISX0208KANJI, 32*94+36-95)); /* unify */
      set_unicode_char(0x9ad9, UMAP(JISX0208KANJI, 25*94+66-95)); /* unify */
      set_unicode_char(0xfa12, UMAP(JISX0208KANJI, 32*94+18-95)); /* unify */
      set_unicode_char(0xfa17, UMAP(JISX0208KANJI, 17*94+55-95)); /* unify */
      set_unicode_char(0xfa18, UMAP(JISX0208KANJI, 46*94+73-95)); /* unify */
      set_unicode_char(0xfa1c, UMAP(JISX0208KANJI, 44*94+87-95)); /* unify */
      set_unicode_char(0xfa1d, UMAP(JISX0208KANJI, 32*94+26-95)); /* unify */
      set_unicode_char(0xfa1e, UMAP(JISX0208KANJI, 17*94+9-95));  /* unify */
      set_unicode_char(0xfa25, UMAP(JISX0208KANJI, 16*94+79-95)); /* unify */
      set_unicode_char(0xfa2a, UMAP(JISX0208KANJI, 40*94+51-95)); /* unify */
      set_unicode_char(0xfa2b, UMAP(JISX0208KANJI, 27*94+84-95)); /* unify */
      set_unicode_char(0xfa2c, UMAP(JISX0208KANJI, 20*94+59-95)); /* unify */
      set_unicode_char(0xfa2d, UMAP(JISX0208KANJI, 36*94+65-95)); /* unify */
    }
    if (scs & SCSJISX0213_2ND) {
      set_unicode_char(0x50f4, UMAP(JISX0213KANJI2, 1*94+93-95)); /* unify */
      set_unicode_char(0x5759, UMAP(JISX0213KANJI2, 8*94+75-95)); /* unify */
    }
  }

  /* save Set Charset */
  save_scs = scs;
}

int32_t ucode_latin[UMAP_JISX0201+1][U_latin] = {
  {
    /*  ISO_8859-1  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x000a1, 0x000a2, 0x000a3, 0x000a4, 0x000a5, 0x000a6, 0x000a7,
    0x000a8, 0x000a9, 0x000aa, 0x000ab, 0x000ac, 0x000ad, 0x000ae, 0x000af,
    0x000b0, 0x000b1, 0x000b2, 0x000b3, 0x000b4, 0x000b5, 0x000b6, 0x000b7,
    0x000b8, 0x000b9, 0x000ba, 0x000bb, 0x000bc, 0x000bd, 0x000be, 0x000bf,
    0x000c0, 0x000c1, 0x000c2, 0x000c3, 0x000c4, 0x000c5, 0x000c6, 0x000c7,
    0x000c8, 0x000c9, 0x000ca, 0x000cb, 0x000cc, 0x000cd, 0x000ce, 0x000cf,
    0x000d0, 0x000d1, 0x000d2, 0x000d3, 0x000d4, 0x000d5, 0x000d6, 0x000d7,
    0x000d8, 0x000d9, 0x000da, 0x000db, 0x000dc, 0x000dd, 0x000de, 0x000df,
    0x000e0, 0x000e1, 0x000e2, 0x000e3, 0x000e4, 0x000e5, 0x000e6, 0x000e7,
    0x000e8, 0x000e9, 0x000ea, 0x000eb, 0x000ec, 0x000ed, 0x000ee, 0x000ef,
    0x000f0, 0x000f1, 0x000f2, 0x000f3, 0x000f4, 0x000f5, 0x000f6, 0x000f7,
    0x000f8, 0x000f9, 0x000fa, 0x000fb, 0x000fc, 0x000fd, 0x000fe, 0x000ff,
  },
  {
    /*  ISO_8859-2  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x00104, 0x002d8, 0x00141, 0x000a4, 0x0013d, 0x0015a, 0x000a7,
    0x000a8, 0x00160, 0x0015e, 0x00164, 0x00179, 0x000ad, 0x0017d, 0x0017b,
    0x000b0, 0x00105, 0x002db, 0x00142, 0x000b4, 0x0013e, 0x0015b, 0x002c7,
    0x000b8, 0x00161, 0x0015f, 0x00165, 0x0017a, 0x002dd, 0x0017e, 0x0017c,
    0x00154, 0x000c1, 0x000c2, 0x00102, 0x000c4, 0x00139, 0x00106, 0x000c7,
    0x0010c, 0x000c9, 0x00118, 0x000cb, 0x0011a, 0x000cd, 0x000ce, 0x0010e,
    0x00110, 0x00143, 0x00147, 0x000d3, 0x000d4, 0x00150, 0x000d6, 0x000d7,
    0x00158, 0x0016e, 0x000da, 0x00170, 0x000dc, 0x000dd, 0x00162, 0x000df,
    0x00155, 0x000e1, 0x000e2, 0x00103, 0x000e4, 0x0013a, 0x00107, 0x000e7,
    0x0010d, 0x000e9, 0x00119, 0x000eb, 0x0011b, 0x000ed, 0x000ee, 0x0010f,
    0x00111, 0x00144, 0x00148, 0x000f3, 0x000f4, 0x00151, 0x000f6, 0x000f7,
    0x00159, 0x0016f, 0x000fa, 0x00171, 0x000fc, 0x000fd, 0x00163, 0x002d9,
  },
  {
    /*  ISO_8859-3  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x00126, 0x002d8, 0x000a3, 0x000a4, U_error, 0x00124, 0x000a7,
    0x000a8, 0x00130, 0x0015e, 0x0011e, 0x00134, 0x000ad, U_error, 0x0017b,
    0x000b0, 0x00127, 0x000b2, 0x000b3, 0x000b4, 0x000b5, 0x00125, 0x000b7,
    0x000b8, 0x00131, 0x0015f, 0x0011f, 0x00135, 0x000bd, U_error, 0x0017c,
    0x000c0, 0x000c1, 0x000c2, U_error, 0x000c4, 0x0010a, 0x00108, 0x000c7,
    0x000c8, 0x000c9, 0x000ca, 0x000cb, 0x000cc, 0x000cd, 0x000ce, 0x000cf,
    U_error, 0x000d1, 0x000d2, 0x000d3, 0x000d4, 0x00120, 0x000d6, 0x000d7,
    0x0011c, 0x000d9, 0x000da, 0x000db, 0x000dc, 0x0016c, 0x0015c, 0x000df,
    0x000e0, 0x000e1, 0x000e2, U_error, 0x000e4, 0x0010b, 0x00109, 0x000e7,
    0x000e8, 0x000e9, 0x000ea, 0x000eb, 0x000ec, 0x000ed, 0x000ee, 0x000ef,
    U_error, 0x000f1, 0x000f2, 0x000f3, 0x000f4, 0x00121, 0x000f6, 0x000f7,
    0x0011d, 0x000f9, 0x000fa, 0x000fb, 0x000fc, 0x0016d, 0x0015d, 0x002d9,
  },
  {
    /*  ISO_8859-4  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x00104, 0x00138, 0x00156, 0x000a4, 0x00128, 0x0013b, 0x000a7,
    0x000a8, 0x00160, 0x00112, 0x00122, 0x00166, 0x000ad, 0x0017d, 0x000af,
    0x000b0, 0x00105, 0x002db, 0x00157, 0x000b4, 0x00129, 0x0013c, 0x002c7,
    0x000b8, 0x00161, 0x00113, 0x00123, 0x00167, 0x0014a, 0x0017e, 0x0014b,
    0x00100, 0x000c1, 0x000c2, 0x000c3, 0x000c4, 0x000c5, 0x000c6, 0x0012e,
    0x0010c, 0x000c9, 0x00118, 0x000cb, 0x00116, 0x000cd, 0x000ce, 0x0012a,
    0x00110, 0x00145, 0x0014c, 0x00136, 0x000d4, 0x000d5, 0x000d6, 0x000d7,
    0x000d8, 0x00172, 0x000da, 0x000db, 0x000dc, 0x00168, 0x0016a, 0x000df,
    0x00101, 0x000e1, 0x000e2, 0x000e3, 0x000e4, 0x000e5, 0x000e6, 0x0012f,
    0x0010d, 0x000e9, 0x00119, 0x000eb, 0x00117, 0x000ed, 0x000ee, 0x0012b,
    0x00111, 0x00146, 0x0014d, 0x00137, 0x000f4, 0x000f5, 0x000f6, 0x000f7,
    0x000f8, 0x00173, 0x000fa, 0x000fb, 0x000fc, 0x00169, 0x0016b, 0x002d9,
  },
  {
    /*  ISO_8859-5  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x00401, 0x00402, 0x00403, 0x00404, 0x00405, 0x00406, 0x00407,
    0x00408, 0x00409, 0x0040a, 0x0040b, 0x0040c, 0x000ad, 0x0040e, 0x0040f,
    0x00410, 0x00411, 0x00412, 0x00413, 0x00414, 0x00415, 0x00416, 0x00417,
    0x00418, 0x00419, 0x0041a, 0x0041b, 0x0041c, 0x0041d, 0x0041e, 0x0041f,
    0x00420, 0x00421, 0x00422, 0x00423, 0x00424, 0x00425, 0x00426, 0x00427,
    0x00428, 0x00429, 0x0042a, 0x0042b, 0x0042c, 0x0042d, 0x0042e, 0x0042f,
    0x00430, 0x00431, 0x00432, 0x00433, 0x00434, 0x00435, 0x00436, 0x00437,
    0x00438, 0x00439, 0x0043a, 0x0043b, 0x0043c, 0x0043d, 0x0043e, 0x0043f,
    0x00440, 0x00441, 0x00442, 0x00443, 0x00444, 0x00445, 0x00446, 0x00447,
    0x00448, 0x00449, 0x0044a, 0x0044b, 0x0044c, 0x0044d, 0x0044e, 0x0044f,
    0x02116, 0x00451, 0x00452, 0x00453, 0x00454, 0x00455, 0x00456, 0x00457,
    0x00458, 0x00459, 0x0045a, 0x0045b, 0x0045c, 0x000a7, 0x0045e, 0x0045f,
  },
  {
    /*  ISO_8859-6  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, U_error, U_error, U_error, 0x000a4, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, 0x0060c, 0x000ad, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, 0x0061b, U_error, U_error, U_error, 0x0061f,
    U_error, 0x00621, 0x00622, 0x00623, 0x00624, 0x00625, 0x00626, 0x00627,
    0x00628, 0x00629, 0x0062a, 0x0062b, 0x0062c, 0x0062d, 0x0062e, 0x0062f,
    0x00630, 0x00631, 0x00632, 0x00633, 0x00634, 0x00635, 0x00636, 0x00637,
    0x00638, 0x00639, 0x0063a, U_error, U_error, U_error, U_error, U_error,
    0x00640, 0x00641, 0x00642, 0x00643, 0x00644, 0x00645, 0x00646, 0x00647,
    0x00648, 0x00649, 0x0064a, 0x0064b, 0x0064c, 0x0064d, 0x0064e, 0x0064f,
    0x00650, 0x00651, 0x00652, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  },
  {
    /*  ISO_8859-7  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x02018, 0x02019, 0x000a3, 0x020ac, 0x020af, 0x000a6, 0x000a7,
    0x000a8, 0x000a9, 0x0037a, 0x000ab, 0x000ac, 0x000ad, U_error, 0x02015,
    0x000b0, 0x000b1, 0x000b2, 0x000b3, 0x00384, 0x00385, 0x00386, 0x000b7,
    0x00388, 0x00389, 0x0038a, 0x000bb, 0x0038c, 0x000bd, 0x0038e, 0x0038f,
    0x00390, 0x00391, 0x00392, 0x00393, 0x00394, 0x00395, 0x00396, 0x00397,
    0x00398, 0x00399, 0x0039a, 0x0039b, 0x0039c, 0x0039d, 0x0039e, 0x0039f,
    0x003a0, 0x003a1, U_error, 0x003a3, 0x003a4, 0x003a5, 0x003a6, 0x003a7,
    0x003a8, 0x003a9, 0x003aa, 0x003ab, 0x003ac, 0x003ad, 0x003ae, 0x003af,
    0x003b0, 0x003b1, 0x003b2, 0x003b3, 0x003b4, 0x003b5, 0x003b6, 0x003b7,
    0x003b8, 0x003b9, 0x003ba, 0x003bb, 0x003bc, 0x003bd, 0x003be, 0x003bf,
    0x003c0, 0x003c1, 0x003c2, 0x003c3, 0x003c4, 0x003c5, 0x003c6, 0x003c7,
    0x003c8, 0x003c9, 0x003ca, 0x003cb, 0x003cc, 0x003cd, 0x003ce, U_error,
  },
  {
    /*  ISO_8859-8  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, U_error, 0x000a2, 0x000a3, 0x000a4, 0x000a5, 0x000a6, 0x000a7,
    0x000a8, 0x000a9, 0x000d7, 0x000ab, 0x000ac, 0x000ad, 0x000ae, 0x000af,
    0x000b0, 0x000b1, 0x000b2, 0x000b3, 0x000b4, 0x000b5, 0x000b6, 0x000b7,
    0x000b8, 0x000b9, 0x000f7, 0x000bb, 0x000bc, 0x000bd, 0x000be, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, 0x02017,
    0x005d0, 0x005d1, 0x005d2, 0x005d3, 0x005d4, 0x005d5, 0x005d6, 0x005d7,
    0x005d8, 0x005d9, 0x005da, 0x005db, 0x005dc, 0x005dd, 0x005de, 0x005df,
    0x005e0, 0x005e1, 0x005e2, 0x005e3, 0x005e4, 0x005e5, 0x005e6, 0x005e7,
    0x005e8, 0x005e9, 0x005ea, U_error, U_error, 0x0200e, 0x0200f, U_error,
  },
  {
    /*  ISO_8859-9  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x000a1, 0x000a2, 0x000a3, 0x000a4, 0x000a5, 0x000a6, 0x000a7,
    0x000a8, 0x000a9, 0x000aa, 0x000ab, 0x000ac, 0x000ad, 0x000ae, 0x000af,
    0x000b0, 0x000b1, 0x000b2, 0x000b3, 0x000b4, 0x000b5, 0x000b6, 0x000b7,
    0x000b8, 0x000b9, 0x000ba, 0x000bb, 0x000bc, 0x000bd, 0x000be, 0x000bf,
    0x000c0, 0x000c1, 0x000c2, 0x000c3, 0x000c4, 0x000c5, 0x000c6, 0x000c7,
    0x000c8, 0x000c9, 0x000ca, 0x000cb, 0x000cc, 0x000cd, 0x000ce, 0x000cf,
    0x0011e, 0x000d1, 0x000d2, 0x000d3, 0x000d4, 0x000d5, 0x000d6, 0x000d7,
    0x000d8, 0x000d9, 0x000da, 0x000db, 0x000dc, 0x00130, 0x0015e, 0x000df,
    0x000e0, 0x000e1, 0x000e2, 0x000e3, 0x000e4, 0x000e5, 0x000e6, 0x000e7,
    0x000e8, 0x000e9, 0x000ea, 0x000eb, 0x000ec, 0x000ed, 0x000ee, 0x000ef,
    0x0011f, 0x000f1, 0x000f2, 0x000f3, 0x000f4, 0x000f5, 0x000f6, 0x000f7,
    0x000f8, 0x000f9, 0x000fa, 0x000fb, 0x000fc, 0x00131, 0x0015f, 0x000ff,
  },
  {
    /*  ISO_8859-10  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x00104, 0x00112, 0x00122, 0x0012a, 0x00128, 0x00136, 0x000a7,
    0x0013b, 0x00110, 0x00160, 0x00166, 0x0017d, 0x000ad, 0x0016a, 0x0014a,
    0x000b0, 0x00105, 0x00113, 0x00123, 0x0012b, 0x00129, 0x00137, 0x000b7,
    0x0013c, 0x00111, 0x00161, 0x00167, 0x0017e, 0x02015, 0x0016b, 0x0014b,
    0x00100, 0x000c1, 0x000c2, 0x000c3, 0x000c4, 0x000c5, 0x000c6, 0x0012e,
    0x0010c, 0x000c9, 0x00118, 0x000cb, 0x00116, 0x000cd, 0x000ce, 0x000cf,
    0x000d0, 0x00145, 0x0014c, 0x000d3, 0x000d4, 0x000d5, 0x000d6, 0x00168,
    0x000d8, 0x00172, 0x000da, 0x000db, 0x000dc, 0x000dd, 0x000de, 0x000df,
    0x00101, 0x000e1, 0x000e2, 0x000e3, 0x000e4, 0x000e5, 0x000e6, 0x0012f,
    0x0010d, 0x000e9, 0x00119, 0x000eb, 0x00117, 0x000ed, 0x000ee, 0x000ef,
    0x000f0, 0x00146, 0x0014d, 0x000f3, 0x000f4, 0x000f5, 0x000f6, 0x00169,
    0x000f8, 0x00173, 0x000fa, 0x000fb, 0x000fc, 0x000fd, 0x000fe, 0x00138,
  },
  {
    /*  ISO_8859-11  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x00e01, 0x00e02, 0x00e03, 0x00e04, 0x00e05, 0x00e06, 0x00e07,
    0x00e08, 0x00e09, 0x00e0a, 0x00e0b, 0x00e0c, 0x00e0d, 0x00e0e, 0x00e0f,
    0x00e10, 0x00e11, 0x00e12, 0x00e13, 0x00e14, 0x00e15, 0x00e16, 0x00e17,
    0x00e18, 0x00e19, 0x00e1a, 0x00e1b, 0x00e1c, 0x00e1d, 0x00e1e, 0x00e1f,
    0x00e20, 0x00e21, 0x00e22, 0x00e23, 0x00e24, 0x00e25, 0x00e26, 0x00e27,
    0x00e28, 0x00e29, 0x00e2a, 0x00e2b, 0x00e2c, 0x00e2d, 0x00e2e, 0x00e2f,
    0x00e30, 0x00e31, 0x00e32, 0x00e33, 0x00e34, 0x00e35, 0x00e36, 0x00e37,
    0x00e38, 0x00e39, 0x00e3a, U_error, U_error, U_error, U_error, 0x00e3f,
    0x00e40, 0x00e41, 0x00e42, 0x00e43, 0x00e44, 0x00e45, 0x00e46, 0x00e47,
    0x00e48, 0x00e49, 0x00e4a, 0x00e4b, 0x00e4c, 0x00e4d, 0x00e4e, 0x00e4f,
    0x00e50, 0x00e51, 0x00e52, 0x00e53, 0x00e54, 0x00e55, 0x00e56, 0x00e57,
    0x00e58, 0x00e59, 0x00e5a, 0x00e5b, U_error, U_error, U_error, U_error,
  },
  {
    /*  ISO_8859-13  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x0201d, 0x000a2, 0x000a3, 0x000a4, 0x0201e, 0x000a6, 0x000a7,
    0x000d8, 0x000a9, 0x00156, 0x000ab, 0x000ac, 0x000ad, 0x000ae, 0x000c6,
    0x000b0, 0x000b1, 0x000b2, 0x000b3, 0x0201c, 0x000b5, 0x000b6, 0x000b7,
    0x000f8, 0x000b9, 0x00157, 0x000bb, 0x000bc, 0x000bd, 0x000be, 0x000e6,
    0x00104, 0x0012e, 0x00100, 0x00106, 0x000c4, 0x000c5, 0x00118, 0x00112,
    0x0010c, 0x000c9, 0x00179, 0x00116, 0x00122, 0x00136, 0x0012a, 0x0013b,
    0x00160, 0x00143, 0x00145, 0x000d3, 0x0014c, 0x000d5, 0x000d6, 0x000d7,
    0x00172, 0x00141, 0x0015a, 0x0016a, 0x000dc, 0x0017b, 0x0017d, 0x000df,
    0x00105, 0x0012f, 0x00101, 0x00107, 0x000e4, 0x000e5, 0x00119, 0x00113,
    0x0010d, 0x000e9, 0x0017a, 0x00117, 0x00123, 0x00137, 0x0012b, 0x0013c,
    0x00161, 0x00144, 0x00146, 0x000f3, 0x0014d, 0x000f5, 0x000f6, 0x000f7,
    0x00173, 0x00142, 0x0015b, 0x0016b, 0x000fc, 0x0017c, 0x0017e, 0x02019,
  },
  {
    /*  ISO_8859-14  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x01e02, 0x01e03, 0x000a3, 0x0010a, 0x0010b, 0x01e0a, 0x000a7,
    0x01e80, 0x000a9, 0x01e82, 0x01e0b, 0x01ef2, 0x000ad, 0x000ae, 0x00178,
    0x01e1e, 0x01e1f, 0x00120, 0x00121, 0x01e40, 0x01e41, 0x000b6, 0x01e56,
    0x01e81, 0x01e57, 0x01e83, 0x01e60, 0x01ef3, 0x01e84, 0x01e85, 0x01e61,
    0x000c0, 0x000c1, 0x000c2, 0x000c3, 0x000c4, 0x000c5, 0x000c6, 0x000c7,
    0x000c8, 0x000c9, 0x000ca, 0x000cb, 0x000cc, 0x000cd, 0x000ce, 0x000cf,
    0x00174, 0x000d1, 0x000d2, 0x000d3, 0x000d4, 0x000d5, 0x000d6, 0x01e6a,
    0x000d8, 0x000d9, 0x000da, 0x000db, 0x000dc, 0x000dd, 0x00176, 0x000df,
    0x000e0, 0x000e1, 0x000e2, 0x000e3, 0x000e4, 0x000e5, 0x000e6, 0x000e7,
    0x000e8, 0x000e9, 0x000ea, 0x000eb, 0x000ec, 0x000ed, 0x000ee, 0x000ef,
    0x00175, 0x000f1, 0x000f2, 0x000f3, 0x000f4, 0x000f5, 0x000f6, 0x01e6b,
    0x000f8, 0x000f9, 0x000fa, 0x000fb, 0x000fc, 0x000fd, 0x00177, 0x000ff,
  },
  {
    /*  ISO_8859-15  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x000a1, 0x000a2, 0x000a3, 0x020ac, 0x000a5, 0x00160, 0x000a7,
    0x00161, 0x000a9, 0x000aa, 0x000ab, 0x000ac, 0x000ad, 0x000ae, 0x000af,
    0x000b0, 0x000b1, 0x000b2, 0x000b3, 0x0017d, 0x000b5, 0x000b6, 0x000b7,
    0x0017e, 0x000b9, 0x000ba, 0x000bb, 0x00152, 0x00153, 0x00178, 0x000bf,
    0x000c0, 0x000c1, 0x000c2, 0x000c3, 0x000c4, 0x000c5, 0x000c6, 0x000c7,
    0x000c8, 0x000c9, 0x000ca, 0x000cb, 0x000cc, 0x000cd, 0x000ce, 0x000cf,
    0x000d0, 0x000d1, 0x000d2, 0x000d3, 0x000d4, 0x000d5, 0x000d6, 0x000d7,
    0x000d8, 0x000d9, 0x000da, 0x000db, 0x000dc, 0x000dd, 0x000de, 0x000df,
    0x000e0, 0x000e1, 0x000e2, 0x000e3, 0x000e4, 0x000e5, 0x000e6, 0x000e7,
    0x000e8, 0x000e9, 0x000ea, 0x000eb, 0x000ec, 0x000ed, 0x000ee, 0x000ef,
    0x000f0, 0x000f1, 0x000f2, 0x000f3, 0x000f4, 0x000f5, 0x000f6, 0x000f7,
    0x000f8, 0x000f9, 0x000fa, 0x000fb, 0x000fc, 0x000fd, 0x000fe, 0x000ff,
  },
  {
    /*  ISO_8859-16  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x0005c, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0007e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    0x000a0, 0x00104, 0x00105, 0x00141, 0x020ac, 0x0201e, 0x00160, 0x000a7,
    0x00161, 0x000a9, 0x00218, 0x000ab, 0x00179, 0x000ad, 0x0017a, 0x0017b,
    0x000b0, 0x000b1, 0x0010c, 0x00142, 0x0017d, 0x0201d, 0x000b6, 0x000b7,
    0x0017e, 0x0010d, 0x00219, 0x000bb, 0x00152, 0x00153, 0x00178, 0x0017c,
    0x000c0, 0x000c1, 0x000c2, 0x00102, 0x000c4, 0x00106, 0x000c6, 0x000c7,
    0x000c8, 0x000c9, 0x000ca, 0x000cb, 0x000cc, 0x000cd, 0x000ce, 0x000cf,
    0x00110, 0x00143, 0x000d2, 0x000d3, 0x000d4, 0x00150, 0x000d6, 0x0015a,
    0x00170, 0x000d9, 0x000da, 0x000db, 0x000dc, 0x00118, 0x0021a, 0x000df,
    0x000e0, 0x000e1, 0x000e2, 0x00103, 0x000e4, 0x00107, 0x000e6, 0x000e7,
    0x000e8, 0x000e9, 0x000ea, 0x000eb, 0x000ec, 0x000ed, 0x000ee, 0x000ef,
    0x00111, 0x00144, 0x000f2, 0x000f3, 0x000f4, 0x00151, 0x000f6, 0x0015b,
    0x00171, 0x000f9, 0x000fa, 0x000fb, 0x000fc, 0x00119, 0x0021b, 0x000ff,
  },
  {
    /*  JIS_X_0201  */
    0x00000, 0x00001, 0x00002, 0x00003, 0x00004, 0x00005, 0x00006, 0x00007,
    0x00008, 0x00009, 0x0000a, 0x0000b, 0x0000c, 0x0000d, 0x0000e, 0x0000f,
    0x00010, 0x00011, 0x00012, 0x00013, 0x00014, 0x00015, 0x00016, 0x00017,
    0x00018, 0x00019, 0x0001a, 0x0001b, 0x0001c, 0x0001d, 0x0001e, 0x0001f,
    0x00020, 0x00021, 0x00022, 0x00023, 0x00024, 0x00025, 0x00026, 0x00027,
    0x00028, 0x00029, 0x0002a, 0x0002b, 0x0002c, 0x0002d, 0x0002e, 0x0002f,
    0x00030, 0x00031, 0x00032, 0x00033, 0x00034, 0x00035, 0x00036, 0x00037,
    0x00038, 0x00039, 0x0003a, 0x0003b, 0x0003c, 0x0003d, 0x0003e, 0x0003f,
    0x00040, 0x00041, 0x00042, 0x00043, 0x00044, 0x00045, 0x00046, 0x00047,
    0x00048, 0x00049, 0x0004a, 0x0004b, 0x0004c, 0x0004d, 0x0004e, 0x0004f,
    0x00050, 0x00051, 0x00052, 0x00053, 0x00054, 0x00055, 0x00056, 0x00057,
    0x00058, 0x00059, 0x0005a, 0x0005b, 0x000a5, 0x0005d, 0x0005e, 0x0005f,
    0x00060, 0x00061, 0x00062, 0x00063, 0x00064, 0x00065, 0x00066, 0x00067,
    0x00068, 0x00069, 0x0006a, 0x0006b, 0x0006c, 0x0006d, 0x0006e, 0x0006f,
    0x00070, 0x00071, 0x00072, 0x00073, 0x00074, 0x00075, 0x00076, 0x00077,
    0x00078, 0x00079, 0x0007a, 0x0007b, 0x0007c, 0x0007d, 0x0203e, 0x0007f,
    0x00080, 0x00081, 0x00082, 0x00083, 0x00084, 0x00085, 0x00086, 0x00087,
    0x00088, 0x00089, 0x0008a, 0x0008b, 0x0008c, 0x0008d, 0x0008e, 0x0008f,
    0x00090, 0x00091, 0x00092, 0x00093, 0x00094, 0x00095, 0x00096, 0x00097,
    0x00098, 0x00099, 0x0009a, 0x0009b, 0x0009c, 0x0009d, 0x0009e, 0x0009f,
    U_error, 0x0ff61, 0x0ff62, 0x0ff63, 0x0ff64, 0x0ff65, 0x0ff66, 0x0ff67,
    0x0ff68, 0x0ff69, 0x0ff6a, 0x0ff6b, 0x0ff6c, 0x0ff6d, 0x0ff6e, 0x0ff6f,
    0x0ff70, 0x0ff71, 0x0ff72, 0x0ff73, 0x0ff74, 0x0ff75, 0x0ff76, 0x0ff77,
    0x0ff78, 0x0ff79, 0x0ff7a, 0x0ff7b, 0x0ff7c, 0x0ff7d, 0x0ff7e, 0x0ff7f,
    0x0ff80, 0x0ff81, 0x0ff82, 0x0ff83, 0x0ff84, 0x0ff85, 0x0ff86, 0x0ff87,
    0x0ff88, 0x0ff89, 0x0ff8a, 0x0ff8b, 0x0ff8c, 0x0ff8d, 0x0ff8e, 0x0ff8f,
    0x0ff90, 0x0ff91, 0x0ff92, 0x0ff93, 0x0ff94, 0x0ff95, 0x0ff96, 0x0ff97,
    0x0ff98, 0x0ff99, 0x0ff9a, 0x0ff9b, 0x0ff9c, 0x0ff9d, 0x0ff9e, 0x0ff9f,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
    U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  },
};

int32_t ucode_kanji1[U_kanji] = {
  0x03000, 0x03001, 0x03002, 0x0ff0c, 0x0ff0e, 0x030fb, 0x0ff1a, 0x0ff1b,
  0x0ff1f, 0x0ff01, 0x0309b, 0x0309c, 0x000b4, 0x0ff40, 0x000a8, 0x0ff3e,
  0x0ffe3, 0x0ff3f, 0x030fd, 0x030fe, 0x0309d, 0x0309e, 0x03003, 0x04edd,
  0x03005, 0x03006, 0x03007, 0x030fc, 0x02014, 0x02010, 0x0ff0f, 0x0ff3c,
  0x0301c, 0x02016, 0x0ff5c, 0x02026, 0x02025, 0x02018, 0x02019, 0x0201c,
  0x0201d, 0x0ff08, 0x0ff09, 0x03014, 0x03015, 0x0ff3b, 0x0ff3d, 0x0ff5b,
  0x0ff5d, 0x03008, 0x03009, 0x0300a, 0x0300b, 0x0300c, 0x0300d, 0x0300e,
  0x0300f, 0x03010, 0x03011, 0x0ff0b, 0x02212, 0x000b1, 0x000d7, 0x000f7,
  0x0ff1d, 0x02260, 0x0ff1c, 0x0ff1e, 0x02266, 0x02267, 0x0221e, 0x02234,
  0x02642, 0x02640, 0x000b0, 0x02032, 0x02033, 0x02103, 0x0ffe5, 0x0ff04,
  0x000a2, 0x000a3, 0x0ff05, 0x0ff03, 0x0ff06, 0x0ff0a, 0x0ff20, 0x000a7,
  0x02606, 0x02605, 0x025cb, 0x025cf, 0x025ce, 0x025c7, 0x025c6, 0x025a1,
  0x025a0, 0x025b3, 0x025b2, 0x025bd, 0x025bc, 0x0203b, 0x03012, 0x02192,
  0x02190, 0x02191, 0x02193, 0x03013, 0x0ff07, 0x0ff02, 0x0ff0d, 0x0ff5e,
  0x03033, 0x03034, 0x03035, 0x0303b, 0x0303c, 0x030ff, 0x0309f, 0x02208,
  0x0220b, 0x02286, 0x02287, 0x02282, 0x02283, 0x0222a, 0x02229, 0x02284,
  0x02285, 0x0228a, 0x0228b, 0x02209, 0x02205, 0x02305, 0x02306, 0x02227,
  0x02228, 0x000ac, 0x021d2, 0x021d4, 0x02200, 0x02203, 0x02295, 0x02296,
  0x02297, 0x02225, 0x02226, 0x0ff5f, 0x0ff60, 0x03018, 0x03019, 0x03016,
  0x03017, 0x02220, 0x022a5, 0x02312, 0x02202, 0x02207, 0x02261, 0x02252,
  0x0226a, 0x0226b, 0x0221a, 0x0223d, 0x0221d, 0x02235, 0x0222b, 0x0222c,
  0x02262, 0x02243, 0x02245, 0x02248, 0x02276, 0x02277, 0x02194, 0x0212b,
  0x02030, 0x0266f, 0x0266d, 0x0266a, 0x02020, 0x02021, 0x000b6, 0x0266e,
  0x0266b, 0x0266c, 0x02669, 0x025ef, 0x025b7, 0x025b6, 0x025c1, 0x025c0,
  0x02197, 0x02198, 0x02196, 0x02199, 0x021c4, 0x021e8, 0x021e6, 0x021e7,
  0x021e9, 0x02934, 0x02935, 0x0ff10, 0x0ff11, 0x0ff12, 0x0ff13, 0x0ff14,
  0x0ff15, 0x0ff16, 0x0ff17, 0x0ff18, 0x0ff19, 0x029bf, 0x025c9, 0x0303d,
  0x0fe46, 0x0fe45, 0x025e6, 0x02022, 0x0ff21, 0x0ff22, 0x0ff23, 0x0ff24,
  0x0ff25, 0x0ff26, 0x0ff27, 0x0ff28, 0x0ff29, 0x0ff2a, 0x0ff2b, 0x0ff2c,
  0x0ff2d, 0x0ff2e, 0x0ff2f, 0x0ff30, 0x0ff31, 0x0ff32, 0x0ff33, 0x0ff34,
  0x0ff35, 0x0ff36, 0x0ff37, 0x0ff38, 0x0ff39, 0x0ff3a, 0x02213, 0x02135,
  0x0210f, 0x033cb, 0x02113, 0x02127, 0x0ff41, 0x0ff42, 0x0ff43, 0x0ff44,
  0x0ff45, 0x0ff46, 0x0ff47, 0x0ff48, 0x0ff49, 0x0ff4a, 0x0ff4b, 0x0ff4c,
  0x0ff4d, 0x0ff4e, 0x0ff4f, 0x0ff50, 0x0ff51, 0x0ff52, 0x0ff53, 0x0ff54,
  0x0ff55, 0x0ff56, 0x0ff57, 0x0ff58, 0x0ff59, 0x0ff5a, 0x030a0, 0x02013,
  0x029fa, 0x029fb, 0x03041, 0x03042, 0x03043, 0x03044, 0x03045, 0x03046,
  0x03047, 0x03048, 0x03049, 0x0304a, 0x0304b, 0x0304c, 0x0304d, 0x0304e,
  0x0304f, 0x03050, 0x03051, 0x03052, 0x03053, 0x03054, 0x03055, 0x03056,
  0x03057, 0x03058, 0x03059, 0x0305a, 0x0305b, 0x0305c, 0x0305d, 0x0305e,
  0x0305f, 0x03060, 0x03061, 0x03062, 0x03063, 0x03064, 0x03065, 0x03066,
  0x03067, 0x03068, 0x03069, 0x0306a, 0x0306b, 0x0306c, 0x0306d, 0x0306e,
  0x0306f, 0x03070, 0x03071, 0x03072, 0x03073, 0x03074, 0x03075, 0x03076,
  0x03077, 0x03078, 0x03079, 0x0307a, 0x0307b, 0x0307c, 0x0307d, 0x0307e,
  0x0307f, 0x03080, 0x03081, 0x03082, 0x03083, 0x03084, 0x03085, 0x03086,
  0x03087, 0x03088, 0x03089, 0x0308a, 0x0308b, 0x0308c, 0x0308d, 0x0308e,
  0x0308f, 0x03090, 0x03091, 0x03092, 0x03093, 0x03094, 0x03095, 0x03096,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  0x030a1, 0x030a2, 0x030a3, 0x030a4, 0x030a5, 0x030a6, 0x030a7, 0x030a8,
  0x030a9, 0x030aa, 0x030ab, 0x030ac, 0x030ad, 0x030ae, 0x030af, 0x030b0,
  0x030b1, 0x030b2, 0x030b3, 0x030b4, 0x030b5, 0x030b6, 0x030b7, 0x030b8,
  0x030b9, 0x030ba, 0x030bb, 0x030bc, 0x030bd, 0x030be, 0x030bf, 0x030c0,
  0x030c1, 0x030c2, 0x030c3, 0x030c4, 0x030c5, 0x030c6, 0x030c7, 0x030c8,
  0x030c9, 0x030ca, 0x030cb, 0x030cc, 0x030cd, 0x030ce, 0x030cf, 0x030d0,
  0x030d1, 0x030d2, 0x030d3, 0x030d4, 0x030d5, 0x030d6, 0x030d7, 0x030d8,
  0x030d9, 0x030da, 0x030db, 0x030dc, 0x030dd, 0x030de, 0x030df, 0x030e0,
  0x030e1, 0x030e2, 0x030e3, 0x030e4, 0x030e5, 0x030e6, 0x030e7, 0x030e8,
  0x030e9, 0x030ea, 0x030eb, 0x030ec, 0x030ed, 0x030ee, 0x030ef, 0x030f0,
  0x030f1, 0x030f2, 0x030f3, 0x030f4, 0x030f5, 0x030f6, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x00391, 0x00392,
  0x00393, 0x00394, 0x00395, 0x00396, 0x00397, 0x00398, 0x00399, 0x0039a,
  0x0039b, 0x0039c, 0x0039d, 0x0039e, 0x0039f, 0x003a0, 0x003a1, 0x003a3,
  0x003a4, 0x003a5, 0x003a6, 0x003a7, 0x003a8, 0x003a9, 0x02664, 0x02660,
  0x02662, 0x02666, 0x02661, 0x02665, 0x02667, 0x02663, 0x003b1, 0x003b2,
  0x003b3, 0x003b4, 0x003b5, 0x003b6, 0x003b7, 0x003b8, 0x003b9, 0x003ba,
  0x003bb, 0x003bc, 0x003bd, 0x003be, 0x003bf, 0x003c0, 0x003c1, 0x003c3,
  0x003c4, 0x003c5, 0x003c6, 0x003c7, 0x003c8, 0x003c9, 0x003c2, 0x024f5,
  0x024f6, 0x024f7, 0x024f8, 0x024f9, 0x024fa, 0x024fb, 0x024fc, 0x024fd,
  0x024fe, 0x02616, 0x02617, 0x03020, 0x0260e, 0x02600, 0x02601, 0x02602,
  0x02603, 0x02668, 0x025b1, 0x031f0, 0x031f1, 0x031f2, 0x031f3, 0x031f4,
  0x031f5, 0x031f6, 0x031f7, 0x031f8, 0x031f9, U_error, 0x031fa, 0x031fb,
  0x031fc, 0x031fd, 0x031fe, 0x031ff, 0x00410, 0x00411, 0x00412, 0x00413,
  0x00414, 0x00415, 0x00401, 0x00416, 0x00417, 0x00418, 0x00419, 0x0041a,
  0x0041b, 0x0041c, 0x0041d, 0x0041e, 0x0041f, 0x00420, 0x00421, 0x00422,
  0x00423, 0x00424, 0x00425, 0x00426, 0x00427, 0x00428, 0x00429, 0x0042a,
  0x0042b, 0x0042c, 0x0042d, 0x0042e, 0x0042f, 0x023be, 0x023bf, 0x023c0,
  0x023c1, 0x023c2, 0x023c3, 0x023c4, 0x023c5, 0x023c6, 0x023c7, 0x023c8,
  0x023c9, 0x023ca, 0x023cb, 0x023cc, 0x00430, 0x00431, 0x00432, 0x00433,
  0x00434, 0x00435, 0x00451, 0x00436, 0x00437, 0x00438, 0x00439, 0x0043a,
  0x0043b, 0x0043c, 0x0043d, 0x0043e, 0x0043f, 0x00440, 0x00441, 0x00442,
  0x00443, 0x00444, 0x00445, 0x00446, 0x00447, 0x00448, 0x00449, 0x0044a,
  0x0044b, 0x0044c, 0x0044d, 0x0044e, 0x0044f, 0x030f7, 0x030f8, 0x030f9,
  0x030fa, 0x022da, 0x022db, 0x02153, 0x02154, 0x02155, 0x02713, 0x02318,
  0x02423, 0x023ce, 0x02500, 0x02502, 0x0250c, 0x02510, 0x02518, 0x02514,
  0x0251c, 0x0252c, 0x02524, 0x02534, 0x0253c, 0x02501, 0x02503, 0x0250f,
  0x02513, 0x0251b, 0x02517, 0x02523, 0x02533, 0x0252b, 0x0253b, 0x0254b,
  0x02520, 0x0252f, 0x02528, 0x02537, 0x0253f, 0x0251d, 0x02530, 0x02525,
  0x02538, 0x02542, 0x03251, 0x03252, 0x03253, 0x03254, 0x03255, 0x03256,
  0x03257, 0x03258, 0x03259, 0x0325a, 0x0325b, 0x0325c, 0x0325d, 0x0325e,
  0x0325f, 0x032b1, 0x032b2, 0x032b3, 0x032b4, 0x032b5, 0x032b6, 0x032b7,
  0x032b8, 0x032b9, 0x032ba, 0x032bb, 0x032bc, 0x032bd, 0x032be, 0x032bf,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  0x025d0, 0x025d1, 0x025d2, 0x025d3, 0x0203c, 0x02047, 0x02048, 0x02049,
  0x001cd, 0x001ce, 0x001d0, 0x01e3e, 0x01e3f, 0x001f8, 0x001f9, 0x001d1,
  0x001d2, 0x001d4, 0x001d6, 0x001d8, 0x001da, 0x001dc, U_error, U_error,
  0x020ac, 0x000a0, 0x000a1, 0x000a4, 0x000a6, 0x000a9, 0x000aa, 0x000ab,
  0x000ad, 0x000ae, 0x000af, 0x000b2, 0x000b3, 0x000b7, 0x000b8, 0x000b9,
  0x000ba, 0x000bb, 0x000bc, 0x000bd, 0x000be, 0x000bf, 0x000c0, 0x000c1,
  0x000c2, 0x000c3, 0x000c4, 0x000c5, 0x000c6, 0x000c7, 0x000c8, 0x000c9,
  0x000ca, 0x000cb, 0x000cc, 0x000cd, 0x000ce, 0x000cf, 0x000d0, 0x000d1,
  0x000d2, 0x000d3, 0x000d4, 0x000d5, 0x000d6, 0x000d8, 0x000d9, 0x000da,
  0x000db, 0x000dc, 0x000dd, 0x000de, 0x000df, 0x000e0, 0x000e1, 0x000e2,
  0x000e3, 0x000e4, 0x000e5, 0x000e6, 0x000e7, 0x000e8, 0x000e9, 0x000ea,
  0x000eb, 0x000ec, 0x000ed, 0x000ee, 0x000ef, 0x000f0, 0x000f1, 0x000f2,
  0x000f3, 0x000f4, 0x000f5, 0x000f6, 0x000f8, 0x000f9, 0x000fa, 0x000fb,
  0x000fc, 0x000fd, 0x000fe, 0x000ff, 0x00100, 0x0012a, 0x0016a, 0x00112,
  0x0014c, 0x00101, 0x0012b, 0x0016b, 0x00113, 0x0014d, 0x00104, 0x002d8,
  0x00141, 0x0013d, 0x0015a, 0x00160, 0x0015e, 0x00164, 0x00179, 0x0017d,
  0x0017b, 0x00105, 0x002db, 0x00142, 0x0013e, 0x0015b, 0x002c7, 0x00161,
  0x0015f, 0x00165, 0x0017a, 0x002dd, 0x0017e, 0x0017c, 0x00154, 0x00102,
  0x00139, 0x00106, 0x0010c, 0x00118, 0x0011a, 0x0010e, 0x00143, 0x00147,
  0x00150, 0x00158, 0x0016e, 0x00170, 0x00162, 0x00155, 0x00103, 0x0013a,
  0x00107, 0x0010d, 0x00119, 0x0011b, 0x0010f, 0x00111, 0x00144, 0x00148,
  0x00151, 0x00159, 0x0016f, 0x00171, 0x00163, 0x002d9, 0x00108, 0x0011c,
  0x00124, 0x00134, 0x0015c, 0x0016c, 0x00109, 0x0011d, 0x00125, 0x00135,
  0x0015d, 0x0016d, 0x00271, 0x0028b, 0x0027e, 0x00283, 0x00292, 0x0026c,
  0x0026e, 0x00279, 0x00288, 0x00256, 0x00273, 0x0027d, 0x00282, 0x00290,
  0x0027b, 0x0026d, 0x0025f, 0x00272, 0x0029d, 0x0028e, 0x00261, 0x0014b,
  0x00270, 0x00281, 0x00127, 0x00295, 0x00294, 0x00266, 0x00298, 0x001c2,
  0x00253, 0x00257, 0x00284, 0x00260, 0x00193, 0x00153, 0x00152, 0x00268,
  0x00289, 0x00258, 0x00275, 0x00259, 0x0025c, 0x0025e, 0x00250, 0x0026f,
  0x0028a, 0x00264, 0x0028c, 0x00254, 0x00251, 0x00252, 0x0028d, 0x00265,
  0x002a2, 0x002a1, 0x00255, 0x00291, 0x0027a, 0x00267, 0x0025a, U_error,
  0x001fd, 0x01f70, 0x01f71, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, 0x01f72, 0x01f73, 0x00361, 0x002c8, 0x002cc,
  0x002d0, 0x002d1, 0x00306, 0x0203f, 0x0030b, 0x00301, 0x00304, 0x00300,
  0x0030f, 0x0030c, 0x00302, 0x002e5, 0x002e6, 0x002e7, 0x002e8, 0x002e9,
  U_error, U_error, 0x00325, 0x0032c, 0x00339, 0x0031c, 0x0031f, 0x00320,
  0x00308, 0x0033d, 0x00329, 0x0032f, 0x002de, 0x00324, 0x00330, 0x0033c,
  0x00334, 0x0031d, 0x0031e, 0x00318, 0x00319, 0x0032a, 0x0033a, 0x0033b,
  0x00303, 0x0031a, 0x02776, 0x02777, 0x02778, 0x02779, 0x0277a, 0x0277b,
  0x0277c, 0x0277d, 0x0277e, 0x0277f, 0x024eb, 0x024ec, 0x024ed, 0x024ee,
  0x024ef, 0x024f0, 0x024f1, 0x024f2, 0x024f3, 0x024f4, 0x02170, 0x02171,
  0x02172, 0x02173, 0x02174, 0x02175, 0x02176, 0x02177, 0x02178, 0x02179,
  0x0217a, 0x0217b, 0x024d0, 0x024d1, 0x024d2, 0x024d3, 0x024d4, 0x024d5,
  0x024d6, 0x024d7, 0x024d8, 0x024d9, 0x024da, 0x024db, 0x024dc, 0x024dd,
  0x024de, 0x024df, 0x024e0, 0x024e1, 0x024e2, 0x024e3, 0x024e4, 0x024e5,
  0x024e6, 0x024e7, 0x024e8, 0x024e9, 0x032d0, 0x032d1, 0x032d2, 0x032d3,
  0x032d4, 0x032d5, 0x032d6, 0x032d7, 0x032d8, 0x032d9, 0x032da, 0x032db,
  0x032dc, 0x032dd, 0x032de, 0x032df, 0x032e0, 0x032e1, 0x032e2, 0x032e3,
  0x032fa, 0x032e9, 0x032e5, 0x032ed, 0x032ec, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x02051, 0x02042,
  0x02460, 0x02461, 0x02462, 0x02463, 0x02464, 0x02465, 0x02466, 0x02467,
  0x02468, 0x02469, 0x0246a, 0x0246b, 0x0246c, 0x0246d, 0x0246e, 0x0246f,
  0x02470, 0x02471, 0x02472, 0x02473, 0x02160, 0x02161, 0x02162, 0x02163,
  0x02164, 0x02165, 0x02166, 0x02167, 0x02168, 0x02169, 0x0216a, 0x03349,
  0x03314, 0x03322, 0x0334d, 0x03318, 0x03327, 0x03303, 0x03336, 0x03351,
  0x03357, 0x0330d, 0x03326, 0x03323, 0x0332b, 0x0334a, 0x0333b, 0x0339c,
  0x0339d, 0x0339e, 0x0338e, 0x0338f, 0x033c4, 0x033a1, 0x0216b, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x0337b, 0x0301d,
  0x0301f, 0x02116, 0x033cd, 0x02121, 0x032a4, 0x032a5, 0x032a6, 0x032a7,
  0x032a8, 0x03231, 0x03232, 0x03239, 0x0337e, 0x0337d, 0x0337c, U_error,
  U_error, U_error, 0x0222e, U_error, U_error, U_error, U_error, 0x0221f,
  0x022bf, U_error, U_error, U_error, 0x02756, 0x0261e, 0x04ff1, 0x2000b,
  0x03402, 0x04e28, 0x04e2f, 0x04e30, 0x04e8d, 0x04ee1, 0x04efd, 0x04eff,
  0x04f03, 0x04f0b, 0x04f60, 0x04f48, 0x04f49, 0x04f56, 0x04f5f, 0x04f6a,
  0x04f6c, 0x04f7e, 0x04f8a, 0x04f94, 0x04f97, 0x0fa30, 0x04fc9, 0x04fe0,
  0x05001, 0x05002, 0x0500e, 0x05018, 0x05027, 0x0502e, 0x05040, 0x0503b,
  0x05041, 0x05094, 0x050cc, 0x050f2, 0x050d0, 0x050e6, 0x0fa31, 0x05106,
  0x05103, 0x0510b, 0x0511e, 0x05135, 0x0514a, 0x0fa32, 0x05155, 0x05157,
  0x034b5, 0x0519d, 0x051c3, 0x051ca, 0x051de, 0x051e2, 0x051ee, 0x05201,
  0x034db, 0x05213, 0x05215, 0x05249, 0x05257, 0x05261, 0x05293, 0x052c8,
  0x0fa33, 0x052cc, 0x052d0, 0x052d6, 0x052db, 0x0fa34, 0x052f0, 0x052fb,
  0x05300, 0x05307, 0x0531c, 0x0fa35, 0x05361, 0x05363, 0x0537d, 0x05393,
  0x0539d, 0x053b2, 0x05412, 0x05427, 0x0544d, 0x0549c, 0x0546b, 0x05474,
  0x0547f, 0x05488, 0x05496, 0x054a1, 0x054a9, 0x054c6, 0x054ff, 0x0550e,
  0x0552b, 0x05535, 0x05550, 0x0555e, 0x05581, 0x05586, 0x0558e, 0x0fa36,
  0x055ad, 0x055ce, 0x0fa37, 0x05608, 0x0560e, 0x0563b, 0x05649, 0x05676,
  0x05666, 0x0fa38, 0x0566f, 0x05671, 0x05672, 0x05699, 0x0569e, 0x056a9,
  0x056ac, 0x056b3, 0x056c9, 0x056ca, 0x0570a, 0x2123d, 0x05721, 0x0572f,
  0x05733, 0x05734, 0x05770, 0x05777, 0x0577c, 0x0579c, 0x0fa0f, 0x2131b,
  0x057b8, 0x057c7, 0x057c8, 0x057cf, 0x057e4, 0x057ed, 0x057f5, 0x057f6,
  0x057ff, 0x05809, 0x0fa10, 0x05861, 0x05864, 0x0fa39, 0x0587c, 0x05889,
  0x0589e, 0x0fa3a, 0x058a9, 0x2146e, 0x058d2, 0x058ce, 0x058d4, 0x058da,
  0x058e0, 0x058e9, 0x0590c, 0x08641, 0x0595d, 0x0596d, 0x0598b, 0x05992,
  0x059a4, 0x059c3, 0x059d2, 0x059dd, 0x05a13, 0x05a23, 0x05a67, 0x05a6d,
  0x05a77, 0x05a7e, 0x05a84, 0x05a9e, 0x05aa7, 0x05ac4, 0x218bd, 0x05b19,
  0x05b25, 0x0525d, 0x04e9c, 0x05516, 0x05a03, 0x0963f, 0x054c0, 0x0611b,
  0x06328, 0x059f6, 0x09022, 0x08475, 0x0831c, 0x07a50, 0x060aa, 0x063e1,
  0x06e25, 0x065ed, 0x08466, 0x082a6, 0x09bf5, 0x06893, 0x05727, 0x065a1,
  0x06271, 0x05b9b, 0x059d0, 0x0867b, 0x098f4, 0x07d62, 0x07dbe, 0x09b8e,
  0x06216, 0x07c9f, 0x088b7, 0x05b89, 0x05eb5, 0x06309, 0x06697, 0x06848,
  0x095c7, 0x0978d, 0x0674f, 0x04ee5, 0x04f0a, 0x04f4d, 0x04f9d, 0x05049,
  0x056f2, 0x05937, 0x059d4, 0x05a01, 0x05c09, 0x060df, 0x0610f, 0x06170,
  0x06613, 0x06905, 0x070ba, 0x0754f, 0x07570, 0x079fb, 0x07dad, 0x07def,
  0x080c3, 0x0840e, 0x08863, 0x08b02, 0x09055, 0x0907a, 0x0533b, 0x04e95,
  0x04ea5, 0x057df, 0x080b2, 0x090c1, 0x078ef, 0x04e00, 0x058f1, 0x06ea2,
  0x09038, 0x07a32, 0x08328, 0x0828b, 0x09c2f, 0x05141, 0x05370, 0x054bd,
  0x054e1, 0x056e0, 0x059fb, 0x05f15, 0x098f2, 0x06deb, 0x080e4, 0x0852d,
  0x09662, 0x09670, 0x096a0, 0x097fb, 0x0540b, 0x053f3, 0x05b87, 0x070cf,
  0x07fbd, 0x08fc2, 0x096e8, 0x0536f, 0x09d5c, 0x07aba, 0x04e11, 0x07893,
  0x081fc, 0x06e26, 0x05618, 0x05504, 0x06b1d, 0x0851a, 0x09c3b, 0x059e5,
  0x053a9, 0x06d66, 0x074dc, 0x0958f, 0x05642, 0x04e91, 0x0904b, 0x096f2,
  0x0834f, 0x0990c, 0x053e1, 0x055b6, 0x05b30, 0x05f71, 0x06620, 0x066f3,
  0x06804, 0x06c38, 0x06cf3, 0x06d29, 0x0745b, 0x076c8, 0x07a4e, 0x09834,
  0x082f1, 0x0885b, 0x08a60, 0x092ed, 0x06db2, 0x075ab, 0x076ca, 0x099c5,
  0x060a6, 0x08b01, 0x08d8a, 0x095b2, 0x0698e, 0x053ad, 0x05186, 0x05712,
  0x05830, 0x05944, 0x05bb4, 0x05ef6, 0x06028, 0x063a9, 0x063f4, 0x06cbf,
  0x06f14, 0x0708e, 0x07114, 0x07159, 0x071d5, 0x0733f, 0x07e01, 0x08276,
  0x082d1, 0x08597, 0x09060, 0x0925b, 0x09d1b, 0x05869, 0x065bc, 0x06c5a,
  0x07525, 0x051f9, 0x0592e, 0x05965, 0x05f80, 0x05fdc, 0x062bc, 0x065fa,
  0x06a2a, 0x06b27, 0x06bb4, 0x0738b, 0x07fc1, 0x08956, 0x09d2c, 0x09d0e,
  0x09ec4, 0x05ca1, 0x06c96, 0x0837b, 0x05104, 0x05c4b, 0x061b6, 0x081c6,
  0x06876, 0x07261, 0x04e59, 0x04ffa, 0x05378, 0x06069, 0x06e29, 0x07a4f,
  0x097f3, 0x04e0b, 0x05316, 0x04eee, 0x04f55, 0x04f3d, 0x04fa1, 0x04f73,
  0x052a0, 0x053ef, 0x05609, 0x0590f, 0x05ac1, 0x05bb6, 0x05be1, 0x079d1,
  0x06687, 0x0679c, 0x067b6, 0x06b4c, 0x06cb3, 0x0706b, 0x073c2, 0x0798d,
  0x079be, 0x07a3c, 0x07b87, 0x082b1, 0x082db, 0x08304, 0x08377, 0x083ef,
  0x083d3, 0x08766, 0x08ab2, 0x05629, 0x08ca8, 0x08fe6, 0x0904e, 0x0971e,
  0x0868a, 0x04fc4, 0x05ce8, 0x06211, 0x07259, 0x0753b, 0x081e5, 0x082bd,
  0x086fe, 0x08cc0, 0x096c5, 0x09913, 0x099d5, 0x04ecb, 0x04f1a, 0x089e3,
  0x056de, 0x0584a, 0x058ca, 0x05efb, 0x05feb, 0x0602a, 0x06094, 0x06062,
  0x061d0, 0x06212, 0x062d0, 0x06539, 0x09b41, 0x06666, 0x068b0, 0x06d77,
  0x07070, 0x0754c, 0x07686, 0x07d75, 0x082a5, 0x087f9, 0x0958b, 0x0968e,
  0x08c9d, 0x051f1, 0x052be, 0x05916, 0x054b3, 0x05bb3, 0x05d16, 0x06168,
  0x06982, 0x06daf, 0x0788d, 0x084cb, 0x08857, 0x08a72, 0x093a7, 0x09ab8,
  0x06d6c, 0x099a8, 0x086d9, 0x057a3, 0x067ff, 0x086ce, 0x0920e, 0x05283,
  0x05687, 0x05404, 0x05ed3, 0x062e1, 0x064b9, 0x0683c, 0x06838, 0x06bbb,
  0x07372, 0x078ba, 0x07a6b, 0x0899a, 0x089d2, 0x08d6b, 0x08f03, 0x090ed,
  0x095a3, 0x09694, 0x09769, 0x05b66, 0x05cb3, 0x0697d, 0x0984d, 0x0984e,
  0x0639b, 0x07b20, 0x06a2b, 0x06a7f, 0x068b6, 0x09c0d, 0x06f5f, 0x05272,
  0x0559d, 0x06070, 0x062ec, 0x06d3b, 0x06e07, 0x06ed1, 0x0845b, 0x08910,
  0x08f44, 0x04e14, 0x09c39, 0x053f6, 0x0691b, 0x06a3a, 0x09784, 0x0682a,
  0x0515c, 0x07ac3, 0x084b2, 0x091dc, 0x0938c, 0x0565b, 0x09d28, 0x06822,
  0x08305, 0x08431, 0x07ca5, 0x05208, 0x082c5, 0x074e6, 0x04e7e, 0x04f83,
  0x051a0, 0x05bd2, 0x0520a, 0x052d8, 0x052e7, 0x05dfb, 0x0559a, 0x0582a,
  0x059e6, 0x05b8c, 0x05b98, 0x05bdb, 0x05e72, 0x05e79, 0x060a3, 0x0611f,
  0x06163, 0x061be, 0x063db, 0x06562, 0x067d1, 0x06853, 0x068fa, 0x06b3e,
  0x06b53, 0x06c57, 0x06f22, 0x06f97, 0x06f45, 0x074b0, 0x07518, 0x076e3,
  0x0770b, 0x07aff, 0x07ba1, 0x07c21, 0x07de9, 0x07f36, 0x07ff0, 0x0809d,
  0x08266, 0x0839e, 0x089b3, 0x08acc, 0x08cab, 0x09084, 0x09451, 0x09593,
  0x09591, 0x095a2, 0x09665, 0x097d3, 0x09928, 0x08218, 0x04e38, 0x0542b,
  0x05cb8, 0x05dcc, 0x073a9, 0x0764c, 0x0773c, 0x05ca9, 0x07feb, 0x08d0b,
  0x096c1, 0x09811, 0x09854, 0x09858, 0x04f01, 0x04f0e, 0x05371, 0x0559c,
  0x05668, 0x057fa, 0x05947, 0x05b09, 0x05bc4, 0x05c90, 0x05e0c, 0x05e7e,
  0x05fcc, 0x063ee, 0x0673a, 0x065d7, 0x065e2, 0x0671f, 0x068cb, 0x068c4,
  0x06a5f, 0x05e30, 0x06bc5, 0x06c17, 0x06c7d, 0x0757f, 0x07948, 0x05b63,
  0x07a00, 0x07d00, 0x05fbd, 0x0898f, 0x08a18, 0x08cb4, 0x08d77, 0x08ecc,
  0x08f1d, 0x098e2, 0x09a0e, 0x09b3c, 0x04e80, 0x0507d, 0x05100, 0x05993,
  0x05b9c, 0x0622f, 0x06280, 0x064ec, 0x06b3a, 0x072a0, 0x07591, 0x07947,
  0x07fa9, 0x087fb, 0x08abc, 0x08b70, 0x063ac, 0x083ca, 0x097a0, 0x05409,
  0x05403, 0x055ab, 0x06854, 0x06a58, 0x08a70, 0x07827, 0x06775, 0x09ecd,
  0x05374, 0x05ba2, 0x0811a, 0x08650, 0x09006, 0x04e18, 0x04e45, 0x04ec7,
  0x04f11, 0x053ca, 0x05438, 0x05bae, 0x05f13, 0x06025, 0x06551, 0x0673d,
  0x06c42, 0x06c72, 0x06ce3, 0x07078, 0x07403, 0x07a76, 0x07aae, 0x07b08,
  0x07d1a, 0x07cfe, 0x07d66, 0x065e7, 0x0725b, 0x053bb, 0x05c45, 0x05de8,
  0x062d2, 0x062e0, 0x06319, 0x06e20, 0x0865a, 0x08a31, 0x08ddd, 0x092f8,
  0x06f01, 0x079a6, 0x09b5a, 0x04ea8, 0x04eab, 0x04eac, 0x04f9b, 0x04fa0,
  0x050d1, 0x05147, 0x07af6, 0x05171, 0x051f6, 0x05354, 0x05321, 0x0537f,
  0x053eb, 0x055ac, 0x05883, 0x05ce1, 0x05f37, 0x05f4a, 0x0602f, 0x06050,
  0x0606d, 0x0631f, 0x06559, 0x06a4b, 0x06cc1, 0x072c2, 0x072ed, 0x077ef,
  0x080f8, 0x08105, 0x08208, 0x0854e, 0x090f7, 0x093e1, 0x097ff, 0x09957,
  0x09a5a, 0x04ef0, 0x051dd, 0x05c2d, 0x06681, 0x0696d, 0x05c40, 0x066f2,
  0x06975, 0x07389, 0x06850, 0x07c81, 0x050c5, 0x052e4, 0x05747, 0x05dfe,
  0x09326, 0x065a4, 0x06b23, 0x06b3d, 0x07434, 0x07981, 0x079bd, 0x07b4b,
  0x07dca, 0x082b9, 0x083cc, 0x0887f, 0x0895f, 0x08b39, 0x08fd1, 0x091d1,
  0x0541f, 0x09280, 0x04e5d, 0x05036, 0x053e5, 0x0533a, 0x072d7, 0x07396,
  0x077e9, 0x082e6, 0x08eaf, 0x099c6, 0x099c8, 0x099d2, 0x05177, 0x0611a,
  0x0865e, 0x055b0, 0x07a7a, 0x05076, 0x05bd3, 0x09047, 0x09685, 0x04e32,
  0x06adb, 0x091e7, 0x05c51, 0x05c48, 0x06398, 0x07a9f, 0x06c93, 0x09774,
  0x08f61, 0x07aaa, 0x0718a, 0x09688, 0x07c82, 0x06817, 0x07e70, 0x06851,
  0x0936c, 0x052f2, 0x0541b, 0x085ab, 0x08a13, 0x07fa4, 0x08ecd, 0x090e1,
  0x05366, 0x08888, 0x07941, 0x04fc2, 0x050be, 0x05211, 0x05144, 0x05553,
  0x0572d, 0x073ea, 0x0578b, 0x05951, 0x05f62, 0x05f84, 0x06075, 0x06176,
  0x06167, 0x061a9, 0x063b2, 0x0643a, 0x0656c, 0x0666f, 0x06842, 0x06e13,
  0x07566, 0x07a3d, 0x07cfb, 0x07d4c, 0x07d99, 0x07e4b, 0x07f6b, 0x0830e,
  0x0834a, 0x086cd, 0x08a08, 0x08a63, 0x08b66, 0x08efd, 0x0981a, 0x09d8f,
  0x082b8, 0x08fce, 0x09be8, 0x05287, 0x0621f, 0x06483, 0x06fc0, 0x09699,
  0x06841, 0x05091, 0x06b20, 0x06c7a, 0x06f54, 0x07a74, 0x07d50, 0x08840,
  0x08a23, 0x06708, 0x04ef6, 0x05039, 0x05026, 0x05065, 0x0517c, 0x05238,
  0x05263, 0x055a7, 0x0570f, 0x05805, 0x05acc, 0x05efa, 0x061b2, 0x061f8,
  0x062f3, 0x06372, 0x0691c, 0x06a29, 0x0727d, 0x072ac, 0x0732e, 0x07814,
  0x0786f, 0x07d79, 0x0770c, 0x080a9, 0x0898b, 0x08b19, 0x08ce2, 0x08ed2,
  0x09063, 0x09375, 0x0967a, 0x09855, 0x09a13, 0x09e78, 0x05143, 0x0539f,
  0x053b3, 0x05e7b, 0x05f26, 0x06e1b, 0x06e90, 0x07384, 0x073fe, 0x07d43,
  0x08237, 0x08a00, 0x08afa, 0x09650, 0x04e4e, 0x0500b, 0x053e4, 0x0547c,
  0x056fa, 0x059d1, 0x05b64, 0x05df1, 0x05eab, 0x05f27, 0x06238, 0x06545,
  0x067af, 0x06e56, 0x072d0, 0x07cca, 0x088b4, 0x080a1, 0x080e1, 0x083f0,
  0x0864e, 0x08a87, 0x08de8, 0x09237, 0x096c7, 0x09867, 0x09f13, 0x04e94,
  0x04e92, 0x04f0d, 0x05348, 0x05449, 0x0543e, 0x05a2f, 0x05f8c, 0x05fa1,
  0x0609f, 0x068a7, 0x06a8e, 0x0745a, 0x07881, 0x08a9e, 0x08aa4, 0x08b77,
  0x09190, 0x04e5e, 0x09bc9, 0x04ea4, 0x04f7c, 0x04faf, 0x05019, 0x05016,
  0x05149, 0x0516c, 0x0529f, 0x052b9, 0x052fe, 0x0539a, 0x053e3, 0x05411,
  0x0540e, 0x05589, 0x05751, 0x057a2, 0x0597d, 0x05b54, 0x05b5d, 0x05b8f,
  0x05de5, 0x05de7, 0x05df7, 0x05e78, 0x05e83, 0x05e9a, 0x05eb7, 0x05f18,
  0x06052, 0x0614c, 0x06297, 0x062d8, 0x063a7, 0x0653b, 0x06602, 0x06643,
  0x066f4, 0x0676d, 0x06821, 0x06897, 0x069cb, 0x06c5f, 0x06d2a, 0x06d69,
  0x06e2f, 0x06e9d, 0x07532, 0x07687, 0x0786c, 0x07a3f, 0x07ce0, 0x07d05,
  0x07d18, 0x07d5e, 0x07db1, 0x08015, 0x08003, 0x080af, 0x080b1, 0x08154,
  0x0818f, 0x0822a, 0x08352, 0x0884c, 0x08861, 0x08b1b, 0x08ca2, 0x08cfc,
  0x090ca, 0x09175, 0x09271, 0x0783f, 0x092fc, 0x095a4, 0x0964d, 0x09805,
  0x09999, 0x09ad8, 0x09d3b, 0x0525b, 0x052ab, 0x053f7, 0x05408, 0x058d5,
  0x062f7, 0x06fe0, 0x08c6a, 0x08f5f, 0x09eb9, 0x0514b, 0x0523b, 0x0544a,
  0x056fd, 0x07a40, 0x09177, 0x09d60, 0x09ed2, 0x07344, 0x06f09, 0x08170,
  0x07511, 0x05ffd, 0x060da, 0x09aa8, 0x072db, 0x08fbc, 0x06b64, 0x09803,
  0x04eca, 0x056f0, 0x05764, 0x058be, 0x05a5a, 0x06068, 0x061c7, 0x0660f,
  0x06606, 0x06839, 0x068b1, 0x06df7, 0x075d5, 0x07d3a, 0x0826e, 0x09b42,
  0x04e9b, 0x04f50, 0x053c9, 0x05506, 0x05d6f, 0x05de6, 0x05dee, 0x067fb,
  0x06c99, 0x07473, 0x07802, 0x08a50, 0x09396, 0x088df, 0x05750, 0x05ea7,
  0x0632b, 0x050b5, 0x050ac, 0x0518d, 0x06700, 0x054c9, 0x0585e, 0x059bb,
  0x05bb0, 0x05f69, 0x0624d, 0x063a1, 0x0683d, 0x06b73, 0x06e08, 0x0707d,
  0x091c7, 0x07280, 0x07815, 0x07826, 0x0796d, 0x0658e, 0x07d30, 0x083dc,
  0x088c1, 0x08f09, 0x0969b, 0x05264, 0x05728, 0x06750, 0x07f6a, 0x08ca1,
  0x051b4, 0x05742, 0x0962a, 0x0583a, 0x0698a, 0x080b4, 0x054b2, 0x05d0e,
  0x057fc, 0x07895, 0x09dfa, 0x04f5c, 0x0524a, 0x0548b, 0x0643e, 0x06628,
  0x06714, 0x067f5, 0x07a84, 0x07b56, 0x07d22, 0x0932f, 0x0685c, 0x09bad,
  0x07b39, 0x05319, 0x0518a, 0x05237, 0x05bdf, 0x062f6, 0x064ae, 0x064e6,
  0x0672d, 0x06bba, 0x085a9, 0x096d1, 0x07690, 0x09bd6, 0x0634c, 0x09306,
  0x09bab, 0x076bf, 0x06652, 0x04e09, 0x05098, 0x053c2, 0x05c71, 0x060e8,
  0x06492, 0x06563, 0x0685f, 0x071e6, 0x073ca, 0x07523, 0x07b97, 0x07e82,
  0x08695, 0x08b83, 0x08cdb, 0x09178, 0x09910, 0x065ac, 0x066ab, 0x06b8b,
  0x04ed5, 0x04ed4, 0x04f3a, 0x04f7f, 0x0523a, 0x053f8, 0x053f2, 0x055e3,
  0x056db, 0x058eb, 0x059cb, 0x059c9, 0x059ff, 0x05b50, 0x05c4d, 0x05e02,
  0x05e2b, 0x05fd7, 0x0601d, 0x06307, 0x0652f, 0x05b5c, 0x065af, 0x065bd,
  0x065e8, 0x0679d, 0x06b62, 0x06b7b, 0x06c0f, 0x07345, 0x07949, 0x079c1,
  0x07cf8, 0x07d19, 0x07d2b, 0x080a2, 0x08102, 0x081f3, 0x08996, 0x08a5e,
  0x08a69, 0x08a66, 0x08a8c, 0x08aee, 0x08cc7, 0x08cdc, 0x096cc, 0x098fc,
  0x06b6f, 0x04e8b, 0x04f3c, 0x04f8d, 0x05150, 0x05b57, 0x05bfa, 0x06148,
  0x06301, 0x06642, 0x06b21, 0x06ecb, 0x06cbb, 0x0723e, 0x074bd, 0x075d4,
  0x078c1, 0x0793a, 0x0800c, 0x08033, 0x081ea, 0x08494, 0x08f9e, 0x06c50,
  0x09e7f, 0x05f0f, 0x08b58, 0x09d2b, 0x07afa, 0x08ef8, 0x05b8d, 0x096eb,
  0x04e03, 0x053f1, 0x057f7, 0x05931, 0x05ac9, 0x05ba4, 0x06089, 0x06e7f,
  0x06f06, 0x075be, 0x08cea, 0x05b9f, 0x08500, 0x07be0, 0x05072, 0x067f4,
  0x0829d, 0x05c61, 0x0854a, 0x07e1e, 0x0820e, 0x05199, 0x05c04, 0x06368,
  0x08d66, 0x0659c, 0x0716e, 0x0793e, 0x07d17, 0x08005, 0x08b1d, 0x08eca,
  0x0906e, 0x086c7, 0x090aa, 0x0501f, 0x052fa, 0x05c3a, 0x06753, 0x0707c,
  0x07235, 0x0914c, 0x091c8, 0x0932b, 0x082e5, 0x05bc2, 0x05f31, 0x060f9,
  0x04e3b, 0x053d6, 0x05b88, 0x0624b, 0x06731, 0x06b8a, 0x072e9, 0x073e0,
  0x07a2e, 0x0816b, 0x08da3, 0x09152, 0x09996, 0x05112, 0x053d7, 0x0546a,
  0x05bff, 0x06388, 0x06a39, 0x07dac, 0x09700, 0x056da, 0x053ce, 0x05468,
  0x05b97, 0x05c31, 0x05dde, 0x04fee, 0x06101, 0x062fe, 0x06d32, 0x079c0,
  0x079cb, 0x07d42, 0x07e4d, 0x07fd2, 0x081ed, 0x0821f, 0x08490, 0x08846,
  0x08972, 0x08b90, 0x08e74, 0x08f2f, 0x09031, 0x0914b, 0x0916c, 0x096c6,
  0x0919c, 0x04ec0, 0x04f4f, 0x05145, 0x05341, 0x05f93, 0x0620e, 0x067d4,
  0x06c41, 0x06e0b, 0x07363, 0x07e26, 0x091cd, 0x09283, 0x053d4, 0x05919,
  0x05bbf, 0x06dd1, 0x0795d, 0x07e2e, 0x07c9b, 0x0587e, 0x0719f, 0x051fa,
  0x08853, 0x08ff0, 0x04fca, 0x05cfb, 0x06625, 0x077ac, 0x07ae3, 0x0821c,
  0x099ff, 0x051c6, 0x05faa, 0x065ec, 0x0696f, 0x06b89, 0x06df3, 0x06e96,
  0x06f64, 0x076fe, 0x07d14, 0x05de1, 0x09075, 0x09187, 0x09806, 0x051e6,
  0x0521d, 0x06240, 0x06691, 0x066d9, 0x06e1a, 0x05eb6, 0x07dd2, 0x07f72,
  0x066f8, 0x085af, 0x085f7, 0x08af8, 0x052a9, 0x053d9, 0x05973, 0x05e8f,
  0x05f90, 0x06055, 0x092e4, 0x09664, 0x050b7, 0x0511f, 0x052dd, 0x05320,
  0x05347, 0x053ec, 0x054e8, 0x05546, 0x05531, 0x05617, 0x05968, 0x059be,
  0x05a3c, 0x05bb5, 0x05c06, 0x05c0f, 0x05c11, 0x05c1a, 0x05e84, 0x05e8a,
  0x05ee0, 0x05f70, 0x0627f, 0x06284, 0x062db, 0x0638c, 0x06377, 0x06607,
  0x0660c, 0x0662d, 0x06676, 0x0677e, 0x068a2, 0x06a1f, 0x06a35, 0x06cbc,
  0x06d88, 0x06e09, 0x06e58, 0x0713c, 0x07126, 0x07167, 0x075c7, 0x07701,
  0x0785d, 0x07901, 0x07965, 0x079f0, 0x07ae0, 0x07b11, 0x07ca7, 0x07d39,
  0x08096, 0x083d6, 0x0848b, 0x08549, 0x0885d, 0x088f3, 0x08a1f, 0x08a3c,
  0x08a54, 0x08a73, 0x08c61, 0x08cde, 0x091a4, 0x09266, 0x0937e, 0x09418,
  0x0969c, 0x09798, 0x04e0a, 0x04e08, 0x04e1e, 0x04e57, 0x05197, 0x05270,
  0x057ce, 0x05834, 0x058cc, 0x05b22, 0x05e38, 0x060c5, 0x064fe, 0x06761,
  0x06756, 0x06d44, 0x072b6, 0x07573, 0x07a63, 0x084b8, 0x08b72, 0x091b8,
  0x09320, 0x05631, 0x057f4, 0x098fe, 0x062ed, 0x0690d, 0x06b96, 0x071ed,
  0x07e54, 0x08077, 0x08272, 0x089e6, 0x098df, 0x08755, 0x08fb1, 0x05c3b,
  0x04f38, 0x04fe1, 0x04fb5, 0x05507, 0x05a20, 0x05bdd, 0x05be9, 0x05fc3,
  0x0614e, 0x0632f, 0x065b0, 0x0664b, 0x068ee, 0x0699b, 0x06d78, 0x06df1,
  0x07533, 0x075b9, 0x0771f, 0x0795e, 0x079e6, 0x07d33, 0x081e3, 0x082af,
  0x085aa, 0x089aa, 0x08a3a, 0x08eab, 0x08f9b, 0x09032, 0x091dd, 0x09707,
  0x04eba, 0x04ec1, 0x05203, 0x05875, 0x058ec, 0x05c0b, 0x0751a, 0x05c3d,
  0x0814e, 0x08a0a, 0x08fc5, 0x09663, 0x0976d, 0x07b25, 0x08acf, 0x09808,
  0x09162, 0x056f3, 0x053a8, 0x09017, 0x05439, 0x05782, 0x05e25, 0x063a8,
  0x06c34, 0x0708a, 0x07761, 0x07c8b, 0x07fe0, 0x08870, 0x09042, 0x09154,
  0x09310, 0x09318, 0x0968f, 0x0745e, 0x09ac4, 0x05d07, 0x05d69, 0x06570,
  0x067a2, 0x08da8, 0x096db, 0x0636e, 0x06749, 0x06919, 0x083c5, 0x09817,
  0x096c0, 0x088fe, 0x06f84, 0x0647a, 0x05bf8, 0x04e16, 0x0702c, 0x0755d,
  0x0662f, 0x051c4, 0x05236, 0x052e2, 0x059d3, 0x05f81, 0x06027, 0x06210,
  0x0653f, 0x06574, 0x0661f, 0x06674, 0x068f2, 0x06816, 0x06b63, 0x06e05,
  0x07272, 0x0751f, 0x076db, 0x07cbe, 0x08056, 0x058f0, 0x088fd, 0x0897f,
  0x08aa0, 0x08a93, 0x08acb, 0x0901d, 0x09192, 0x09752, 0x09759, 0x06589,
  0x07a0e, 0x08106, 0x096bb, 0x05e2d, 0x060dc, 0x0621a, 0x065a5, 0x06614,
  0x06790, 0x077f3, 0x07a4d, 0x07c4d, 0x07e3e, 0x0810a, 0x08cac, 0x08d64,
  0x08de1, 0x08e5f, 0x078a9, 0x05207, 0x062d9, 0x063a5, 0x06442, 0x06298,
  0x08a2d, 0x07a83, 0x07bc0, 0x08aac, 0x096ea, 0x07d76, 0x0820c, 0x08749,
  0x04ed9, 0x05148, 0x05343, 0x05360, 0x05ba3, 0x05c02, 0x05c16, 0x05ddd,
  0x06226, 0x06247, 0x064b0, 0x06813, 0x06834, 0x06cc9, 0x06d45, 0x06d17,
  0x067d3, 0x06f5c, 0x0714e, 0x0717d, 0x065cb, 0x07a7f, 0x07bad, 0x07dda,
  0x07e4a, 0x07fa8, 0x0817a, 0x0821b, 0x08239, 0x085a6, 0x08a6e, 0x08cce,
  0x08df5, 0x09078, 0x09077, 0x092ad, 0x09291, 0x09583, 0x09bae, 0x0524d,
  0x05584, 0x06f38, 0x07136, 0x05168, 0x07985, 0x07e55, 0x081b3, 0x07cce,
  0x0564c, 0x05851, 0x05ca8, 0x063aa, 0x066fe, 0x066fd, 0x0695a, 0x072d9,
  0x0758f, 0x0758e, 0x0790e, 0x07956, 0x079df, 0x07c97, 0x07d20, 0x07d44,
  0x08607, 0x08a34, 0x0963b, 0x09061, 0x09f20, 0x050e7, 0x05275, 0x053cc,
  0x053e2, 0x05009, 0x055aa, 0x058ee, 0x0594f, 0x0723d, 0x05b8b, 0x05c64,
  0x0531d, 0x060e3, 0x060f3, 0x0635c, 0x06383, 0x0633f, 0x063bb, 0x064cd,
  0x065e9, 0x066f9, 0x05de3, 0x069cd, 0x069fd, 0x06f15, 0x071e5, 0x04e89,
  0x075e9, 0x076f8, 0x07a93, 0x07cdf, 0x07dcf, 0x07d9c, 0x08061, 0x08349,
  0x08358, 0x0846c, 0x084bc, 0x085fb, 0x088c5, 0x08d70, 0x09001, 0x0906d,
  0x09397, 0x0971c, 0x09a12, 0x050cf, 0x05897, 0x0618e, 0x081d3, 0x08535,
  0x08d08, 0x09020, 0x04fc3, 0x05074, 0x05247, 0x05373, 0x0606f, 0x06349,
  0x0675f, 0x06e2c, 0x08db3, 0x0901f, 0x04fd7, 0x05c5e, 0x08cca, 0x065cf,
  0x07d9a, 0x05352, 0x08896, 0x05176, 0x063c3, 0x05b58, 0x05b6b, 0x05c0a,
  0x0640d, 0x06751, 0x0905c, 0x04ed6, 0x0591a, 0x0592a, 0x06c70, 0x08a51,
  0x0553e, 0x05815, 0x059a5, 0x060f0, 0x06253, 0x067c1, 0x08235, 0x06955,
  0x09640, 0x099c4, 0x09a28, 0x04f53, 0x05806, 0x05bfe, 0x08010, 0x05cb1,
  0x05e2f, 0x05f85, 0x06020, 0x0614b, 0x06234, 0x066ff, 0x06cf0, 0x06ede,
  0x080ce, 0x0817f, 0x082d4, 0x0888b, 0x08cb8, 0x09000, 0x0902e, 0x0968a,
  0x09edb, 0x09bdb, 0x04ee3, 0x053f0, 0x05927, 0x07b2c, 0x0918d, 0x0984c,
  0x09df9, 0x06edd, 0x07027, 0x05353, 0x05544, 0x05b85, 0x06258, 0x0629e,
  0x062d3, 0x06ca2, 0x06fef, 0x07422, 0x08a17, 0x09438, 0x06fc1, 0x08afe,
  0x08338, 0x051e7, 0x086f8, 0x053ea, 0x053e9, 0x04f46, 0x09054, 0x08fb0,
  0x0596a, 0x08131, 0x05dfd, 0x07aea, 0x08fbf, 0x068da, 0x08c37, 0x072f8,
  0x09c48, 0x06a3d, 0x08ab0, 0x04e39, 0x05358, 0x05606, 0x05766, 0x062c5,
  0x063a2, 0x065e6, 0x06b4e, 0x06de1, 0x06e5b, 0x070ad, 0x077ed, 0x07aef,
  0x07baa, 0x07dbb, 0x0803d, 0x080c6, 0x086cb, 0x08a95, 0x0935b, 0x056e3,
  0x058c7, 0x05f3e, 0x065ad, 0x06696, 0x06a80, 0x06bb5, 0x07537, 0x08ac7,
  0x05024, 0x077e5, 0x05730, 0x05f1b, 0x06065, 0x0667a, 0x06c60, 0x075f4,
  0x07a1a, 0x07f6e, 0x081f4, 0x08718, 0x09045, 0x099b3, 0x07bc9, 0x0755c,
  0x07af9, 0x07b51, 0x084c4, 0x09010, 0x079e9, 0x07a92, 0x08336, 0x05ae1,
  0x07740, 0x04e2d, 0x04ef2, 0x05b99, 0x05fe0, 0x062bd, 0x0663c, 0x067f1,
  0x06ce8, 0x0866b, 0x08877, 0x08a3b, 0x0914e, 0x092f3, 0x099d0, 0x06a17,
  0x07026, 0x0732a, 0x082e7, 0x08457, 0x08caf, 0x04e01, 0x05146, 0x051cb,
  0x0558b, 0x05bf5, 0x05e16, 0x05e33, 0x05e81, 0x05f14, 0x05f35, 0x05f6b,
  0x05fb4, 0x061f2, 0x06311, 0x066a2, 0x0671d, 0x06f6e, 0x07252, 0x0753a,
  0x0773a, 0x08074, 0x08139, 0x08178, 0x08776, 0x08abf, 0x08adc, 0x08d85,
  0x08df3, 0x0929a, 0x09577, 0x09802, 0x09ce5, 0x052c5, 0x06357, 0x076f4,
  0x06715, 0x06c88, 0x073cd, 0x08cc3, 0x093ae, 0x09673, 0x06d25, 0x0589c,
  0x0690e, 0x069cc, 0x08ffd, 0x0939a, 0x075db, 0x0901a, 0x0585a, 0x06802,
  0x063b4, 0x069fb, 0x04f43, 0x06f2c, 0x067d8, 0x08fbb, 0x08526, 0x07db4,
  0x09354, 0x0693f, 0x06f70, 0x0576a, 0x058f7, 0x05b2c, 0x07d2c, 0x0722a,
  0x0540a, 0x091e3, 0x09db4, 0x04ead, 0x04f4e, 0x0505c, 0x05075, 0x05243,
  0x08c9e, 0x05448, 0x05824, 0x05b9a, 0x05e1d, 0x05e95, 0x05ead, 0x05ef7,
  0x05f1f, 0x0608c, 0x062b5, 0x0633a, 0x063d0, 0x068af, 0x06c40, 0x07887,
  0x0798e, 0x07a0b, 0x07de0, 0x08247, 0x08a02, 0x08ae6, 0x08e44, 0x09013,
  0x090b8, 0x0912d, 0x091d8, 0x09f0e, 0x06ce5, 0x06458, 0x064e2, 0x06575,
  0x06ef4, 0x07684, 0x07b1b, 0x09069, 0x093d1, 0x06eba, 0x054f2, 0x05fb9,
  0x064a4, 0x08f4d, 0x08fed, 0x09244, 0x05178, 0x0586b, 0x05929, 0x05c55,
  0x05e97, 0x06dfb, 0x07e8f, 0x0751c, 0x08cbc, 0x08ee2, 0x0985b, 0x070b9,
  0x04f1d, 0x06bbf, 0x06fb1, 0x07530, 0x096fb, 0x0514e, 0x05410, 0x05835,
  0x05857, 0x059ac, 0x05c60, 0x05f92, 0x06597, 0x0675c, 0x06e21, 0x0767b,
  0x083df, 0x08ced, 0x09014, 0x090fd, 0x0934d, 0x07825, 0x0783a, 0x052aa,
  0x05ea6, 0x0571f, 0x05974, 0x06012, 0x05012, 0x0515a, 0x051ac, 0x051cd,
  0x05200, 0x05510, 0x05854, 0x05858, 0x05957, 0x05b95, 0x05cf6, 0x05d8b,
  0x060bc, 0x06295, 0x0642d, 0x06771, 0x06843, 0x068bc, 0x068df, 0x076d7,
  0x06dd8, 0x06e6f, 0x06d9b, 0x0706f, 0x071c8, 0x05f53, 0x075d8, 0x07977,
  0x07b49, 0x07b54, 0x07b52, 0x07cd6, 0x07d71, 0x05230, 0x08463, 0x08569,
  0x085e4, 0x08a0e, 0x08b04, 0x08c46, 0x08e0f, 0x09003, 0x0900f, 0x09419,
  0x09676, 0x0982d, 0x09a30, 0x095d8, 0x050cd, 0x052d5, 0x0540c, 0x05802,
  0x05c0e, 0x061a7, 0x0649e, 0x06d1e, 0x077b3, 0x07ae5, 0x080f4, 0x08404,
  0x09053, 0x09285, 0x05ce0, 0x09d07, 0x0533f, 0x05f97, 0x05fb3, 0x06d9c,
  0x07279, 0x07763, 0x079bf, 0x07be4, 0x06bd2, 0x072ec, 0x08aad, 0x06803,
  0x06a61, 0x051f8, 0x07a81, 0x06934, 0x05c4a, 0x09cf6, 0x082eb, 0x05bc5,
  0x09149, 0x0701e, 0x05678, 0x05c6f, 0x060c7, 0x06566, 0x06c8c, 0x08c5a,
  0x09041, 0x09813, 0x05451, 0x066c7, 0x0920d, 0x05948, 0x090a3, 0x05185,
  0x04e4d, 0x051ea, 0x08599, 0x08b0e, 0x07058, 0x0637a, 0x0934b, 0x06962,
  0x099b4, 0x07e04, 0x07577, 0x05357, 0x06960, 0x08edf, 0x096e3, 0x06c5d,
  0x04e8c, 0x05c3c, 0x05f10, 0x08fe9, 0x05302, 0x08cd1, 0x08089, 0x08679,
  0x05eff, 0x065e5, 0x04e73, 0x05165, 0x05982, 0x05c3f, 0x097ee, 0x04efb,
  0x0598a, 0x05fcd, 0x08a8d, 0x06fe1, 0x079b0, 0x07962, 0x05be7, 0x08471,
  0x0732b, 0x071b1, 0x05e74, 0x05ff5, 0x0637b, 0x0649a, 0x071c3, 0x07c98,
  0x04e43, 0x05efc, 0x04e4b, 0x057dc, 0x056a2, 0x060a9, 0x06fc3, 0x07d0d,
  0x080fd, 0x08133, 0x081bf, 0x08fb2, 0x08997, 0x086a4, 0x05df4, 0x0628a,
  0x064ad, 0x08987, 0x06777, 0x06ce2, 0x06d3e, 0x07436, 0x07834, 0x05a46,
  0x07f75, 0x082ad, 0x099ac, 0x04ff3, 0x05ec3, 0x062dd, 0x06392, 0x06557,
  0x0676f, 0x076c3, 0x0724c, 0x080cc, 0x080ba, 0x08f29, 0x0914d, 0x0500d,
  0x057f9, 0x05a92, 0x06885, 0x06973, 0x07164, 0x072fd, 0x08cb7, 0x058f2,
  0x08ce0, 0x0966a, 0x09019, 0x0877f, 0x079e4, 0x077e7, 0x08429, 0x04f2f,
  0x05265, 0x0535a, 0x062cd, 0x067cf, 0x06cca, 0x0767d, 0x07b94, 0x07c95,
  0x08236, 0x08584, 0x08feb, 0x066dd, 0x06f20, 0x07206, 0x07e1b, 0x083ab,
  0x099c1, 0x09ea6, 0x051fd, 0x07bb1, 0x07872, 0x07bb8, 0x08087, 0x07b48,
  0x06ae8, 0x05e61, 0x0808c, 0x07551, 0x07560, 0x0516b, 0x09262, 0x06e8c,
  0x0767a, 0x09197, 0x09aea, 0x04f10, 0x07f70, 0x0629c, 0x07b4f, 0x095a5,
  0x09ce9, 0x0567a, 0x05859, 0x086e4, 0x096bc, 0x04f34, 0x05224, 0x0534a,
  0x053cd, 0x053db, 0x05e06, 0x0642c, 0x06591, 0x0677f, 0x06c3e, 0x06c4e,
  0x07248, 0x072af, 0x073ed, 0x07554, 0x07e41, 0x0822c, 0x085e9, 0x08ca9,
  0x07bc4, 0x091c6, 0x07169, 0x09812, 0x098ef, 0x0633d, 0x06669, 0x0756a,
  0x076e4, 0x078d0, 0x08543, 0x086ee, 0x0532a, 0x05351, 0x05426, 0x05983,
  0x05e87, 0x05f7c, 0x060b2, 0x06249, 0x06279, 0x062ab, 0x06590, 0x06bd4,
  0x06ccc, 0x075b2, 0x076ae, 0x07891, 0x079d8, 0x07dcb, 0x07f77, 0x080a5,
  0x088ab, 0x08ab9, 0x08cbb, 0x0907f, 0x0975e, 0x098db, 0x06a0b, 0x07c38,
  0x05099, 0x05c3e, 0x05fae, 0x06787, 0x06bd8, 0x07435, 0x07709, 0x07f8e,
  0x09f3b, 0x067ca, 0x07a17, 0x05339, 0x0758b, 0x09aed, 0x05f66, 0x0819d,
  0x083f1, 0x08098, 0x05f3c, 0x05fc5, 0x07562, 0x07b46, 0x0903c, 0x06867,
  0x059eb, 0x05a9b, 0x07d10, 0x0767e, 0x08b2c, 0x04ff5, 0x05f6a, 0x06a19,
  0x06c37, 0x06f02, 0x074e2, 0x07968, 0x08868, 0x08a55, 0x08c79, 0x05edf,
  0x063cf, 0x075c5, 0x079d2, 0x082d7, 0x09328, 0x092f2, 0x0849c, 0x086ed,
  0x09c2d, 0x054c1, 0x05f6c, 0x0658c, 0x06d5c, 0x07015, 0x08ca7, 0x08cd3,
  0x0983b, 0x0654f, 0x074f6, 0x04e0d, 0x04ed8, 0x057e0, 0x0592b, 0x05a66,
  0x05bcc, 0x051a8, 0x05e03, 0x05e9c, 0x06016, 0x06276, 0x06577, 0x065a7,
  0x0666e, 0x06d6e, 0x07236, 0x07b26, 0x08150, 0x0819a, 0x08299, 0x08b5c,
  0x08ca0, 0x08ce6, 0x08d74, 0x0961c, 0x09644, 0x04fae, 0x064ab, 0x06b66,
  0x0821e, 0x08461, 0x0856a, 0x090e8, 0x05c01, 0x06953, 0x098a8, 0x0847a,
  0x08557, 0x04f0f, 0x0526f, 0x05fa9, 0x05e45, 0x0670d, 0x0798f, 0x08179,
  0x08907, 0x08986, 0x06df5, 0x05f17, 0x06255, 0x06cb8, 0x04ecf, 0x07269,
  0x09b92, 0x05206, 0x0543b, 0x05674, 0x058b3, 0x061a4, 0x0626e, 0x0711a,
  0x0596e, 0x07c89, 0x07cde, 0x07d1b, 0x096f0, 0x06587, 0x0805e, 0x04e19,
  0x04f75, 0x05175, 0x05840, 0x05e63, 0x05e73, 0x05f0a, 0x067c4, 0x04e26,
  0x0853d, 0x09589, 0x0965b, 0x07c73, 0x09801, 0x050fb, 0x058c1, 0x07656,
  0x078a7, 0x05225, 0x077a5, 0x08511, 0x07b86, 0x0504f, 0x05909, 0x07247,
  0x07bc7, 0x07de8, 0x08fba, 0x08fd4, 0x0904d, 0x04fbf, 0x052c9, 0x05a29,
  0x05f01, 0x097ad, 0x04fdd, 0x08217, 0x092ea, 0x05703, 0x06355, 0x06b69,
  0x0752b, 0x088dc, 0x08f14, 0x07a42, 0x052df, 0x05893, 0x06155, 0x0620a,
  0x066ae, 0x06bcd, 0x07c3f, 0x083e9, 0x05023, 0x04ff8, 0x05305, 0x05446,
  0x05831, 0x05949, 0x05b9d, 0x05cf0, 0x05cef, 0x05d29, 0x05e96, 0x062b1,
  0x06367, 0x0653e, 0x065b9, 0x0670b, 0x06cd5, 0x06ce1, 0x070f9, 0x07832,
  0x07e2b, 0x080de, 0x082b3, 0x0840c, 0x084ec, 0x08702, 0x08912, 0x08a2a,
  0x08c4a, 0x090a6, 0x092d2, 0x098fd, 0x09cf3, 0x09d6c, 0x04e4f, 0x04ea1,
  0x0508d, 0x05256, 0x0574a, 0x059a8, 0x05e3d, 0x05fd8, 0x05fd9, 0x0623f,
  0x066b4, 0x0671b, 0x067d0, 0x068d2, 0x05192, 0x07d21, 0x080aa, 0x081a8,
  0x08b00, 0x08c8c, 0x08cbf, 0x0927e, 0x09632, 0x05420, 0x0982c, 0x05317,
  0x050d5, 0x0535c, 0x058a8, 0x064b2, 0x06734, 0x07267, 0x07766, 0x07a46,
  0x091e6, 0x052c3, 0x06ca1, 0x06b86, 0x05800, 0x05e4c, 0x05954, 0x0672c,
  0x07ffb, 0x051e1, 0x076c6, 0x06469, 0x078e8, 0x09b54, 0x09ebb, 0x057cb,
  0x059b9, 0x06627, 0x0679a, 0x06bce, 0x054e9, 0x069d9, 0x05e55, 0x0819c,
  0x06795, 0x09baa, 0x067fe, 0x09c52, 0x0685d, 0x04ea6, 0x04fe3, 0x053c8,
  0x062b9, 0x0672b, 0x06cab, 0x08fc4, 0x04fad, 0x07e6d, 0x09ebf, 0x04e07,
  0x06162, 0x06e80, 0x06f2b, 0x08513, 0x05473, 0x0672a, 0x09b45, 0x05df3,
  0x07b95, 0x05cac, 0x05bc6, 0x0871c, 0x06e4a, 0x084d1, 0x07a14, 0x08108,
  0x05999, 0x07c8d, 0x06c11, 0x07720, 0x052d9, 0x05922, 0x07121, 0x0725f,
  0x077db, 0x09727, 0x09d61, 0x0690b, 0x05a7f, 0x05a18, 0x051a5, 0x0540d,
  0x0547d, 0x0660e, 0x076df, 0x08ff7, 0x09298, 0x09cf4, 0x059ea, 0x0725d,
  0x06ec5, 0x0514d, 0x068c9, 0x07dbf, 0x07dec, 0x09762, 0x09eba, 0x06478,
  0x06a21, 0x08302, 0x05984, 0x05b5f, 0x06bdb, 0x0731b, 0x076f2, 0x07db2,
  0x08017, 0x08499, 0x05132, 0x06728, 0x09ed9, 0x076ee, 0x06762, 0x052ff,
  0x09905, 0x05c24, 0x0623b, 0x07c7e, 0x08cb0, 0x0554f, 0x060b6, 0x07d0b,
  0x09580, 0x05301, 0x04e5f, 0x051b6, 0x0591c, 0x0723a, 0x08036, 0x091ce,
  0x05f25, 0x077e2, 0x05384, 0x05f79, 0x07d04, 0x085ac, 0x08a33, 0x08e8d,
  0x09756, 0x067f3, 0x085ae, 0x09453, 0x06109, 0x06108, 0x06cb9, 0x07652,
  0x08aed, 0x08f38, 0x0552f, 0x04f51, 0x0512a, 0x052c7, 0x053cb, 0x05ba5,
  0x05e7d, 0x060a0, 0x06182, 0x063d6, 0x06709, 0x067da, 0x06e67, 0x06d8c,
  0x07336, 0x07337, 0x07531, 0x07950, 0x088d5, 0x08a98, 0x0904a, 0x09091,
  0x090f5, 0x096c4, 0x0878d, 0x05915, 0x04e88, 0x04f59, 0x04e0e, 0x08a89,
  0x08f3f, 0x09810, 0x050ad, 0x05e7c, 0x05996, 0x05bb9, 0x05eb8, 0x063da,
  0x063fa, 0x064c1, 0x066dc, 0x0694a, 0x069d8, 0x06d0b, 0x06eb6, 0x07194,
  0x07528, 0x07aaf, 0x07f8a, 0x08000, 0x08449, 0x084c9, 0x08981, 0x08b21,
  0x08e0a, 0x09065, 0x0967d, 0x0990a, 0x0617e, 0x06291, 0x06b32, 0x06c83,
  0x06d74, 0x07fcc, 0x07ffc, 0x06dc0, 0x07f85, 0x087ba, 0x088f8, 0x06765,
  0x083b1, 0x0983c, 0x096f7, 0x06d1b, 0x07d61, 0x0843d, 0x0916a, 0x04e71,
  0x05375, 0x05d50, 0x06b04, 0x06feb, 0x085cd, 0x0862d, 0x089a7, 0x05229,
  0x0540f, 0x05c65, 0x0674e, 0x068a8, 0x07406, 0x07483, 0x075e2, 0x088cf,
  0x088e1, 0x091cc, 0x096e2, 0x09678, 0x05f8b, 0x07387, 0x07acb, 0x0844e,
  0x063a0, 0x07565, 0x05289, 0x06d41, 0x06e9c, 0x07409, 0x07559, 0x0786b,
  0x07c92, 0x09686, 0x07adc, 0x09f8d, 0x04fb6, 0x0616e, 0x065c5, 0x0865c,
  0x04e86, 0x04eae, 0x050da, 0x04e21, 0x051cc, 0x05bee, 0x06599, 0x06881,
  0x06dbc, 0x0731f, 0x07642, 0x077ad, 0x07a1c, 0x07ce7, 0x0826f, 0x08ad2,
  0x0907c, 0x091cf, 0x09675, 0x09818, 0x0529b, 0x07dd1, 0x0502b, 0x05398,
  0x06797, 0x06dcb, 0x071d0, 0x07433, 0x081e8, 0x08f2a, 0x096a3, 0x09c57,
  0x09e9f, 0x07460, 0x05841, 0x06d99, 0x07d2f, 0x0985e, 0x04ee4, 0x04f36,
  0x04f8b, 0x051b7, 0x052b1, 0x05dba, 0x0601c, 0x073b2, 0x0793c, 0x082d3,
  0x09234, 0x096b7, 0x096f6, 0x0970a, 0x09e97, 0x09f62, 0x066a6, 0x06b74,
  0x05217, 0x052a3, 0x070c8, 0x088c2, 0x05ec9, 0x0604b, 0x06190, 0x06f23,
  0x07149, 0x07c3e, 0x07df4, 0x0806f, 0x084ee, 0x09023, 0x0932c, 0x05442,
  0x09b6f, 0x06ad3, 0x07089, 0x08cc2, 0x08def, 0x09732, 0x052b4, 0x05a41,
  0x05eca, 0x05f04, 0x06717, 0x0697c, 0x06994, 0x06d6a, 0x06f0f, 0x07262,
  0x072fc, 0x07bed, 0x08001, 0x0807e, 0x0874b, 0x090ce, 0x0516d, 0x09e93,
  0x07984, 0x0808b, 0x09332, 0x08ad6, 0x0502d, 0x0548c, 0x08a71, 0x06b6a,
  0x08cc4, 0x08107, 0x060d1, 0x067a0, 0x09df2, 0x04e99, 0x04e98, 0x09c10,
  0x08a6b, 0x085c1, 0x08568, 0x06900, 0x06e7e, 0x07897, 0x08155, 0x20b9f,
  0x05b41, 0x05b56, 0x05b7d, 0x05b93, 0x05bd8, 0x05bec, 0x05c12, 0x05c1e,
  0x05c23, 0x05c2b, 0x0378d, 0x05c62, 0x0fa3b, 0x0fa3c, 0x216b4, 0x05c7a,
  0x05c8f, 0x05c9f, 0x05ca3, 0x05caa, 0x05cba, 0x05ccb, 0x05cd0, 0x05cd2,
  0x05cf4, 0x21e34, 0x037e2, 0x05d0d, 0x05d27, 0x0fa11, 0x05d46, 0x05d47,
  0x05d53, 0x05d4a, 0x05d6d, 0x05d81, 0x05da0, 0x05da4, 0x05da7, 0x05db8,
  0x05dcb, 0x0541e, 0x05f0c, 0x04e10, 0x04e15, 0x04e2a, 0x04e31, 0x04e36,
  0x04e3c, 0x04e3f, 0x04e42, 0x04e56, 0x04e58, 0x04e82, 0x04e85, 0x08c6b,
  0x04e8a, 0x08212, 0x05f0d, 0x04e8e, 0x04e9e, 0x04e9f, 0x04ea0, 0x04ea2,
  0x04eb0, 0x04eb3, 0x04eb6, 0x04ece, 0x04ecd, 0x04ec4, 0x04ec6, 0x04ec2,
  0x04ed7, 0x04ede, 0x04eed, 0x04edf, 0x04ef7, 0x04f09, 0x04f5a, 0x04f30,
  0x04f5b, 0x04f5d, 0x04f57, 0x04f47, 0x04f76, 0x04f88, 0x04f8f, 0x04f98,
  0x04f7b, 0x04f69, 0x04f70, 0x04f91, 0x04f6f, 0x04f86, 0x04f96, 0x05118,
  0x04fd4, 0x04fdf, 0x04fce, 0x04fd8, 0x04fdb, 0x04fd1, 0x04fda, 0x04fd0,
  0x04fe4, 0x04fe5, 0x0501a, 0x05028, 0x05014, 0x0502a, 0x05025, 0x05005,
  0x04f1c, 0x04ff6, 0x05021, 0x05029, 0x0502c, 0x04ffe, 0x04fef, 0x05011,
  0x05006, 0x05043, 0x05047, 0x06703, 0x05055, 0x05050, 0x05048, 0x0505a,
  0x05056, 0x0506c, 0x05078, 0x05080, 0x0509a, 0x05085, 0x050b4, 0x050b2,
  0x050c9, 0x050ca, 0x050b3, 0x050c2, 0x050d6, 0x050de, 0x050e5, 0x050ed,
  0x050e3, 0x050ee, 0x050f9, 0x050f5, 0x05109, 0x05101, 0x05102, 0x05116,
  0x05115, 0x05114, 0x0511a, 0x05121, 0x0513a, 0x05137, 0x0513c, 0x0513b,
  0x0513f, 0x05140, 0x05152, 0x0514c, 0x05154, 0x05162, 0x07af8, 0x05169,
  0x0516a, 0x0516e, 0x05180, 0x05182, 0x056d8, 0x0518c, 0x05189, 0x0518f,
  0x05191, 0x05193, 0x05195, 0x05196, 0x051a4, 0x051a6, 0x051a2, 0x051a9,
  0x051aa, 0x051ab, 0x051b3, 0x051b1, 0x051b2, 0x051b0, 0x051b5, 0x051bd,
  0x051c5, 0x051c9, 0x051db, 0x051e0, 0x08655, 0x051e9, 0x051ed, 0x051f0,
  0x051f5, 0x051fe, 0x05204, 0x0520b, 0x05214, 0x0520e, 0x05227, 0x0522a,
  0x0522e, 0x05233, 0x05239, 0x0524f, 0x05244, 0x0524b, 0x0524c, 0x0525e,
  0x05254, 0x0526a, 0x05274, 0x05269, 0x05273, 0x0527f, 0x0527d, 0x0528d,
  0x05294, 0x05292, 0x05271, 0x05288, 0x05291, 0x08fa8, 0x08fa7, 0x052ac,
  0x052ad, 0x052bc, 0x052b5, 0x052c1, 0x052cd, 0x052d7, 0x052de, 0x052e3,
  0x052e6, 0x098ed, 0x052e0, 0x052f3, 0x052f5, 0x052f8, 0x052f9, 0x05306,
  0x05308, 0x07538, 0x0530d, 0x05310, 0x0530f, 0x05315, 0x0531a, 0x05323,
  0x0532f, 0x05331, 0x05333, 0x05338, 0x05340, 0x05346, 0x05345, 0x04e17,
  0x05349, 0x0534d, 0x051d6, 0x0535e, 0x05369, 0x0536e, 0x05918, 0x0537b,
  0x05377, 0x05382, 0x05396, 0x053a0, 0x053a6, 0x053a5, 0x053ae, 0x053b0,
  0x053b6, 0x053c3, 0x07c12, 0x096d9, 0x053df, 0x066fc, 0x071ee, 0x053ee,
  0x053e8, 0x053ed, 0x053fa, 0x05401, 0x0543d, 0x05440, 0x0542c, 0x0542d,
  0x0543c, 0x0542e, 0x05436, 0x05429, 0x0541d, 0x0544e, 0x0548f, 0x05475,
  0x0548e, 0x0545f, 0x05471, 0x05477, 0x05470, 0x05492, 0x0547b, 0x05480,
  0x05476, 0x05484, 0x05490, 0x05486, 0x054c7, 0x054a2, 0x054b8, 0x054a5,
  0x054ac, 0x054c4, 0x054c8, 0x054a8, 0x054ab, 0x054c2, 0x054a4, 0x054be,
  0x054bc, 0x054d8, 0x054e5, 0x054e6, 0x0550f, 0x05514, 0x054fd, 0x054ee,
  0x054ed, 0x054fa, 0x054e2, 0x05539, 0x05540, 0x05563, 0x0554c, 0x0552e,
  0x0555c, 0x05545, 0x05556, 0x05557, 0x05538, 0x05533, 0x0555d, 0x05599,
  0x05580, 0x054af, 0x0558a, 0x0559f, 0x0557b, 0x0557e, 0x05598, 0x0559e,
  0x055ae, 0x0557c, 0x05583, 0x055a9, 0x05587, 0x055a8, 0x055da, 0x055c5,
  0x055df, 0x055c4, 0x055dc, 0x055e4, 0x055d4, 0x05614, 0x055f7, 0x05616,
  0x055fe, 0x055fd, 0x0561b, 0x055f9, 0x0564e, 0x05650, 0x071df, 0x05634,
  0x05636, 0x05632, 0x05638, 0x0566b, 0x05664, 0x0562f, 0x0566c, 0x0566a,
  0x05686, 0x05680, 0x0568a, 0x056a0, 0x05694, 0x0568f, 0x056a5, 0x056ae,
  0x056b6, 0x056b4, 0x056c2, 0x056bc, 0x056c1, 0x056c3, 0x056c0, 0x056c8,
  0x056ce, 0x056d1, 0x056d3, 0x056d7, 0x056ee, 0x056f9, 0x05700, 0x056ff,
  0x05704, 0x05709, 0x05708, 0x0570b, 0x0570d, 0x05713, 0x05718, 0x05716,
  0x055c7, 0x0571c, 0x05726, 0x05737, 0x05738, 0x0574e, 0x0573b, 0x05740,
  0x0574f, 0x05769, 0x057c0, 0x05788, 0x05761, 0x0577f, 0x05789, 0x05793,
  0x057a0, 0x057b3, 0x057a4, 0x057aa, 0x057b0, 0x057c3, 0x057c6, 0x057d4,
  0x057d2, 0x057d3, 0x0580a, 0x057d6, 0x057e3, 0x0580b, 0x05819, 0x0581d,
  0x05872, 0x05821, 0x05862, 0x0584b, 0x05870, 0x06bc0, 0x05852, 0x0583d,
  0x05879, 0x05885, 0x058b9, 0x0589f, 0x058ab, 0x058ba, 0x058de, 0x058bb,
  0x058b8, 0x058ae, 0x058c5, 0x058d3, 0x058d1, 0x058d7, 0x058d9, 0x058d8,
  0x058e5, 0x058dc, 0x058e4, 0x058df, 0x058ef, 0x058fa, 0x058f9, 0x058fb,
  0x058fc, 0x058fd, 0x05902, 0x0590a, 0x05910, 0x0591b, 0x068a6, 0x05925,
  0x0592c, 0x0592d, 0x05932, 0x05938, 0x0593e, 0x07ad2, 0x05955, 0x05950,
  0x0594e, 0x0595a, 0x05958, 0x05962, 0x05960, 0x05967, 0x0596c, 0x05969,
  0x05978, 0x05981, 0x0599d, 0x04f5e, 0x04fab, 0x059a3, 0x059b2, 0x059c6,
  0x059e8, 0x059dc, 0x0598d, 0x059d9, 0x059da, 0x05a25, 0x05a1f, 0x05a11,
  0x05a1c, 0x05a09, 0x05a1a, 0x05a40, 0x05a6c, 0x05a49, 0x05a35, 0x05a36,
  0x05a62, 0x05a6a, 0x05a9a, 0x05abc, 0x05abe, 0x05acb, 0x05ac2, 0x05abd,
  0x05ae3, 0x05ad7, 0x05ae6, 0x05ae9, 0x05ad6, 0x05afa, 0x05afb, 0x05b0c,
  0x05b0b, 0x05b16, 0x05b32, 0x05ad0, 0x05b2a, 0x05b36, 0x05b3e, 0x05b43,
  0x05b45, 0x05b40, 0x05b51, 0x05b55, 0x05b5a, 0x05b5b, 0x05b65, 0x05b69,
  0x05b70, 0x05b73, 0x05b75, 0x05b78, 0x06588, 0x05b7a, 0x05b80, 0x05b83,
  0x05ba6, 0x05bb8, 0x05bc3, 0x05bc7, 0x05bc9, 0x05bd4, 0x05bd0, 0x05be4,
  0x05be6, 0x05be2, 0x05bde, 0x05be5, 0x05beb, 0x05bf0, 0x05bf6, 0x05bf3,
  0x05c05, 0x05c07, 0x05c08, 0x05c0d, 0x05c13, 0x05c20, 0x05c22, 0x05c28,
  0x05c38, 0x05c39, 0x05c41, 0x05c46, 0x05c4e, 0x05c53, 0x05c50, 0x05c4f,
  0x05b71, 0x05c6c, 0x05c6e, 0x04e62, 0x05c76, 0x05c79, 0x05c8c, 0x05c91,
  0x05c94, 0x0599b, 0x05cab, 0x05cbb, 0x05cb6, 0x05cbc, 0x05cb7, 0x05cc5,
  0x05cbe, 0x05cc7, 0x05cd9, 0x05ce9, 0x05cfd, 0x05cfa, 0x05ced, 0x05d8c,
  0x05cea, 0x05d0b, 0x05d15, 0x05d17, 0x05d5c, 0x05d1f, 0x05d1b, 0x05d11,
  0x05d14, 0x05d22, 0x05d1a, 0x05d19, 0x05d18, 0x05d4c, 0x05d52, 0x05d4e,
  0x05d4b, 0x05d6c, 0x05d73, 0x05d76, 0x05d87, 0x05d84, 0x05d82, 0x05da2,
  0x05d9d, 0x05dac, 0x05dae, 0x05dbd, 0x05d90, 0x05db7, 0x05dbc, 0x05dc9,
  0x05dcd, 0x05dd3, 0x05dd2, 0x05dd6, 0x05ddb, 0x05deb, 0x05df2, 0x05df5,
  0x05e0b, 0x05e1a, 0x05e19, 0x05e11, 0x05e1b, 0x05e36, 0x05e37, 0x05e44,
  0x05e43, 0x05e40, 0x05e4e, 0x05e57, 0x05e54, 0x05e5f, 0x05e62, 0x05e64,
  0x05e47, 0x05e75, 0x05e76, 0x05e7a, 0x09ebc, 0x05e7f, 0x05ea0, 0x05ec1,
  0x05ec2, 0x05ec8, 0x05ed0, 0x05ecf, 0x05ed6, 0x05ee3, 0x05edd, 0x05eda,
  0x05edb, 0x05ee2, 0x05ee1, 0x05ee8, 0x05ee9, 0x05eec, 0x05ef1, 0x05ef3,
  0x05ef0, 0x05ef4, 0x05ef8, 0x05efe, 0x05f03, 0x05f09, 0x05f5d, 0x05f5c,
  0x05f0b, 0x05f11, 0x05f16, 0x05f29, 0x05f2d, 0x05f38, 0x05f41, 0x05f48,
  0x05f4c, 0x05f4e, 0x05f2f, 0x05f51, 0x05f56, 0x05f57, 0x05f59, 0x05f61,
  0x05f6d, 0x05f73, 0x05f77, 0x05f83, 0x05f82, 0x05f7f, 0x05f8a, 0x05f88,
  0x05f91, 0x05f87, 0x05f9e, 0x05f99, 0x05f98, 0x05fa0, 0x05fa8, 0x05fad,
  0x05fbc, 0x05fd6, 0x05ffb, 0x05fe4, 0x05ff8, 0x05ff1, 0x05fdd, 0x060b3,
  0x05fff, 0x06021, 0x06060, 0x06019, 0x06010, 0x06029, 0x0600e, 0x06031,
  0x0601b, 0x06015, 0x0602b, 0x06026, 0x0600f, 0x0603a, 0x0605a, 0x06041,
  0x0606a, 0x06077, 0x0605f, 0x0604a, 0x06046, 0x0604d, 0x06063, 0x06043,
  0x06064, 0x06042, 0x0606c, 0x0606b, 0x06059, 0x06081, 0x0608d, 0x060e7,
  0x06083, 0x0609a, 0x06084, 0x0609b, 0x06096, 0x06097, 0x06092, 0x060a7,
  0x0608b, 0x060e1, 0x060b8, 0x060e0, 0x060d3, 0x060b4, 0x05ff0, 0x060bd,
  0x060c6, 0x060b5, 0x060d8, 0x0614d, 0x06115, 0x06106, 0x060f6, 0x060f7,
  0x06100, 0x060f4, 0x060fa, 0x06103, 0x06121, 0x060fb, 0x060f1, 0x0610d,
  0x0610e, 0x06147, 0x0613e, 0x06128, 0x06127, 0x0614a, 0x0613f, 0x0613c,
  0x0612c, 0x06134, 0x0613d, 0x06142, 0x06144, 0x06173, 0x06177, 0x06158,
  0x06159, 0x0615a, 0x0616b, 0x06174, 0x0616f, 0x06165, 0x06171, 0x0615f,
  0x0615d, 0x06153, 0x06175, 0x06199, 0x06196, 0x06187, 0x061ac, 0x06194,
  0x0619a, 0x0618a, 0x06191, 0x061ab, 0x061ae, 0x061cc, 0x061ca, 0x061c9,
  0x061f7, 0x061c8, 0x061c3, 0x061c6, 0x061ba, 0x061cb, 0x07f79, 0x061cd,
  0x061e6, 0x061e3, 0x061f6, 0x061fa, 0x061f4, 0x061ff, 0x061fd, 0x061fc,
  0x061fe, 0x06200, 0x06208, 0x06209, 0x0620d, 0x0620c, 0x06214, 0x0621b,
  0x0621e, 0x06221, 0x0622a, 0x0622e, 0x06230, 0x06232, 0x06233, 0x06241,
  0x0624e, 0x0625e, 0x06263, 0x0625b, 0x06260, 0x06268, 0x0627c, 0x06282,
  0x06289, 0x0627e, 0x06292, 0x06293, 0x06296, 0x062d4, 0x06283, 0x06294,
  0x062d7, 0x062d1, 0x062bb, 0x062cf, 0x062ff, 0x062c6, 0x064d4, 0x062c8,
  0x062dc, 0x062cc, 0x062ca, 0x062c2, 0x062c7, 0x0629b, 0x062c9, 0x0630c,
  0x062ee, 0x062f1, 0x06327, 0x06302, 0x06308, 0x062ef, 0x062f5, 0x06350,
  0x0633e, 0x0634d, 0x0641c, 0x0634f, 0x06396, 0x0638e, 0x06380, 0x063ab,
  0x06376, 0x063a3, 0x0638f, 0x06389, 0x0639f, 0x063b5, 0x0636b, 0x06369,
  0x063be, 0x063e9, 0x063c0, 0x063c6, 0x063e3, 0x063c9, 0x063d2, 0x063f6,
  0x063c4, 0x06416, 0x06434, 0x06406, 0x06413, 0x06426, 0x06436, 0x0651d,
  0x06417, 0x06428, 0x0640f, 0x06467, 0x0646f, 0x06476, 0x0644e, 0x0652a,
  0x06495, 0x06493, 0x064a5, 0x064a9, 0x06488, 0x064bc, 0x064da, 0x064d2,
  0x064c5, 0x064c7, 0x064bb, 0x064d8, 0x064c2, 0x064f1, 0x064e7, 0x08209,
  0x064e0, 0x064e1, 0x062ac, 0x064e3, 0x064ef, 0x0652c, 0x064f6, 0x064f4,
  0x064f2, 0x064fa, 0x06500, 0x064fd, 0x06518, 0x0651c, 0x06505, 0x06524,
  0x06523, 0x0652b, 0x06534, 0x06535, 0x06537, 0x06536, 0x06538, 0x0754b,
  0x06548, 0x06556, 0x06555, 0x0654d, 0x06558, 0x0655e, 0x0655d, 0x06572,
  0x06578, 0x06582, 0x06583, 0x08b8a, 0x0659b, 0x0659f, 0x065ab, 0x065b7,
  0x065c3, 0x065c6, 0x065c1, 0x065c4, 0x065cc, 0x065d2, 0x065db, 0x065d9,
  0x065e0, 0x065e1, 0x065f1, 0x06772, 0x0660a, 0x06603, 0x065fb, 0x06773,
  0x06635, 0x06636, 0x06634, 0x0661c, 0x0664f, 0x06644, 0x06649, 0x06641,
  0x0665e, 0x0665d, 0x06664, 0x06667, 0x06668, 0x0665f, 0x06662, 0x06670,
  0x06683, 0x06688, 0x0668e, 0x06689, 0x06684, 0x06698, 0x0669d, 0x066c1,
  0x066b9, 0x066c9, 0x066be, 0x066bc, 0x066c4, 0x066b8, 0x066d6, 0x066da,
  0x066e0, 0x0663f, 0x066e6, 0x066e9, 0x066f0, 0x066f5, 0x066f7, 0x0670f,
  0x06716, 0x0671e, 0x06726, 0x06727, 0x09738, 0x0672e, 0x0673f, 0x06736,
  0x06741, 0x06738, 0x06737, 0x06746, 0x0675e, 0x06760, 0x06759, 0x06763,
  0x06764, 0x06789, 0x06770, 0x067a9, 0x0677c, 0x0676a, 0x0678c, 0x0678b,
  0x067a6, 0x067a1, 0x06785, 0x067b7, 0x067ef, 0x067b4, 0x067ec, 0x067b3,
  0x067e9, 0x067b8, 0x067e4, 0x067de, 0x067dd, 0x067e2, 0x067ee, 0x067b9,
  0x067ce, 0x067c6, 0x067e7, 0x06a9c, 0x0681e, 0x06846, 0x06829, 0x06840,
  0x0684d, 0x06832, 0x0684e, 0x068b3, 0x0682b, 0x06859, 0x06863, 0x06877,
  0x0687f, 0x0689f, 0x0688f, 0x068ad, 0x06894, 0x0689d, 0x0689b, 0x06883,
  0x06aae, 0x068b9, 0x06874, 0x068b5, 0x068a0, 0x068ba, 0x0690f, 0x0688d,
  0x0687e, 0x06901, 0x068ca, 0x06908, 0x068d8, 0x06922, 0x06926, 0x068e1,
  0x0690c, 0x068cd, 0x068d4, 0x068e7, 0x068d5, 0x06936, 0x06912, 0x06904,
  0x068d7, 0x068e3, 0x06925, 0x068f9, 0x068e0, 0x068ef, 0x06928, 0x0692a,
  0x0691a, 0x06923, 0x06921, 0x068c6, 0x06979, 0x06977, 0x0695c, 0x06978,
  0x0696b, 0x06954, 0x0697e, 0x0696e, 0x06939, 0x06974, 0x0693d, 0x06959,
  0x06930, 0x06961, 0x0695e, 0x0695d, 0x06981, 0x0696a, 0x069b2, 0x069ae,
  0x069d0, 0x069bf, 0x069c1, 0x069d3, 0x069be, 0x069ce, 0x05be8, 0x069ca,
  0x069dd, 0x069bb, 0x069c3, 0x069a7, 0x06a2e, 0x06991, 0x069a0, 0x0699c,
  0x06995, 0x069b4, 0x069de, 0x069e8, 0x06a02, 0x06a1b, 0x069ff, 0x06b0a,
  0x069f9, 0x069f2, 0x069e7, 0x06a05, 0x069b1, 0x06a1e, 0x069ed, 0x06a14,
  0x069eb, 0x06a0a, 0x06a12, 0x06ac1, 0x06a23, 0x06a13, 0x06a44, 0x06a0c,
  0x06a72, 0x06a36, 0x06a78, 0x06a47, 0x06a62, 0x06a59, 0x06a66, 0x06a48,
  0x06a38, 0x06a22, 0x06a90, 0x06a8d, 0x06aa0, 0x06a84, 0x06aa2, 0x06aa3,
  0x06a97, 0x08617, 0x06abb, 0x06ac3, 0x06ac2, 0x06ab8, 0x06ab3, 0x06aac,
  0x06ade, 0x06ad1, 0x06adf, 0x06aaa, 0x06ada, 0x06aea, 0x06afb, 0x06b05,
  0x08616, 0x06afa, 0x06b12, 0x06b16, 0x09b31, 0x06b1f, 0x06b38, 0x06b37,
  0x076dc, 0x06b39, 0x098ee, 0x06b47, 0x06b43, 0x06b49, 0x06b50, 0x06b59,
  0x06b54, 0x06b5b, 0x06b5f, 0x06b61, 0x06b78, 0x06b79, 0x06b7f, 0x06b80,
  0x06b84, 0x06b83, 0x06b8d, 0x06b98, 0x06b95, 0x06b9e, 0x06ba4, 0x06baa,
  0x06bab, 0x06baf, 0x06bb2, 0x06bb1, 0x06bb3, 0x06bb7, 0x06bbc, 0x06bc6,
  0x06bcb, 0x06bd3, 0x06bdf, 0x06bec, 0x06beb, 0x06bf3, 0x06bef, 0x09ebe,
  0x06c08, 0x06c13, 0x06c14, 0x06c1b, 0x06c24, 0x06c23, 0x06c5e, 0x06c55,
  0x06c62, 0x06c6a, 0x06c82, 0x06c8d, 0x06c9a, 0x06c81, 0x06c9b, 0x06c7e,
  0x06c68, 0x06c73, 0x06c92, 0x06c90, 0x06cc4, 0x06cf1, 0x06cd3, 0x06cbd,
  0x06cd7, 0x06cc5, 0x06cdd, 0x06cae, 0x06cb1, 0x06cbe, 0x06cba, 0x06cdb,
  0x06cef, 0x06cd9, 0x06cea, 0x06d1f, 0x0884d, 0x06d36, 0x06d2b, 0x06d3d,
  0x06d38, 0x06d19, 0x06d35, 0x06d33, 0x06d12, 0x06d0c, 0x06d63, 0x06d93,
  0x06d64, 0x06d5a, 0x06d79, 0x06d59, 0x06d8e, 0x06d95, 0x06fe4, 0x06d85,
  0x06df9, 0x06e15, 0x06e0a, 0x06db5, 0x06dc7, 0x06de6, 0x06db8, 0x06dc6,
  0x06dec, 0x06dde, 0x06dcc, 0x06de8, 0x06dd2, 0x06dc5, 0x06dfa, 0x06dd9,
  0x06de4, 0x06dd5, 0x06dea, 0x06dee, 0x06e2d, 0x06e6e, 0x06e2e, 0x06e19,
  0x06e72, 0x06e5f, 0x06e3e, 0x06e23, 0x06e6b, 0x06e2b, 0x06e76, 0x06e4d,
  0x06e1f, 0x06e43, 0x06e3a, 0x06e4e, 0x06e24, 0x06eff, 0x06e1d, 0x06e38,
  0x06e82, 0x06eaa, 0x06e98, 0x06ec9, 0x06eb7, 0x06ed3, 0x06ebd, 0x06eaf,
  0x06ec4, 0x06eb2, 0x06ed4, 0x06ed5, 0x06e8f, 0x06ea5, 0x06ec2, 0x06e9f,
  0x06f41, 0x06f11, 0x0704c, 0x06eec, 0x06ef8, 0x06efe, 0x06f3f, 0x06ef2,
  0x06f31, 0x06eef, 0x06f32, 0x06ecc, 0x06f3e, 0x06f13, 0x06ef7, 0x06f86,
  0x06f7a, 0x06f78, 0x06f81, 0x06f80, 0x06f6f, 0x06f5b, 0x06ff3, 0x06f6d,
  0x06f82, 0x06f7c, 0x06f58, 0x06f8e, 0x06f91, 0x06fc2, 0x06f66, 0x06fb3,
  0x06fa3, 0x06fa1, 0x06fa4, 0x06fb9, 0x06fc6, 0x06faa, 0x06fdf, 0x06fd5,
  0x06fec, 0x06fd4, 0x06fd8, 0x06ff1, 0x06fee, 0x06fdb, 0x07009, 0x0700b,
  0x06ffa, 0x07011, 0x07001, 0x0700f, 0x06ffe, 0x0701b, 0x0701a, 0x06f74,
  0x0701d, 0x07018, 0x0701f, 0x07030, 0x0703e, 0x07032, 0x07051, 0x07063,
  0x07099, 0x07092, 0x070af, 0x070f1, 0x070ac, 0x070b8, 0x070b3, 0x070ae,
  0x070df, 0x070cb, 0x070dd, 0x070d9, 0x07109, 0x070fd, 0x0711c, 0x07119,
  0x07165, 0x07155, 0x07188, 0x07166, 0x07162, 0x0714c, 0x07156, 0x0716c,
  0x0718f, 0x071fb, 0x07184, 0x07195, 0x071a8, 0x071ac, 0x071d7, 0x071b9,
  0x071be, 0x071d2, 0x071c9, 0x071d4, 0x071ce, 0x071e0, 0x071ec, 0x071e7,
  0x071f5, 0x071fc, 0x071f9, 0x071ff, 0x0720d, 0x07210, 0x0721b, 0x07228,
  0x0722d, 0x0722c, 0x07230, 0x07232, 0x0723b, 0x0723c, 0x0723f, 0x07240,
  0x07246, 0x0724b, 0x07258, 0x07274, 0x0727e, 0x07282, 0x07281, 0x07287,
  0x07292, 0x07296, 0x072a2, 0x072a7, 0x072b9, 0x072b2, 0x072c3, 0x072c6,
  0x072c4, 0x072ce, 0x072d2, 0x072e2, 0x072e0, 0x072e1, 0x072f9, 0x072f7,
  0x0500f, 0x07317, 0x0730a, 0x0731c, 0x07316, 0x0731d, 0x07334, 0x0732f,
  0x07329, 0x07325, 0x0733e, 0x0734e, 0x0734f, 0x09ed8, 0x07357, 0x0736a,
  0x07368, 0x07370, 0x07378, 0x07375, 0x0737b, 0x0737a, 0x073c8, 0x073b3,
  0x073ce, 0x073bb, 0x073c0, 0x073e5, 0x073ee, 0x073de, 0x074a2, 0x07405,
  0x0746f, 0x07425, 0x073f8, 0x07432, 0x0743a, 0x07455, 0x0743f, 0x0745f,
  0x07459, 0x07441, 0x0745c, 0x07469, 0x07470, 0x07463, 0x0746a, 0x07476,
  0x0747e, 0x0748b, 0x0749e, 0x074a7, 0x074ca, 0x074cf, 0x074d4, 0x073f1,
  0x074e0, 0x074e3, 0x074e7, 0x074e9, 0x074ee, 0x074f2, 0x074f0, 0x074f1,
  0x074f8, 0x074f7, 0x07504, 0x07503, 0x07505, 0x0750c, 0x0750e, 0x0750d,
  0x07515, 0x07513, 0x0751e, 0x07526, 0x0752c, 0x0753c, 0x07544, 0x0754d,
  0x0754a, 0x07549, 0x0755b, 0x07546, 0x0755a, 0x07569, 0x07564, 0x07567,
  0x0756b, 0x0756d, 0x07578, 0x07576, 0x07586, 0x07587, 0x07574, 0x0758a,
  0x07589, 0x07582, 0x07594, 0x0759a, 0x0759d, 0x075a5, 0x075a3, 0x075c2,
  0x075b3, 0x075c3, 0x075b5, 0x075bd, 0x075b8, 0x075bc, 0x075b1, 0x075cd,
  0x075ca, 0x075d2, 0x075d9, 0x075e3, 0x075de, 0x075fe, 0x075ff, 0x075fc,
  0x07601, 0x075f0, 0x075fa, 0x075f2, 0x075f3, 0x0760b, 0x0760d, 0x07609,
  0x0761f, 0x07627, 0x07620, 0x07621, 0x07622, 0x07624, 0x07634, 0x07630,
  0x0763b, 0x07647, 0x07648, 0x07646, 0x0765c, 0x07658, 0x07661, 0x07662,
  0x07668, 0x07669, 0x0766a, 0x07667, 0x0766c, 0x07670, 0x07672, 0x07676,
  0x07678, 0x0767c, 0x07680, 0x07683, 0x07688, 0x0768b, 0x0768e, 0x07696,
  0x07693, 0x07699, 0x0769a, 0x076b0, 0x076b4, 0x076b8, 0x076b9, 0x076ba,
  0x076c2, 0x076cd, 0x076d6, 0x076d2, 0x076de, 0x076e1, 0x076e5, 0x076e7,
  0x076ea, 0x0862f, 0x076fb, 0x07708, 0x07707, 0x07704, 0x07729, 0x07724,
  0x0771e, 0x07725, 0x07726, 0x0771b, 0x07737, 0x07738, 0x07747, 0x0775a,
  0x07768, 0x0776b, 0x0775b, 0x07765, 0x0777f, 0x0777e, 0x07779, 0x0778e,
  0x0778b, 0x07791, 0x077a0, 0x0779e, 0x077b0, 0x077b6, 0x077b9, 0x077bf,
  0x077bc, 0x077bd, 0x077bb, 0x077c7, 0x077cd, 0x077d7, 0x077da, 0x077dc,
  0x077e3, 0x077ee, 0x077fc, 0x0780c, 0x07812, 0x07926, 0x07820, 0x0792a,
  0x07845, 0x0788e, 0x07874, 0x07886, 0x0787c, 0x0789a, 0x0788c, 0x078a3,
  0x078b5, 0x078aa, 0x078af, 0x078d1, 0x078c6, 0x078cb, 0x078d4, 0x078be,
  0x078bc, 0x078c5, 0x078ca, 0x078ec, 0x078e7, 0x078da, 0x078fd, 0x078f4,
  0x07907, 0x07912, 0x07911, 0x07919, 0x0792c, 0x0792b, 0x07940, 0x07960,
  0x07957, 0x0795f, 0x0795a, 0x07955, 0x07953, 0x0797a, 0x0797f, 0x0798a,
  0x0799d, 0x079a7, 0x09f4b, 0x079aa, 0x079ae, 0x079b3, 0x079b9, 0x079ba,
  0x079c9, 0x079d5, 0x079e7, 0x079ec, 0x079e1, 0x079e3, 0x07a08, 0x07a0d,
  0x07a18, 0x07a19, 0x07a20, 0x07a1f, 0x07980, 0x07a31, 0x07a3b, 0x07a3e,
  0x07a37, 0x07a43, 0x07a57, 0x07a49, 0x07a61, 0x07a62, 0x07a69, 0x09f9d,
  0x07a70, 0x07a79, 0x07a7d, 0x07a88, 0x07a97, 0x07a95, 0x07a98, 0x07a96,
  0x07aa9, 0x07ac8, 0x07ab0, 0x07ab6, 0x07ac5, 0x07ac4, 0x07abf, 0x09083,
  0x07ac7, 0x07aca, 0x07acd, 0x07acf, 0x07ad5, 0x07ad3, 0x07ad9, 0x07ada,
  0x07add, 0x07ae1, 0x07ae2, 0x07ae6, 0x07aed, 0x07af0, 0x07b02, 0x07b0f,
  0x07b0a, 0x07b06, 0x07b33, 0x07b18, 0x07b19, 0x07b1e, 0x07b35, 0x07b28,
  0x07b36, 0x07b50, 0x07b7a, 0x07b04, 0x07b4d, 0x07b0b, 0x07b4c, 0x07b45,
  0x07b75, 0x07b65, 0x07b74, 0x07b67, 0x07b70, 0x07b71, 0x07b6c, 0x07b6e,
  0x07b9d, 0x07b98, 0x07b9f, 0x07b8d, 0x07b9c, 0x07b9a, 0x07b8b, 0x07b92,
  0x07b8f, 0x07b5d, 0x07b99, 0x07bcb, 0x07bc1, 0x07bcc, 0x07bcf, 0x07bb4,
  0x07bc6, 0x07bdd, 0x07be9, 0x07c11, 0x07c14, 0x07be6, 0x07be5, 0x07c60,
  0x07c00, 0x07c07, 0x07c13, 0x07bf3, 0x07bf7, 0x07c17, 0x07c0d, 0x07bf6,
  0x07c23, 0x07c27, 0x07c2a, 0x07c1f, 0x07c37, 0x07c2b, 0x07c3d, 0x07c4c,
  0x07c43, 0x07c54, 0x07c4f, 0x07c40, 0x07c50, 0x07c58, 0x07c5f, 0x07c64,
  0x07c56, 0x07c65, 0x07c6c, 0x07c75, 0x07c83, 0x07c90, 0x07ca4, 0x07cad,
  0x07ca2, 0x07cab, 0x07ca1, 0x07ca8, 0x07cb3, 0x07cb2, 0x07cb1, 0x07cae,
  0x07cb9, 0x07cbd, 0x07cc0, 0x07cc5, 0x07cc2, 0x07cd8, 0x07cd2, 0x07cdc,
  0x07ce2, 0x09b3b, 0x07cef, 0x07cf2, 0x07cf4, 0x07cf6, 0x07cfa, 0x07d06,
  0x07d02, 0x07d1c, 0x07d15, 0x07d0a, 0x07d45, 0x07d4b, 0x07d2e, 0x07d32,
  0x07d3f, 0x07d35, 0x07d46, 0x07d73, 0x07d56, 0x07d4e, 0x07d72, 0x07d68,
  0x07d6e, 0x07d4f, 0x07d63, 0x07d93, 0x07d89, 0x07d5b, 0x07d8f, 0x07d7d,
  0x07d9b, 0x07dba, 0x07dae, 0x07da3, 0x07db5, 0x07dc7, 0x07dbd, 0x07dab,
  0x07e3d, 0x07da2, 0x07daf, 0x07ddc, 0x07db8, 0x07d9f, 0x07db0, 0x07dd8,
  0x07ddd, 0x07de4, 0x07dde, 0x07dfb, 0x07df2, 0x07de1, 0x07e05, 0x07e0a,
  0x07e23, 0x07e21, 0x07e12, 0x07e31, 0x07e1f, 0x07e09, 0x07e0b, 0x07e22,
  0x07e46, 0x07e66, 0x07e3b, 0x07e35, 0x07e39, 0x07e43, 0x07e37, 0x07e32,
  0x07e3a, 0x07e67, 0x07e5d, 0x07e56, 0x07e5e, 0x07e59, 0x07e5a, 0x07e79,
  0x07e6a, 0x07e69, 0x07e7c, 0x07e7b, 0x07e83, 0x07dd5, 0x07e7d, 0x08fae,
  0x07e7f, 0x07e88, 0x07e89, 0x07e8c, 0x07e92, 0x07e90, 0x07e93, 0x07e94,
  0x07e96, 0x07e8e, 0x07e9b, 0x07e9c, 0x07f38, 0x07f3a, 0x07f45, 0x07f4c,
  0x07f4d, 0x07f4e, 0x07f50, 0x07f51, 0x07f55, 0x07f54, 0x07f58, 0x07f5f,
  0x07f60, 0x07f68, 0x07f69, 0x07f67, 0x07f78, 0x07f82, 0x07f86, 0x07f83,
  0x07f88, 0x07f87, 0x07f8c, 0x07f94, 0x07f9e, 0x07f9d, 0x07f9a, 0x07fa3,
  0x07faf, 0x07fb2, 0x07fb9, 0x07fae, 0x07fb6, 0x07fb8, 0x08b71, 0x07fc5,
  0x07fc6, 0x07fca, 0x07fd5, 0x07fd4, 0x07fe1, 0x07fe6, 0x07fe9, 0x07ff3,
  0x07ff9, 0x098dc, 0x08006, 0x08004, 0x0800b, 0x08012, 0x08018, 0x08019,
  0x0801c, 0x08021, 0x08028, 0x0803f, 0x0803b, 0x0804a, 0x08046, 0x08052,
  0x08058, 0x0805a, 0x0805f, 0x08062, 0x08068, 0x08073, 0x08072, 0x08070,
  0x08076, 0x08079, 0x0807d, 0x0807f, 0x08084, 0x08086, 0x08085, 0x0809b,
  0x08093, 0x0809a, 0x080ad, 0x05190, 0x080ac, 0x080db, 0x080e5, 0x080d9,
  0x080dd, 0x080c4, 0x080da, 0x080d6, 0x08109, 0x080ef, 0x080f1, 0x0811b,
  0x08129, 0x08123, 0x0812f, 0x0814b, 0x0968b, 0x08146, 0x0813e, 0x08153,
  0x08151, 0x080fc, 0x08171, 0x0816e, 0x08165, 0x08166, 0x08174, 0x08183,
  0x08188, 0x0818a, 0x08180, 0x08182, 0x081a0, 0x08195, 0x081a4, 0x081a3,
  0x0815f, 0x08193, 0x081a9, 0x081b0, 0x081b5, 0x081be, 0x081b8, 0x081bd,
  0x081c0, 0x081c2, 0x081ba, 0x081c9, 0x081cd, 0x081d1, 0x081d9, 0x081d8,
  0x081c8, 0x081da, 0x081df, 0x081e0, 0x081e7, 0x081fa, 0x081fb, 0x081fe,
  0x08201, 0x08202, 0x08205, 0x08207, 0x0820a, 0x0820d, 0x08210, 0x08216,
  0x08229, 0x0822b, 0x08238, 0x08233, 0x08240, 0x08259, 0x08258, 0x0825d,
  0x0825a, 0x0825f, 0x08264, 0x08262, 0x08268, 0x0826a, 0x0826b, 0x0822e,
  0x08271, 0x08277, 0x08278, 0x0827e, 0x0828d, 0x08292, 0x082ab, 0x0829f,
  0x082bb, 0x082ac, 0x082e1, 0x082e3, 0x082df, 0x082d2, 0x082f4, 0x082f3,
  0x082fa, 0x08393, 0x08303, 0x082fb, 0x082f9, 0x082de, 0x08306, 0x082dc,
  0x08309, 0x082d9, 0x08335, 0x08334, 0x08316, 0x08332, 0x08331, 0x08340,
  0x08339, 0x08350, 0x08345, 0x0832f, 0x0832b, 0x08317, 0x08318, 0x08385,
  0x0839a, 0x083aa, 0x0839f, 0x083a2, 0x08396, 0x08323, 0x0838e, 0x08387,
  0x0838a, 0x0837c, 0x083b5, 0x08373, 0x08375, 0x083a0, 0x08389, 0x083a8,
  0x083f4, 0x08413, 0x083eb, 0x083ce, 0x083fd, 0x08403, 0x083d8, 0x0840b,
  0x083c1, 0x083f7, 0x08407, 0x083e0, 0x083f2, 0x0840d, 0x08422, 0x08420,
  0x083bd, 0x08438, 0x08506, 0x083fb, 0x0846d, 0x0842a, 0x0843c, 0x0855a,
  0x08484, 0x08477, 0x0846b, 0x084ad, 0x0846e, 0x08482, 0x08469, 0x08446,
  0x0842c, 0x0846f, 0x08479, 0x08435, 0x084ca, 0x08462, 0x084b9, 0x084bf,
  0x0849f, 0x084d9, 0x084cd, 0x084bb, 0x084da, 0x084d0, 0x084c1, 0x084c6,
  0x084d6, 0x084a1, 0x08521, 0x084ff, 0x084f4, 0x08517, 0x08518, 0x0852c,
  0x0851f, 0x08515, 0x08514, 0x084fc, 0x08540, 0x08563, 0x08558, 0x08548,
  0x08541, 0x08602, 0x0854b, 0x08555, 0x08580, 0x085a4, 0x08588, 0x08591,
  0x0858a, 0x085a8, 0x0856d, 0x08594, 0x0859b, 0x085ea, 0x08587, 0x0859c,
  0x08577, 0x0857e, 0x08590, 0x085c9, 0x085ba, 0x085cf, 0x085b9, 0x085d0,
  0x085d5, 0x085dd, 0x085e5, 0x085dc, 0x085f9, 0x0860a, 0x08613, 0x0860b,
  0x085fe, 0x085fa, 0x08606, 0x08622, 0x0861a, 0x08630, 0x0863f, 0x0864d,
  0x04e55, 0x08654, 0x0865f, 0x08667, 0x08671, 0x08693, 0x086a3, 0x086a9,
  0x086aa, 0x0868b, 0x0868c, 0x086b6, 0x086af, 0x086c4, 0x086c6, 0x086b0,
  0x086c9, 0x08823, 0x086ab, 0x086d4, 0x086de, 0x086e9, 0x086ec, 0x086df,
  0x086db, 0x086ef, 0x08712, 0x08706, 0x08708, 0x08700, 0x08703, 0x086fb,
  0x08711, 0x08709, 0x0870d, 0x086f9, 0x0870a, 0x08734, 0x0873f, 0x08737,
  0x0873b, 0x08725, 0x08729, 0x0871a, 0x08760, 0x0875f, 0x08778, 0x0874c,
  0x0874e, 0x08774, 0x08757, 0x08768, 0x0876e, 0x08759, 0x08753, 0x08763,
  0x0876a, 0x08805, 0x087a2, 0x0879f, 0x08782, 0x087af, 0x087cb, 0x087bd,
  0x087c0, 0x087d0, 0x096d6, 0x087ab, 0x087c4, 0x087b3, 0x087c7, 0x087c6,
  0x087bb, 0x087ef, 0x087f2, 0x087e0, 0x0880f, 0x0880d, 0x087fe, 0x087f6,
  0x087f7, 0x0880e, 0x087d2, 0x08811, 0x08816, 0x08815, 0x08822, 0x08821,
  0x08831, 0x08836, 0x08839, 0x08827, 0x0883b, 0x08844, 0x08842, 0x08852,
  0x08859, 0x0885e, 0x08862, 0x0886b, 0x08881, 0x0887e, 0x0889e, 0x08875,
  0x0887d, 0x088b5, 0x08872, 0x08882, 0x08897, 0x08892, 0x088ae, 0x08899,
  0x088a2, 0x0888d, 0x088a4, 0x088b0, 0x088bf, 0x088b1, 0x088c3, 0x088c4,
  0x088d4, 0x088d8, 0x088d9, 0x088dd, 0x088f9, 0x08902, 0x088fc, 0x088f4,
  0x088e8, 0x088f2, 0x08904, 0x0890c, 0x0890a, 0x08913, 0x08943, 0x0891e,
  0x08925, 0x0892a, 0x0892b, 0x08941, 0x08944, 0x0893b, 0x08936, 0x08938,
  0x0894c, 0x0891d, 0x08960, 0x0895e, 0x08966, 0x08964, 0x0896d, 0x0896a,
  0x0896f, 0x08974, 0x08977, 0x0897e, 0x08983, 0x08988, 0x0898a, 0x08993,
  0x08998, 0x089a1, 0x089a9, 0x089a6, 0x089ac, 0x089af, 0x089b2, 0x089ba,
  0x089bd, 0x089bf, 0x089c0, 0x089da, 0x089dc, 0x089dd, 0x089e7, 0x089f4,
  0x089f8, 0x08a03, 0x08a16, 0x08a10, 0x08a0c, 0x08a1b, 0x08a1d, 0x08a25,
  0x08a36, 0x08a41, 0x08a5b, 0x08a52, 0x08a46, 0x08a48, 0x08a7c, 0x08a6d,
  0x08a6c, 0x08a62, 0x08a85, 0x08a82, 0x08a84, 0x08aa8, 0x08aa1, 0x08a91,
  0x08aa5, 0x08aa6, 0x08a9a, 0x08aa3, 0x08ac4, 0x08acd, 0x08ac2, 0x08ada,
  0x08aeb, 0x08af3, 0x08ae7, 0x08ae4, 0x08af1, 0x08b14, 0x08ae0, 0x08ae2,
  0x08af7, 0x08ade, 0x08adb, 0x08b0c, 0x08b07, 0x08b1a, 0x08ae1, 0x08b16,
  0x08b10, 0x08b17, 0x08b20, 0x08b33, 0x097ab, 0x08b26, 0x08b2b, 0x08b3e,
  0x08b28, 0x08b41, 0x08b4c, 0x08b4f, 0x08b4e, 0x08b49, 0x08b56, 0x08b5b,
  0x08b5a, 0x08b6b, 0x08b5f, 0x08b6c, 0x08b6f, 0x08b74, 0x08b7d, 0x08b80,
  0x08b8c, 0x08b8e, 0x08b92, 0x08b93, 0x08b96, 0x08b99, 0x08b9a, 0x08c3a,
  0x08c41, 0x08c3f, 0x08c48, 0x08c4c, 0x08c4e, 0x08c50, 0x08c55, 0x08c62,
  0x08c6c, 0x08c78, 0x08c7a, 0x08c82, 0x08c89, 0x08c85, 0x08c8a, 0x08c8d,
  0x08c8e, 0x08c94, 0x08c7c, 0x08c98, 0x0621d, 0x08cad, 0x08caa, 0x08cbd,
  0x08cb2, 0x08cb3, 0x08cae, 0x08cb6, 0x08cc8, 0x08cc1, 0x08ce4, 0x08ce3,
  0x08cda, 0x08cfd, 0x08cfa, 0x08cfb, 0x08d04, 0x08d05, 0x08d0a, 0x08d07,
  0x08d0f, 0x08d0d, 0x08d10, 0x09f4e, 0x08d13, 0x08ccd, 0x08d14, 0x08d16,
  0x08d67, 0x08d6d, 0x08d71, 0x08d73, 0x08d81, 0x08d99, 0x08dc2, 0x08dbe,
  0x08dba, 0x08dcf, 0x08dda, 0x08dd6, 0x08dcc, 0x08ddb, 0x08dcb, 0x08dea,
  0x08deb, 0x08ddf, 0x08de3, 0x08dfc, 0x08e08, 0x08e09, 0x08dff, 0x08e1d,
  0x08e1e, 0x08e10, 0x08e1f, 0x08e42, 0x08e35, 0x08e30, 0x08e34, 0x08e4a,
  0x08e47, 0x08e49, 0x08e4c, 0x08e50, 0x08e48, 0x08e59, 0x08e64, 0x08e60,
  0x08e2a, 0x08e63, 0x08e55, 0x08e76, 0x08e72, 0x08e7c, 0x08e81, 0x08e87,
  0x08e85, 0x08e84, 0x08e8b, 0x08e8a, 0x08e93, 0x08e91, 0x08e94, 0x08e99,
  0x08eaa, 0x08ea1, 0x08eac, 0x08eb0, 0x08ec6, 0x08eb1, 0x08ebe, 0x08ec5,
  0x08ec8, 0x08ecb, 0x08edb, 0x08ee3, 0x08efc, 0x08efb, 0x08eeb, 0x08efe,
  0x08f0a, 0x08f05, 0x08f15, 0x08f12, 0x08f19, 0x08f13, 0x08f1c, 0x08f1f,
  0x08f1b, 0x08f0c, 0x08f26, 0x08f33, 0x08f3b, 0x08f39, 0x08f45, 0x08f42,
  0x08f3e, 0x08f4c, 0x08f49, 0x08f46, 0x08f4e, 0x08f57, 0x08f5c, 0x08f62,
  0x08f63, 0x08f64, 0x08f9c, 0x08f9f, 0x08fa3, 0x08fad, 0x08faf, 0x08fb7,
  0x08fda, 0x08fe5, 0x08fe2, 0x08fea, 0x08fef, 0x09087, 0x08ff4, 0x09005,
  0x08ff9, 0x08ffa, 0x09011, 0x09015, 0x09021, 0x0900d, 0x0901e, 0x09016,
  0x0900b, 0x09027, 0x09036, 0x09035, 0x09039, 0x08ff8, 0x0904f, 0x09050,
  0x09051, 0x09052, 0x0900e, 0x09049, 0x0903e, 0x09056, 0x09058, 0x0905e,
  0x09068, 0x0906f, 0x09076, 0x096a8, 0x09072, 0x09082, 0x0907d, 0x09081,
  0x09080, 0x0908a, 0x09089, 0x0908f, 0x090a8, 0x090af, 0x090b1, 0x090b5,
  0x090e2, 0x090e4, 0x06248, 0x090db, 0x09102, 0x09112, 0x09119, 0x09132,
  0x09130, 0x0914a, 0x09156, 0x09158, 0x09163, 0x09165, 0x09169, 0x09173,
  0x09172, 0x0918b, 0x09189, 0x09182, 0x091a2, 0x091ab, 0x091af, 0x091aa,
  0x091b5, 0x091b4, 0x091ba, 0x091c0, 0x091c1, 0x091c9, 0x091cb, 0x091d0,
  0x091d6, 0x091df, 0x091e1, 0x091db, 0x091fc, 0x091f5, 0x091f6, 0x0921e,
  0x091ff, 0x09214, 0x0922c, 0x09215, 0x09211, 0x0925e, 0x09257, 0x09245,
  0x09249, 0x09264, 0x09248, 0x09295, 0x0923f, 0x0924b, 0x09250, 0x0929c,
  0x09296, 0x09293, 0x0929b, 0x0925a, 0x092cf, 0x092b9, 0x092b7, 0x092e9,
  0x0930f, 0x092fa, 0x09344, 0x0932e, 0x09319, 0x09322, 0x0931a, 0x09323,
  0x0933a, 0x09335, 0x0933b, 0x0935c, 0x09360, 0x0937c, 0x0936e, 0x09356,
  0x093b0, 0x093ac, 0x093ad, 0x09394, 0x093b9, 0x093d6, 0x093d7, 0x093e8,
  0x093e5, 0x093d8, 0x093c3, 0x093dd, 0x093d0, 0x093c8, 0x093e4, 0x0941a,
  0x09414, 0x09413, 0x09403, 0x09407, 0x09410, 0x09436, 0x0942b, 0x09435,
  0x09421, 0x0943a, 0x09441, 0x09452, 0x09444, 0x0945b, 0x09460, 0x09462,
  0x0945e, 0x0946a, 0x09229, 0x09470, 0x09475, 0x09477, 0x0947d, 0x0945a,
  0x0947c, 0x0947e, 0x09481, 0x0947f, 0x09582, 0x09587, 0x0958a, 0x09594,
  0x09596, 0x09598, 0x09599, 0x095a0, 0x095a8, 0x095a7, 0x095ad, 0x095bc,
  0x095bb, 0x095b9, 0x095be, 0x095ca, 0x06ff6, 0x095c3, 0x095cd, 0x095cc,
  0x095d5, 0x095d4, 0x095d6, 0x095dc, 0x095e1, 0x095e5, 0x095e2, 0x09621,
  0x09628, 0x0962e, 0x0962f, 0x09642, 0x0964c, 0x0964f, 0x0964b, 0x09677,
  0x0965c, 0x0965e, 0x0965d, 0x0965f, 0x09666, 0x09672, 0x0966c, 0x0968d,
  0x09698, 0x09695, 0x09697, 0x096aa, 0x096a7, 0x096b1, 0x096b2, 0x096b0,
  0x096b4, 0x096b6, 0x096b8, 0x096b9, 0x096ce, 0x096cb, 0x096c9, 0x096cd,
  0x0894d, 0x096dc, 0x0970d, 0x096d5, 0x096f9, 0x09704, 0x09706, 0x09708,
  0x09713, 0x0970e, 0x09711, 0x0970f, 0x09716, 0x09719, 0x09724, 0x0972a,
  0x09730, 0x09739, 0x0973d, 0x0973e, 0x09744, 0x09746, 0x09748, 0x09742,
  0x09749, 0x0975c, 0x09760, 0x09764, 0x09766, 0x09768, 0x052d2, 0x0976b,
  0x09771, 0x09779, 0x09785, 0x0977c, 0x09781, 0x0977a, 0x09786, 0x0978b,
  0x0978f, 0x09790, 0x0979c, 0x097a8, 0x097a6, 0x097a3, 0x097b3, 0x097b4,
  0x097c3, 0x097c6, 0x097c8, 0x097cb, 0x097dc, 0x097ed, 0x09f4f, 0x097f2,
  0x07adf, 0x097f6, 0x097f5, 0x0980f, 0x0980c, 0x09838, 0x09824, 0x09821,
  0x09837, 0x0983d, 0x09846, 0x0984f, 0x0984b, 0x0986b, 0x0986f, 0x09870,
  0x09871, 0x09874, 0x09873, 0x098aa, 0x098af, 0x098b1, 0x098b6, 0x098c4,
  0x098c3, 0x098c6, 0x098e9, 0x098eb, 0x09903, 0x09909, 0x09912, 0x09914,
  0x09918, 0x09921, 0x0991d, 0x0991e, 0x09924, 0x09920, 0x0992c, 0x0992e,
  0x0993d, 0x0993e, 0x09942, 0x09949, 0x09945, 0x09950, 0x0994b, 0x09951,
  0x09952, 0x0994c, 0x09955, 0x09997, 0x09998, 0x099a5, 0x099ad, 0x099ae,
  0x099bc, 0x099df, 0x099db, 0x099dd, 0x099d8, 0x099d1, 0x099ed, 0x099ee,
  0x099f1, 0x099f2, 0x099fb, 0x099f8, 0x09a01, 0x09a0f, 0x09a05, 0x099e2,
  0x09a19, 0x09a2b, 0x09a37, 0x09a45, 0x09a42, 0x09a40, 0x09a43, 0x09a3e,
  0x09a55, 0x09a4d, 0x09a5b, 0x09a57, 0x09a5f, 0x09a62, 0x09a65, 0x09a64,
  0x09a69, 0x09a6b, 0x09a6a, 0x09aad, 0x09ab0, 0x09abc, 0x09ac0, 0x09acf,
  0x09ad1, 0x09ad3, 0x09ad4, 0x09ade, 0x09adf, 0x09ae2, 0x09ae3, 0x09ae6,
  0x09aef, 0x09aeb, 0x09aee, 0x09af4, 0x09af1, 0x09af7, 0x09afb, 0x09b06,
  0x09b18, 0x09b1a, 0x09b1f, 0x09b22, 0x09b23, 0x09b25, 0x09b27, 0x09b28,
  0x09b29, 0x09b2a, 0x09b2e, 0x09b2f, 0x09b32, 0x09b44, 0x09b43, 0x09b4f,
  0x09b4d, 0x09b4e, 0x09b51, 0x09b58, 0x09b74, 0x09b93, 0x09b83, 0x09b91,
  0x09b96, 0x09b97, 0x09b9f, 0x09ba0, 0x09ba8, 0x09bb4, 0x09bc0, 0x09bca,
  0x09bb9, 0x09bc6, 0x09bcf, 0x09bd1, 0x09bd2, 0x09be3, 0x09be2, 0x09be4,
  0x09bd4, 0x09be1, 0x09c3a, 0x09bf2, 0x09bf1, 0x09bf0, 0x09c15, 0x09c14,
  0x09c09, 0x09c13, 0x09c0c, 0x09c06, 0x09c08, 0x09c12, 0x09c0a, 0x09c04,
  0x09c2e, 0x09c1b, 0x09c25, 0x09c24, 0x09c21, 0x09c30, 0x09c47, 0x09c32,
  0x09c46, 0x09c3e, 0x09c5a, 0x09c60, 0x09c67, 0x09c76, 0x09c78, 0x09ce7,
  0x09cec, 0x09cf0, 0x09d09, 0x09d08, 0x09ceb, 0x09d03, 0x09d06, 0x09d2a,
  0x09d26, 0x09daf, 0x09d23, 0x09d1f, 0x09d44, 0x09d15, 0x09d12, 0x09d41,
  0x09d3f, 0x09d3e, 0x09d46, 0x09d48, 0x09d5d, 0x09d5e, 0x09d64, 0x09d51,
  0x09d50, 0x09d59, 0x09d72, 0x09d89, 0x09d87, 0x09dab, 0x09d6f, 0x09d7a,
  0x09d9a, 0x09da4, 0x09da9, 0x09db2, 0x09dc4, 0x09dc1, 0x09dbb, 0x09db8,
  0x09dba, 0x09dc6, 0x09dcf, 0x09dc2, 0x09dd9, 0x09dd3, 0x09df8, 0x09de6,
  0x09ded, 0x09def, 0x09dfd, 0x09e1a, 0x09e1b, 0x09e1e, 0x09e75, 0x09e79,
  0x09e7d, 0x09e81, 0x09e88, 0x09e8b, 0x09e8c, 0x09e92, 0x09e95, 0x09e91,
  0x09e9d, 0x09ea5, 0x09ea9, 0x09eb8, 0x09eaa, 0x09ead, 0x09761, 0x09ecc,
  0x09ece, 0x09ecf, 0x09ed0, 0x09ed4, 0x09edc, 0x09ede, 0x09edd, 0x09ee0,
  0x09ee5, 0x09ee8, 0x09eef, 0x09ef4, 0x09ef6, 0x09ef7, 0x09ef9, 0x09efb,
  0x09efc, 0x09efd, 0x09f07, 0x09f08, 0x076b7, 0x09f15, 0x09f21, 0x09f2c,
  0x09f3e, 0x09f4a, 0x09f52, 0x09f54, 0x09f63, 0x09f5f, 0x09f60, 0x09f61,
  0x09f66, 0x09f67, 0x09f6c, 0x09f6a, 0x09f77, 0x09f72, 0x09f76, 0x09f95,
  0x09f9c, 0x09fa0, 0x0582f, 0x069c7, 0x09059, 0x07464, 0x051dc, 0x07199,
  0x05653, 0x05de2, 0x05e14, 0x05e18, 0x05e58, 0x05e5e, 0x05ebe, 0x0f928,
  0x05ecb, 0x05ef9, 0x05f00, 0x05f02, 0x05f07, 0x05f1d, 0x05f23, 0x05f34,
  0x05f36, 0x05f3d, 0x05f40, 0x05f45, 0x05f54, 0x05f58, 0x05f64, 0x05f67,
  0x05f7d, 0x05f89, 0x05f9c, 0x05fa7, 0x05faf, 0x05fb5, 0x05fb7, 0x05fc9,
  0x05fde, 0x05fe1, 0x05fe9, 0x0600d, 0x06014, 0x06018, 0x06033, 0x06035,
  0x06047, 0x0fa3d, 0x0609d, 0x0609e, 0x060cb, 0x060d4, 0x060d5, 0x060dd,
  0x060f8, 0x0611c, 0x0612b, 0x06130, 0x06137, 0x0fa3e, 0x0618d, 0x0fa3f,
  0x061bc, 0x061b9, 0x0fa40, 0x06222, 0x0623e, 0x06243, 0x06256, 0x0625a,
  0x0626f, 0x06285, 0x062c4, 0x062d6, 0x062fc, 0x0630a, 0x06318, 0x06339,
  0x06343, 0x06365, 0x0637c, 0x063e5, 0x063ed, 0x063f5, 0x06410, 0x06414,
  0x06422, 0x06479, 0x06451, 0x06460, 0x0646d, 0x064ce, 0x064be, 0x064bf,
  0x064c4, 0x064ca, 0x064d0, 0x064f7, 0x064fb, 0x06522, 0x06529, 0x0fa41,
  0x06567, 0x0659d, 0x0fa42, 0x06600, 0x06609, 0x06615, 0x0661e, 0x0663a,
  0x06622, 0x06624, 0x0662b, 0x06630, 0x06631, 0x06633, 0x066fb, 0x06648,
  0x0664c, 0x231c4, 0x06659, 0x0665a, 0x06661, 0x06665, 0x06673, 0x06677,
  0x06678, 0x0668d, 0x0fa43, 0x066a0, 0x066b2, 0x066bb, 0x066c6, 0x066c8,
  0x03b22, 0x066db, 0x066e8, 0x066fa, 0x06713, 0x0f929, 0x06733, 0x06766,
  0x06747, 0x06748, 0x0677b, 0x06781, 0x06793, 0x06798, 0x0679b, 0x067bb,
  0x067f9, 0x067c0, 0x067d7, 0x067fc, 0x06801, 0x06852, 0x0681d, 0x0682c,
  0x06831, 0x0685b, 0x06872, 0x06875, 0x0fa44, 0x068a3, 0x068a5, 0x068b2,
  0x068c8, 0x068d0, 0x068e8, 0x068ed, 0x068f0, 0x068f1, 0x068fc, 0x0690a,
  0x06949, 0x235c4, 0x06935, 0x06942, 0x06957, 0x06963, 0x06964, 0x06968,
  0x06980, 0x0fa14, 0x069a5, 0x069ad, 0x069cf, 0x03bb6, 0x03bc3, 0x069e2,
  0x069e9, 0x069ea, 0x069f5, 0x069f6, 0x06a0f, 0x06a15, 0x2373f, 0x06a3b,
  0x06a3e, 0x06a45, 0x06a50, 0x06a56, 0x06a5b, 0x06a6b, 0x06a73, 0x23763,
  0x06a89, 0x06a94, 0x06a9d, 0x06a9e, 0x06aa5, 0x06ae4, 0x06ae7, 0x03c0f,
  0x0f91d, 0x06b1b, 0x06b1e, 0x06b2c, 0x06b35, 0x06b46, 0x06b56, 0x06b60,
  0x06b65, 0x06b67, 0x06b77, 0x06b82, 0x06ba9, 0x06bad, 0x0f970, 0x06bcf,
  0x06bd6, 0x06bd7, 0x06bff, 0x06c05, 0x06c10, 0x06c33, 0x06c59, 0x06c5c,
  0x06caa, 0x06c74, 0x06c76, 0x06c85, 0x06c86, 0x06c98, 0x06c9c, 0x06cfb,
  0x06cc6, 0x06cd4, 0x06ce0, 0x06ceb, 0x06cee, 0x23cfe, 0x06d04, 0x06d0e,
  0x06d2e, 0x06d31, 0x06d39, 0x06d3f, 0x06d58, 0x06d65, 0x0fa45, 0x06d82,
  0x06d87, 0x06d89, 0x06d94, 0x06daa, 0x06dac, 0x06dbf, 0x06dc4, 0x06dd6,
  0x06dda, 0x06ddb, 0x06ddd, 0x06dfc, 0x0fa46, 0x06e34, 0x06e44, 0x06e5c,
  0x06e5e, 0x06eab, 0x06eb1, 0x06ec1, 0x06ec7, 0x06ece, 0x06f10, 0x06f1a,
  0x0fa47, 0x06f2a, 0x06f2f, 0x06f33, 0x06f51, 0x06f59, 0x06f5e, 0x06f61,
  0x06f62, 0x06f7e, 0x06f88, 0x06f8c, 0x06f8d, 0x06f94, 0x06fa0, 0x06fa7,
  0x06fb6, 0x06fbc, 0x06fc7, 0x06fca, 0x06ff9, 0x06ff0, 0x06ff5, 0x07005,
  0x07006, 0x07028, 0x0704a, 0x0705d, 0x0705e, 0x0704e, 0x07064, 0x07075,
  0x07085, 0x070a4, 0x070ab, 0x070b7, 0x070d4, 0x070d8, 0x070e4, 0x0710f,
  0x0712b, 0x0711e, 0x07120, 0x0712e, 0x07130, 0x07146, 0x07147, 0x07151,
  0x0fa48, 0x07152, 0x0715c, 0x07160, 0x07168, 0x0fa15, 0x07185, 0x07187,
  0x07192, 0x071c1, 0x071ba, 0x071c4, 0x071fe, 0x07200, 0x07215, 0x07255,
  0x07256, 0x03e3f, 0x0728d, 0x0729b, 0x072be, 0x072c0, 0x072fb, 0x247f1,
  0x07327, 0x07328, 0x0fa16, 0x07350, 0x07366, 0x0737c, 0x07395, 0x0739f,
  0x073a0, 0x073a2, 0x073a6, 0x073ab, 0x073c9, 0x073cf, 0x073d6, 0x073d9,
  0x073e3, 0x073e9, 0x07407, 0x0740a, 0x0741a, 0x0741b, 0x0fa4a, 0x07426,
  0x07428, 0x0742a, 0x0742b, 0x0742c, 0x0742e, 0x0742f, 0x07430, 0x07444,
  0x07446, 0x07447, 0x0744b, 0x07457, 0x07462, 0x0746b, 0x0746d, 0x07486,
  0x07487, 0x07489, 0x07498, 0x0749c, 0x0749f, 0x074a3, 0x07490, 0x074a6,
  0x074a8, 0x074a9, 0x074b5, 0x074bf, 0x074c8, 0x074c9, 0x074da, 0x074ff,
  0x07501, 0x07517, 0x0752f, 0x0756f, 0x07579, 0x07592, 0x03f72, 0x075ce,
  0x075e4, 0x07600, 0x07602, 0x07608, 0x07615, 0x07616, 0x07619, 0x0761e,
  0x0762d, 0x07635, 0x07643, 0x0764b, 0x07664, 0x07665, 0x0766d, 0x0766f,
  0x07671, 0x07681, 0x0769b, 0x0769d, 0x0769e, 0x076a6, 0x076aa, 0x076b6,
  0x076c5, 0x076cc, 0x076ce, 0x076d4, 0x076e6, 0x076f1, 0x076fc, 0x0770a,
  0x07719, 0x07734, 0x07736, 0x07746, 0x0774d, 0x0774e, 0x0775c, 0x0775f,
  0x07762, 0x0777a, 0x07780, 0x07794, 0x077aa, 0x077e0, 0x0782d, 0x2548e,
  0x07843, 0x0784e, 0x0784f, 0x07851, 0x07868, 0x0786e, 0x0fa4b, 0x078b0,
  0x2550e, 0x078ad, 0x078e4, 0x078f2, 0x07900, 0x078f7, 0x0791c, 0x0792e,
  0x07931, 0x07934, 0x0fa4c, 0x0fa4d, 0x07945, 0x07946, 0x0fa4e, 0x0fa4f,
  0x0fa50, 0x0795c, 0x0fa51, 0x0fa19, 0x0fa1a, 0x07979, 0x0fa52, 0x0fa53,
  0x0fa1b, 0x07998, 0x079b1, 0x079b8, 0x079c8, 0x079ca, 0x25771, 0x079d4,
  0x079de, 0x079eb, 0x079ed, 0x07a03, 0x0fa54, 0x07a39, 0x07a5d, 0x07a6d,
  0x0fa55, 0x07a85, 0x07aa0, 0x259c4, 0x07ab3, 0x07abb, 0x07ace, 0x07aeb,
  0x07afd, 0x07b12, 0x07b2d, 0x07b3b, 0x07b47, 0x07b4e, 0x07b60, 0x07b6d,
  0x07b6f, 0x07b72, 0x07b9e, 0x0fa56, 0x07bd7, 0x07bd9, 0x07c01, 0x07c31,
  0x07c1e, 0x07c20, 0x07c33, 0x07c36, 0x04264, 0x25da1, 0x07c59, 0x07c6d,
  0x07c79, 0x07c8f, 0x07c94, 0x07ca0, 0x07cbc, 0x07cd5, 0x07cd9, 0x07cdd,
  0x07d07, 0x07d08, 0x07d13, 0x07d1d, 0x07d23, 0x07d31, 0x07d41, 0x07d48,
  0x07d53, 0x07d5c, 0x07d7a, 0x07d83, 0x07d8b, 0x07da0, 0x07da6, 0x07dc2,
  0x07dcc, 0x07dd6, 0x07de3, 0x0fa57, 0x07e28, 0x07e08, 0x07e11, 0x07e15,
  0x0fa59, 0x07e47, 0x07e52, 0x07e61, 0x07e8a, 0x07e8d, 0x07f47, 0x0fa5a,
  0x07f91, 0x07f97, 0x07fbf, 0x07fce, 0x07fdb, 0x07fdf, 0x07fec, 0x07fee,
  0x07ffa, 0x0fa5b, 0x08014, 0x08026, 0x08035, 0x08037, 0x0803c, 0x080ca,
  0x080d7, 0x080e0, 0x080f3, 0x08118, 0x0814a, 0x08160, 0x08167, 0x08168,
  0x0816d, 0x081bb, 0x081ca, 0x081cf, 0x081d7, 0x0fa5c, 0x04453, 0x0445b,
  0x08260, 0x08274, 0x26aff, 0x0828e, 0x082a1, 0x082a3, 0x082a4, 0x082a9,
  0x082ae, 0x082b7, 0x082be, 0x082bf, 0x082c6, 0x082d5, 0x082fd, 0x082fe,
  0x08300, 0x08301, 0x08362, 0x08322, 0x0832d, 0x0833a, 0x08343, 0x08347,
  0x08351, 0x08355, 0x0837d, 0x08386, 0x08392, 0x08398, 0x083a7, 0x083a9,
  0x083bf, 0x083c0, 0x083c7, 0x083cf, 0x083d1, 0x083e1, 0x083ea, 0x08401,
  0x08406, 0x0840a, 0x0fa5f, 0x08448, 0x0845f, 0x08470, 0x08473, 0x08485,
  0x0849e, 0x084af, 0x084b4, 0x084ba, 0x084c0, 0x084c2, 0x26e40, 0x08532,
  0x0851e, 0x08523, 0x0852f, 0x08559, 0x08564, 0x0fa1f, 0x085ad, 0x0857a,
  0x0858c, 0x0858f, 0x085a2, 0x085b0, 0x085cb, 0x085ce, 0x085ed, 0x08612,
  0x085ff, 0x08604, 0x08605, 0x08610, 0x270f4, 0x08618, 0x08629, 0x08638,
  0x08657, 0x0865b, 0x0f936, 0x08662, 0x0459d, 0x0866c, 0x08675, 0x08698,
  0x086b8, 0x086fa, 0x086fc, 0x086fd, 0x0870b, 0x08771, 0x08787, 0x08788,
  0x087ac, 0x087ad, 0x087b5, 0x045ea, 0x087d6, 0x087ec, 0x08806, 0x0880a,
  0x08810, 0x08814, 0x0881f, 0x08898, 0x088aa, 0x088ca, 0x088ce, 0x27684,
  0x088f5, 0x0891c, 0x0fa60, 0x08918, 0x08919, 0x0891a, 0x08927, 0x08930,
  0x08932, 0x08939, 0x08940, 0x08994, 0x0fa61, 0x089d4, 0x089e5, 0x089f6,
  0x08a12, 0x08a15, 0x08a22, 0x08a37, 0x08a47, 0x08a4e, 0x08a5d, 0x08a61,
  0x08a75, 0x08a79, 0x08aa7, 0x08ad0, 0x08adf, 0x08af4, 0x08af6, 0x0fa22,
  0x0fa62, 0x0fa63, 0x08b46, 0x08b54, 0x08b59, 0x08b69, 0x08b9d, 0x08c49,
  0x08c68, 0x0fa64, 0x08ce1, 0x08cf4, 0x08cf8, 0x08cfe, 0x0fa65, 0x08d12,
  0x08d1b, 0x08daf, 0x08dce, 0x08dd1, 0x08dd7, 0x08e20, 0x08e23, 0x08e3d,
  0x08e70, 0x08e7b, 0x28277, 0x08ec0, 0x04844, 0x08efa, 0x08f1e, 0x08f2d,
  0x08f36, 0x08f54, 0x283cd, 0x08fa6, 0x08fb5, 0x08fe4, 0x08fe8, 0x08fee,
  0x09008, 0x0902d, 0x0fa67, 0x09088, 0x09095, 0x09097, 0x09099, 0x0909b,
  0x090a2, 0x090b3, 0x090be, 0x090c4, 0x090c5, 0x090c7, 0x090d7, 0x090dd,
  0x090de, 0x090ef, 0x090f4, 0x0fa26, 0x09114, 0x09115, 0x09116, 0x09122,
  0x09123, 0x09127, 0x0912f, 0x09131, 0x09134, 0x0913d, 0x09148, 0x0915b,
  0x09183, 0x0919e, 0x091ac, 0x091b1, 0x091bc, 0x091d7, 0x091fb, 0x091e4,
  0x091e5, 0x091ed, 0x091f1, 0x09207, 0x09210, 0x09238, 0x09239, 0x0923a,
  0x0923c, 0x09240, 0x09243, 0x0924f, 0x09278, 0x09288, 0x092c2, 0x092cb,
  0x092cc, 0x092d3, 0x092e0, 0x092ff, 0x09304, 0x0931f, 0x09321, 0x09325,
  0x09348, 0x09349, 0x0934a, 0x09364, 0x09365, 0x0936a, 0x09370, 0x0939b,
  0x093a3, 0x093ba, 0x093c6, 0x093de, 0x093df, 0x09404, 0x093fd, 0x09433,
  0x0944a, 0x09463, 0x0946b, 0x09471, 0x09472, 0x0958e, 0x0959f, 0x095a6,
  0x095a9, 0x095ac, 0x095b6, 0x095bd, 0x095cb, 0x095d0, 0x095d3, 0x049b0,
  0x095da, 0x095de, 0x09658, 0x09684, 0x0f9dc, 0x0969d, 0x096a4, 0x096a5,
  0x096d2, 0x096de, 0x0fa68, 0x096e9, 0x096ef, 0x09733, 0x0973b, 0x0974d,
  0x0974e, 0x0974f, 0x0975a, 0x0976e, 0x09773, 0x09795, 0x097ae, 0x097ba,
  0x097c1, 0x097c9, 0x097de, 0x097db, 0x097f4, 0x0fa69, 0x0980a, 0x0981e,
  0x0982b, 0x09830, 0x0fa6a, 0x09852, 0x09853, 0x09856, 0x09857, 0x09859,
  0x0985a, 0x0f9d0, 0x09865, 0x0986c, 0x098ba, 0x098c8, 0x098e7, 0x09958,
  0x0999e, 0x09a02, 0x09a03, 0x09a24, 0x09a2d, 0x09a2e, 0x09a38, 0x09a4a,
  0x09a4e, 0x09a52, 0x09ab6, 0x09ac1, 0x09ac3, 0x09ace, 0x09ad6, 0x09af9,
  0x09b02, 0x09b08, 0x09b20, 0x04c17, 0x09b2d, 0x09b5e, 0x09b79, 0x09b66,
  0x09b72, 0x09b75, 0x09b84, 0x09b8a, 0x09b8f, 0x09b9e, 0x09ba7, 0x09bc1,
  0x09bce, 0x09be5, 0x09bf8, 0x09bfd, 0x09c00, 0x09c23, 0x09c41, 0x09c4f,
  0x09c50, 0x09c53, 0x09c63, 0x09c65, 0x09c77, 0x09d1d, 0x09d1e, 0x09d43,
  0x09d47, 0x09d52, 0x09d63, 0x09d70, 0x09d7c, 0x09d8a, 0x09d96, 0x09dc0,
  0x09dac, 0x09dbc, 0x09dd7, 0x2a190, 0x09de7, 0x09e07, 0x09e15, 0x09e7c,
  0x09e9e, 0x09ea4, 0x09eac, 0x09eaf, 0x09eb4, 0x09eb5, 0x09ec3, 0x09ed1,
  0x09f10, 0x09f39, 0x09f57, 0x09f90, 0x09f94, 0x09f97, 0x09fa2, 0x059f8,
  0x05c5b, 0x05e77, 0x07626, 0x07e6b, 
};

int32_t ucode_kanji2[U_kanji] = {
  0x20089, 0x04e02, 0x04e0f, 0x04e12, 0x04e29, 0x04e2b, 0x04e2e, 0x04e40,
  0x04e47, 0x04e48, 0x200a2, 0x04e51, 0x03406, 0x200a4, 0x04e5a, 0x04e69,
  0x04e9d, 0x0342c, 0x0342e, 0x04eb9, 0x04ebb, 0x201a2, 0x04ebc, 0x04ec3,
  0x04ec8, 0x04ed0, 0x04eeb, 0x04eda, 0x04ef1, 0x04ef5, 0x04f00, 0x04f16,
  0x04f64, 0x04f37, 0x04f3e, 0x04f54, 0x04f58, 0x20213, 0x04f77, 0x04f78,
  0x04f7a, 0x04f7d, 0x04f82, 0x04f85, 0x04f92, 0x04f9a, 0x04fe6, 0x04fb2,
  0x04fbe, 0x04fc5, 0x04fcb, 0x04fcf, 0x04fd2, 0x0346a, 0x04ff2, 0x05000,
  0x05010, 0x05013, 0x0501c, 0x0501e, 0x05022, 0x03468, 0x05042, 0x05046,
  0x0504e, 0x05053, 0x05057, 0x05063, 0x05066, 0x0506a, 0x05070, 0x050a3,
  0x05088, 0x05092, 0x05093, 0x05095, 0x05096, 0x0509c, 0x050aa, 0x2032b,
  0x050b1, 0x050ba, 0x050bb, 0x050c4, 0x050c7, 0x050f3, 0x20381, 0x050ce,
  0x20371, 0x050d4, 0x050d9, 0x050e1, 0x050e9, 0x03492, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, 0x002d8, 0x002c7, 0x000b8, 0x002d9,
  0x002dd, 0x000af, 0x002db, 0x002da, 0x0ff5e, 0x00384, 0x00385, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, 0x000a1,
  0x000a6, 0x000bf, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  0x000ba, 0x000aa, 0x000a9, 0x000ae, 0x02122, 0x000a4, 0x02116, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, 0x05108, 0x203f9, 0x05117, 0x0511b,
  0x2044a, 0x05160, 0x20509, 0x05173, 0x05183, 0x0518b, 0x034bc, 0x05198,
  0x051a3, 0x051ad, 0x034c7, 0x051bc, 0x205d6, 0x20628, 0x051f3, 0x051f4,
  0x05202, 0x05212, 0x05216, 0x2074f, 0x05255, 0x0525c, 0x0526c, 0x05277,
  0x05284, 0x05282, 0x20807, 0x05298, 0x2083a, 0x052a4, 0x052a6, 0x052af,
  0x052ba, 0x052bb, 0x052ca, 0x0351f, 0x052d1, 0x208b9, 0x052f7, 0x0530a,
  0x0530b, 0x05324, 0x05335, 0x0533e, 0x05342, 0x2097c, 0x2099d, 0x05367,
  0x0536c, 0x0537a, 0x053a4, 0x053b4, 0x20ad3, 0x053b7, 0x053c0, 0x20b1d,
  0x0355d, 0x0355e, 0x053d5, 0x053da, 0x03563, 0x053f4, 0x053f5, 0x05455,
  0x05424, 0x05428, 0x0356e, 0x05443, 0x05462, 0x05466, 0x0546c, 0x0548a,
  0x0548d, 0x05495, 0x054a0, 0x054a6, 0x054ad, 0x054ae, 0x054b7, 0x054ba,
  0x054bf, 0x054c3, 0x20d45, 0x054ec, 0x054ef, 0x054f1, 0x054f3, 0x05500,
  0x05501, 0x05509, 0x0553c, 0x05541, 0x035a6, 0x05547, 0x0554a, 0x035a8,
  0x05560, 0x05561, 0x05564, 0x20de1, 0x0557d, 0x05582, 0x05588, 0x05591,
  0x035c5, 0x055d2, 0x20e95, 0x20e6d, 0x055bf, 0x055c9, 0x055cc, 0x055d1,
  0x055dd, 0x035da, 0x055e2, 0x20e64, 0x055e9, 0x05628, 0x20f5f, 0x05607,
  0x05610, 0x05630, 0x05637, 0x035f4, 0x0563d, 0x0563f, 0x05640, 0x05647,
  0x0565e, 0x05660, 0x0566d, 0x03605, 0x05688, 0x0568c, 0x05695, 0x0569a,
  0x0569d, 0x056a8, 0x056ad, 0x056b2, 0x056c5, 0x056cd, 0x056df, 0x056e8,
  0x056f6, 0x056f7, 0x21201, 0x05715, 0x05723, 0x21255, 0x05729, 0x2127b,
  0x05745, 0x05746, 0x0574c, 0x0574d, 0x21274, 0x05768, 0x0576f, 0x05773,
  0x05774, 0x05775, 0x0577b, 0x212e4, 0x212d7, 0x057ac, 0x0579a, 0x0579d,
  0x0579e, 0x057a8, 0x057d7, 0x212fd, 0x057cc, 0x21336, 0x21344, 0x057de,
  0x057e6, 0x057f0, 0x0364a, 0x057f8, 0x057fb, 0x057fd, 0x05804, 0x0581e,
  0x05820, 0x05827, 0x05832, 0x05839, 0x213c4, 0x05849, 0x0584c, 0x05867,
  0x0588a, 0x0588b, 0x0588d, 0x0588f, 0x05890, 0x05894, 0x0589d, 0x058aa,
  0x058b1, 0x2146d, 0x058c3, 0x058cd, 0x058e2, 0x058f3, 0x058f4, 0x05905,
  0x05906, 0x0590b, 0x0590d, 0x05914, 0x05924, 0x215d7, 0x03691, 0x0593d,
  0x03699, 0x05946, 0x03696, 0x26c29, 0x0595b, 0x0595f, 0x21647, 0x05975,
  0x05976, 0x0597c, 0x0599f, 0x059ae, 0x059bc, 0x059c8, 0x059cd, 0x059de,
  0x059e3, 0x059e4, 0x059e7, 0x059ee, 0x21706, 0x21742, 0x036cf, 0x05a0c,
  0x05a0d, 0x05a17, 0x05a27, 0x05a2d, 0x05a55, 0x05a65, 0x05a7a, 0x05a8b,
  0x05a9c, 0x05a9f, 0x05aa0, 0x05aa2, 0x05ab1, 0x05ab3, 0x05ab5, 0x05aba,
  0x05abf, 0x05ada, 0x05adc, 0x05ae0, 0x05ae5, 0x05af0, 0x05aee, 0x05af5,
  0x05b00, 0x05b08, 0x05b17, 0x05b34, 0x05b2d, 0x05b4c, 0x05b52, 0x05b68,
  0x05b6f, 0x05b7c, 0x05b7f, 0x05b81, 0x05b84, 0x219c3, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x00386, 0x00388,
  0x00389, 0x0038a, 0x003aa, U_error, 0x0038c, U_error, 0x0038e, 0x003ab,
  U_error, 0x0038f, U_error, U_error, U_error, U_error, 0x003ac, 0x003ad,
  0x003ae, 0x003af, 0x003ca, 0x00390, 0x003cc, 0x003c2, 0x003cd, 0x003cb,
  0x003b0, 0x003ce, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, 0x00402, 0x00403, 0x00404,
  0x00405, 0x00406, 0x00407, 0x00408, 0x00409, 0x0040a, 0x0040b, 0x0040c,
  0x0040e, 0x0040f, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, 0x00452, 0x00453, 0x00454,
  0x00455, 0x00456, 0x00457, 0x00458, 0x00459, 0x0045a, 0x0045b, 0x0045c,
  0x0045e, 0x0045f, 0x05b96, 0x05bac, 0x03761, 0x05bc0, 0x03762, 0x05bce,
  0x05bd6, 0x0376c, 0x0376b, 0x05bf1, 0x05bfd, 0x03775, 0x05c03, 0x05c29,
  0x05c30, 0x21c56, 0x05c5f, 0x05c63, 0x05c67, 0x05c68, 0x05c69, 0x05c70,
  0x21d2d, 0x21d45, 0x05c7c, 0x21d78, 0x21d62, 0x05c88, 0x05c8a, 0x037c1,
  0x21da1, 0x21d9c, 0x05ca0, 0x05ca2, 0x05ca6, 0x05ca7, 0x21d92, 0x05cad,
  0x05cb5, 0x21db7, 0x05cc9, 0x21de0, 0x21e33, 0x05d06, 0x05d10, 0x05d2b,
  0x05d1d, 0x05d20, 0x05d24, 0x05d26, 0x05d31, 0x05d39, 0x05d42, 0x037e8,
  0x05d61, 0x05d6a, 0x037f4, 0x05d70, 0x21f1e, 0x037fd, 0x05d88, 0x03800,
  0x05d92, 0x05d94, 0x05d97, 0x05d99, 0x05db0, 0x05db2, 0x05db4, 0x21f76,
  0x05db9, 0x05dd1, 0x05dd7, 0x05dd8, 0x05de0, 0x21ffa, 0x05de4, 0x05de9,
  0x0382f, 0x05e00, 0x03836, 0x05e12, 0x05e15, 0x03840, 0x05e1f, 0x05e2e,
  0x05e3e, 0x05e49, 0x0385c, 0x05e56, 0x03861, 0x05e6b, 0x05e6c, 0x05e6d,
  0x000c6, 0x00110, U_error, 0x00126, U_error, 0x00132, U_error, 0x00141,
  0x0013f, U_error, 0x0014a, 0x000d8, 0x00152, U_error, 0x00166, 0x000de,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  0x000e6, 0x00111, 0x000f0, 0x00127, 0x00131, 0x00133, 0x00138, 0x00142,
  0x00140, 0x00149, 0x0014b, 0x000f8, 0x00153, 0x000df, 0x00167, 0x000fe,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x000c1, 0x000c0,
  0x000c4, 0x000c2, 0x00102, 0x001cd, 0x00100, 0x00104, 0x000c5, 0x000c3,
  0x00106, 0x00108, 0x0010c, 0x000c7, 0x0010a, 0x0010e, 0x000c9, 0x000c8,
  0x000cb, 0x000ca, 0x0011a, 0x00116, 0x00112, 0x00118, U_error, 0x0011c,
  0x0011e, 0x00122, 0x00120, 0x00124, 0x000cd, 0x000cc, 0x000cf, 0x000ce,
  0x001cf, 0x00130, 0x0012a, 0x0012e, 0x00128, 0x00134, 0x00136, 0x00139,
  0x0013d, 0x0013b, 0x00143, 0x00147, 0x00145, 0x000d1, 0x000d3, 0x000d2,
  0x000d6, 0x000d4, 0x001d1, 0x00150, 0x0014c, 0x000d5, 0x00154, 0x00158,
  0x00156, 0x0015a, 0x0015c, 0x00160, 0x0015e, 0x00164, 0x00162, 0x000da,
  0x000d9, 0x000dc, 0x000db, 0x0016c, 0x001d3, 0x00170, 0x0016a, 0x00172,
  0x0016e, 0x00168, 0x001d7, 0x001db, 0x001d9, 0x001d5, 0x00174, 0x000dd,
  0x00178, 0x00176, 0x00179, 0x0017d, 0x0017b, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, 0x000e1, 0x000e0, 0x000e4, 0x000e2,
  0x00103, 0x001ce, 0x00101, 0x00105, 0x000e5, 0x000e3, 0x00107, 0x00109,
  0x0010d, 0x000e7, 0x0010b, 0x0010f, 0x000e9, 0x000e8, 0x000eb, 0x000ea,
  0x0011b, 0x00117, 0x00113, 0x00119, 0x001f5, 0x0011d, 0x0011f, U_error,
  0x00121, 0x00125, 0x000ed, 0x000ec, 0x000ef, 0x000ee, 0x001d0, U_error,
  0x0012b, 0x0012f, 0x00129, 0x00135, 0x00137, 0x0013a, 0x0013e, 0x0013c,
  0x00144, 0x00148, 0x00146, 0x000f1, 0x000f3, 0x000f2, 0x000f6, 0x000f4,
  0x001d2, 0x00151, 0x0014d, 0x000f5, 0x00155, 0x00159, 0x00157, 0x0015b,
  0x0015d, 0x00161, 0x0015f, 0x00165, 0x00163, 0x000fa, 0x000f9, 0x000fc,
  0x000fb, 0x0016d, 0x001d4, 0x00171, 0x0016b, 0x00173, 0x0016f, 0x00169,
  0x001d8, 0x001dc, 0x001da, 0x001d6, 0x00175, 0x000fd, 0x000ff, 0x00177,
  0x0017a, 0x0017e, 0x0017c, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, 0x05e6e, 0x2217b, 0x05ea5, 0x05eaa, 0x05eac, 0x05eb9,
  0x05ebf, 0x05ec6, 0x05ed2, 0x05ed9, 0x2231e, 0x05efd, 0x05f08, 0x05f0e,
  0x05f1c, 0x223ad, 0x05f1e, 0x05f47, 0x05f63, 0x05f72, 0x05f7e, 0x05f8f,
  0x05fa2, 0x05fa4, 0x05fb8, 0x05fc4, 0x038fa, 0x05fc7, 0x05fcb, 0x05fd2,
  0x05fd3, 0x05fd4, 0x05fe2, 0x05fee, 0x05fef, 0x05ff3, 0x05ffc, 0x03917,
  0x06017, 0x06022, 0x06024, 0x0391a, 0x0604c, 0x0607f, 0x0608a, 0x06095,
  0x060a8, 0x226f3, 0x060b0, 0x060b1, 0x060be, 0x060c8, 0x060d9, 0x060db,
  0x060ee, 0x060f2, 0x060f5, 0x06110, 0x06112, 0x06113, 0x06119, 0x0611e,
  0x0613a, 0x0396f, 0x06141, 0x06146, 0x06160, 0x0617c, 0x2285b, 0x06192,
  0x06193, 0x06197, 0x06198, 0x061a5, 0x061a8, 0x061ad, 0x228ab, 0x061d5,
  0x061dd, 0x061df, 0x061f5, 0x2298f, 0x06215, 0x06223, 0x06229, 0x06246,
  0x0624c, 0x06251, 0x06252, 0x06261, 0x06264, 0x0627b, 0x0626d, 0x06273,
  0x06299, 0x062a6, 0x062d5, 0x22ab8, 0x062fd, 0x06303, 0x0630d, 0x06310,
  0x22b4f, 0x22b50, 0x06332, 0x06335, 0x0633b, 0x0633c, 0x06341, 0x06344,
  0x0634e, 0x22b46, 0x06359, 0x22c1d, 0x22ba6, 0x0636c, 0x06384, 0x06399,
  0x22c24, 0x06394, 0x063bd, 0x063f7, 0x063d4, 0x063d5, 0x063dc, 0x063e0,
  0x063eb, 0x063ec, 0x063f2, 0x06409, 0x0641e, 0x06425, 0x06429, 0x0642f,
  0x0645a, 0x0645b, 0x0645d, 0x06473, 0x0647d, 0x06487, 0x06491, 0x0649d,
  0x0649f, 0x064cb, 0x064cc, 0x064d5, 0x064d7, 0x22de1, 0x064e4, 0x064e5,
  0x064ff, 0x06504, 0x03a6e, 0x0650f, 0x06514, 0x06516, 0x03a73, 0x0651e,
  0x06532, 0x06544, 0x06554, 0x0656b, 0x0657a, 0x06581, 0x06584, 0x06585,
  0x0658a, 0x065b2, 0x065b5, 0x065b8, 0x065bf, 0x065c2, 0x065c9, 0x065d4,
  0x03ad6, 0x065f2, 0x065f9, 0x065fc, 0x06604, 0x06608, 0x06621, 0x0662a,
  0x06645, 0x06651, 0x0664e, 0x03aea, 0x231c3, 0x06657, 0x0665b, 0x06663,
  0x231f5, 0x231b6, 0x0666a, 0x0666b, 0x0666c, 0x0666d, 0x0667b, 0x06680,
  0x06690, 0x06692, 0x06699, 0x03b0e, 0x066ad, 0x066b1, 0x066b5, 0x03b1a,
  0x066bf, 0x03b1c, 0x066ec, 0x03ad7, 0x06701, 0x06705, 0x06712, 0x23372,
  0x06719, 0x233d3, 0x233d2, 0x0674c, 0x0674d, 0x06754, 0x0675d, 0x233d0,
  0x233e4, 0x233d5, 0x06774, 0x06776, 0x233da, 0x06792, 0x233df, 0x08363,
  0x06810, 0x067b0, 0x067b2, 0x067c3, 0x067c8, 0x067d2, 0x067d9, 0x067db,
  0x067f0, 0x067f7, 0x2344a, 0x23451, 0x2344b, 0x06818, 0x0681f, 0x0682d,
  0x23465, 0x06833, 0x0683b, 0x0683e, 0x06844, 0x06845, 0x06849, 0x0684c,
  0x06855, 0x06857, 0x03b77, 0x0686b, 0x0686e, 0x0687a, 0x0687c, 0x06882,
  0x06890, 0x06896, 0x03b6d, 0x06898, 0x06899, 0x0689a, 0x0689c, 0x068aa,
  0x068ab, 0x068b4, 0x068bb, 0x068fb, 0x234e4, 0x2355a, 0x0fa13, 0x068c3,
  0x068c5, 0x068cc, 0x068cf, 0x068d6, 0x068d9, 0x068e4, 0x068e5, 0x068ec,
  0x068f7, 0x06903, 0x06907, 0x03b87, 0x03b88, 0x23594, 0x0693b, 0x03b8d,
  0x06946, 0x06969, 0x0696c, 0x06972, 0x0697a, 0x0697f, 0x06992, 0x03ba4,
  0x06996, 0x06998, 0x069a6, 0x069b0, 0x069b7, 0x069ba, 0x069bc, 0x069c0,
  0x069d1, 0x069d6, 0x23639, 0x23647, 0x06a30, 0x23638, 0x2363a, 0x069e3,
  0x069ee, 0x069ef, 0x069f3, 0x03bcd, 0x069f4, 0x069fe, 0x06a11, 0x06a1a,
  0x06a1d, 0x2371c, 0x06a32, 0x06a33, 0x06a34, 0x06a3f, 0x06a46, 0x06a49,
  0x06a7a, 0x06a4e, 0x06a52, 0x06a64, 0x2370c, 0x06a7e, 0x06a83, 0x06a8b,
  0x03bf0, 0x06a91, 0x06a9f, 0x06aa1, 0x23764, 0x06aab, 0x06abd, 0x06ac6,
  0x06ad4, 0x06ad0, 0x06adc, 0x06add, 0x237ff, 0x237e7, 0x06aec, 0x06af1,
  0x06af2, 0x06af3, 0x06afd, 0x23824, 0x06b0b, 0x06b0f, 0x06b10, 0x06b11,
  0x2383d, 0x06b17, 0x03c26, 0x06b2f, 0x06b4a, 0x06b58, 0x06b6c, 0x06b75,
  0x06b7a, 0x06b81, 0x04e02, 0x04e04, 0x04e05, 0x04e0c, 0x04e12, 0x04e1f,
  0x04e23, 0x04e24, 0x04e28, 0x04e2b, 0x04e2e, 0x04e2f, 0x04e30, 0x04e35,
  0x04e40, 0x04e41, 0x04e44, 0x04e47, 0x04e51, 0x04e5a, 0x04e5c, 0x04e63,
  0x04e68, 0x04e69, 0x04e74, 0x04e75, 0x04e79, 0x04e7f, 0x04e8d, 0x04e96,
  0x04e97, 0x04e9d, 0x04eaf, 0x04eb9, 0x04ec3, 0x04ed0, 0x04eda, 0x04edb,
  0x04ee0, 0x04ee1, 0x04ee2, 0x04ee8, 0x04eef, 0x04ef1, 0x04ef3, 0x04ef5,
  0x04efd, 0x04efe, 0x04eff, 0x04f00, 0x04f02, 0x04f03, 0x04f08, 0x04f0b,
  0x04f0c, 0x04f12, 0x04f15, 0x04f16, 0x04f17, 0x04f19, 0x04f2e, 0x04f31,
  0x04f60, 0x04f33, 0x04f35, 0x04f37, 0x04f39, 0x04f3b, 0x04f3e, 0x04f40,
  0x04f42, 0x04f48, 0x04f49, 0x04f4b, 0x04f4c, 0x04f52, 0x04f54, 0x04f56,
  0x04f58, 0x04f5f, 0x04f63, 0x04f6a, 0x04f6c, 0x04f6e, 0x04f71, 0x04f77,
  0x04f78, 0x04f79, 0x04f7a, 0x04f7d, 0x04f7e, 0x04f81, 0x04f82, 0x04f84,
  0x04f85, 0x04f89, 0x04f8a, 0x04f8c, 0x04f8e, 0x04f90, 0x04f92, 0x04f93,
  0x04f94, 0x04f97, 0x04f99, 0x04f9a, 0x04f9e, 0x04f9f, 0x04fb2, 0x04fb7,
  0x04fb9, 0x04fbb, 0x04fbc, 0x04fbd, 0x04fbe, 0x04fc0, 0x04fc1, 0x04fc5,
  0x04fc6, 0x04fc8, 0x04fc9, 0x04fcb, 0x04fcc, 0x04fcd, 0x04fcf, 0x04fd2,
  0x04fdc, 0x04fe0, 0x04fe2, 0x04ff0, 0x04ff2, 0x04ffc, 0x04ffd, 0x04fff,
  0x05000, 0x05001, 0x05004, 0x05007, 0x0500a, 0x0500c, 0x0500e, 0x05010,
  0x05013, 0x05017, 0x05018, 0x0501b, 0x0501c, 0x0501d, 0x0501e, 0x05022,
  0x05027, 0x0502e, 0x05030, 0x05032, 0x05033, 0x05035, 0x05040, 0x05041,
  0x05042, 0x05045, 0x05046, 0x0504a, 0x0504c, 0x0504e, 0x05051, 0x05052,
  0x05053, 0x05057, 0x05059, 0x0505f, 0x05060, 0x05062, 0x05063, 0x05066,
  0x05067, 0x0506a, 0x0506d, 0x05070, 0x05071, 0x0503b, 0x05081, 0x05083,
  0x05084, 0x05086, 0x0508a, 0x0508e, 0x0508f, 0x05090, 0x05092, 0x05093,
  0x05094, 0x05096, 0x0509b, 0x0509c, 0x0509e, 0x0509f, 0x050a0, 0x050a1,
  0x050a2, 0x050aa, 0x050af, 0x050b0, 0x050b9, 0x050ba, 0x050bd, 0x050c0,
  0x050c3, 0x050c4, 0x050c7, 0x050cc, 0x050ce, 0x050d0, 0x050d3, 0x050d4,
  0x050d8, 0x050dc, 0x050dd, 0x050df, 0x050e2, 0x050e4, 0x050e6, 0x050e8,
  0x050e9, 0x050ef, 0x050f1, 0x050f6, 0x050fa, 0x050fe, 0x05103, 0x05106,
  0x05107, 0x05108, 0x0510b, 0x0510c, 0x0510d, 0x0510e, 0x050f2, 0x05110,
  0x05117, 0x05119, 0x0511b, 0x0511c, 0x0511d, 0x0511e, 0x05123, 0x05127,
  0x05128, 0x0512c, 0x0512d, 0x0512f, 0x05131, 0x05133, 0x05134, 0x05135,
  0x05138, 0x05139, 0x05142, 0x0514a, 0x0514f, 0x05153, 0x05155, 0x05157,
  0x05158, 0x0515f, 0x05164, 0x05166, 0x0517e, 0x05183, 0x05184, 0x0518b,
  0x0518e, 0x05198, 0x0519d, 0x051a1, 0x051a3, 0x051ad, 0x051b8, 0x051ba,
  0x051bc, 0x051be, 0x051bf, 0x051c2, 0x051c8, 0x051cf, 0x051d1, 0x051d2,
  0x051d3, 0x051d5, 0x051d8, 0x051de, 0x051e2, 0x051e5, 0x051ee, 0x051f2,
  0x051f3, 0x051f4, 0x051f7, 0x05201, 0x05202, 0x05205, 0x05212, 0x05213,
  0x05215, 0x05216, 0x05218, 0x05222, 0x05228, 0x05231, 0x05232, 0x05235,
  0x0523c, 0x05245, 0x05249, 0x05255, 0x05257, 0x05258, 0x0525a, 0x0525c,
  0x0525f, 0x05260, 0x05261, 0x05266, 0x0526e, 0x05277, 0x05278, 0x05279,
  0x05280, 0x05282, 0x05285, 0x0528a, 0x0528c, 0x05293, 0x05295, 0x05296,
  0x05297, 0x05298, 0x0529a, 0x0529c, 0x052a4, 0x052a5, 0x052a6, 0x052a7,
  0x052af, 0x052b0, 0x052b6, 0x052b7, 0x052b8, 0x052ba, 0x052bb, 0x052bd,
  0x052c0, 0x052c4, 0x052c6, 0x052c8, 0x052cc, 0x052cf, 0x052d1, 0x052d4,
  0x052d6, 0x052db, 0x052dc, 0x052e1, 0x052e5, 0x052e8, 0x052e9, 0x052ea,
  0x052ec, 0x052f0, 0x052f1, 0x052f4, 0x052f6, 0x052f7, 0x05300, 0x05303,
  0x0530a, 0x0530b, 0x0530c, 0x05311, 0x05313, 0x05318, 0x0531b, 0x0531c,
  0x0531e, 0x0531f, 0x05325, 0x05327, 0x05328, 0x05329, 0x0532b, 0x0532c,
  0x0532d, 0x05330, 0x05332, 0x05335, 0x0533c, 0x0533d, 0x0533e, 0x05342,
  0x0534c, 0x0534b, 0x05359, 0x0535b, 0x05361, 0x05363, 0x05365, 0x0536c,
  0x0536d, 0x05372, 0x05379, 0x0537e, 0x05383, 0x05387, 0x05388, 0x0538e,
  0x05393, 0x05394, 0x05399, 0x0539d, 0x053a1, 0x053a4, 0x053aa, 0x053ab,
  0x053af, 0x053b2, 0x053b4, 0x053b5, 0x053b7, 0x053b8, 0x053ba, 0x053bd,
  0x053c0, 0x053c5, 0x053cf, 0x053d2, 0x053d3, 0x053d5, 0x053da, 0x053dd,
  0x053de, 0x053e0, 0x053e6, 0x053e7, 0x053f5, 0x05402, 0x05413, 0x0541a,
  0x05421, 0x05427, 0x05428, 0x0542a, 0x0542f, 0x05431, 0x05434, 0x05435,
  0x05443, 0x05444, 0x05447, 0x0544d, 0x0544f, 0x0545e, 0x05462, 0x05464,
  0x05466, 0x05467, 0x05469, 0x0546b, 0x0546d, 0x0546e, 0x05474, 0x0547f,
  0x05481, 0x05483, 0x05485, 0x05488, 0x05489, 0x0548d, 0x05491, 0x05495,
  0x05496, 0x0549c, 0x0549f, 0x054a1, 0x054a6, 0x054a7, 0x054a9, 0x054aa,
  0x054ad, 0x054ae, 0x054b1, 0x054b7, 0x054b9, 0x054ba, 0x054bb, 0x054bf,
  0x054c6, 0x054ca, 0x054cd, 0x054ce, 0x054e0, 0x054ea, 0x054ec, 0x054ef,
  0x054f6, 0x054fc, 0x054fe, 0x054ff, 0x05500, 0x05501, 0x05505, 0x05508,
  0x05509, 0x0550c, 0x0550d, 0x0550e, 0x05515, 0x0552a, 0x0552b, 0x05532,
  0x05535, 0x05536, 0x0553b, 0x0553c, 0x0553d, 0x05541, 0x05547, 0x05549,
  0x0554a, 0x0554d, 0x05550, 0x05551, 0x05558, 0x0555a, 0x0555b, 0x0555e,
  0x05560, 0x05561, 0x05564, 0x05566, 0x0557f, 0x05581, 0x05582, 0x05586,
  0x05588, 0x0558e, 0x0558f, 0x05591, 0x05592, 0x05593, 0x05594, 0x05597,
  0x055a3, 0x055a4, 0x055ad, 0x055b2, 0x055bf, 0x055c1, 0x055c3, 0x055c6,
  0x055c9, 0x055cb, 0x055cc, 0x055ce, 0x055d1, 0x055d2, 0x055d3, 0x055d7,
  0x055d8, 0x055db, 0x055de, 0x055e2, 0x055e9, 0x055f6, 0x055ff, 0x05605,
  0x05608, 0x0560a, 0x0560d, 0x0560e, 0x0560f, 0x05610, 0x05611, 0x05612,
  0x05619, 0x0562c, 0x05630, 0x05633, 0x05635, 0x05637, 0x05639, 0x0563b,
  0x0563c, 0x0563d, 0x0563f, 0x05640, 0x05641, 0x05643, 0x05644, 0x05646,
  0x05649, 0x0564b, 0x0564d, 0x0564f, 0x05654, 0x0565e, 0x05660, 0x05661,
  0x05662, 0x05663, 0x05666, 0x05669, 0x0566d, 0x0566f, 0x05671, 0x05672,
  0x05675, 0x05684, 0x05685, 0x05688, 0x0568b, 0x0568c, 0x05695, 0x05699,
  0x0569a, 0x0569d, 0x0569e, 0x0569f, 0x056a6, 0x056a7, 0x056a8, 0x056a9,
  0x056ab, 0x056ac, 0x056ad, 0x056b1, 0x056b3, 0x056b7, 0x056be, 0x056c5,
  0x056c9, 0x056ca, 0x056cb, 0x056cf, 0x056d0, 0x056cc, 0x056cd, 0x056d9,
  0x056dc, 0x056dd, 0x056df, 0x056e1, 0x056e4, 0x056e5, 0x056e6, 0x056e7,
  0x056e8, 0x056f1, 0x056eb, 0x056ed, 0x056f6, 0x056f7, 0x05701, 0x05702,
  0x05707, 0x0570a, 0x0570c, 0x05711, 0x05715, 0x0571a, 0x0571b, 0x0571d,
  0x05720, 0x05722, 0x05723, 0x05724, 0x05725, 0x05729, 0x0572a, 0x0572c,
  0x0572e, 0x0572f, 0x05733, 0x05734, 0x0573d, 0x0573e, 0x0573f, 0x05745,
  0x05746, 0x0574c, 0x0574d, 0x05752, 0x05762, 0x05765, 0x05767, 0x05768,
  0x0576b, 0x0576d, 0x0576e, 0x0576f, 0x05770, 0x05771, 0x05773, 0x05774,
  0x05775, 0x05777, 0x05779, 0x0577a, 0x0577b, 0x0577c, 0x0577e, 0x05781,
  0x05783, 0x0578c, 0x05794, 0x05797, 0x05799, 0x0579a, 0x0579c, 0x0579d,
  0x0579e, 0x0579f, 0x057a1, 0x05795, 0x057a7, 0x057a8, 0x057a9, 0x057ac,
  0x057b8, 0x057bd, 0x057c7, 0x057c8, 0x057cc, 0x057cf, 0x057d5, 0x057dd,
  0x057de, 0x057e4, 0x057e6, 0x057e7, 0x057e9, 0x057ed, 0x057f0, 0x057f5,
  0x057f6, 0x057f8, 0x057fd, 0x057fe, 0x057ff, 0x05803, 0x05804, 0x05808,
  0x05809, 0x057e1, 0x0580c, 0x0580d, 0x0581b, 0x0581e, 0x0581f, 0x05820,
  0x05826, 0x05827, 0x0582d, 0x05832, 0x05839, 0x0583f, 0x05849, 0x0584c,
  0x0584d, 0x0584f, 0x05850, 0x05855, 0x0585f, 0x05861, 0x05864, 0x05867,
  0x05868, 0x05878, 0x0587c, 0x0587f, 0x05880, 0x05881, 0x05887, 0x05888,
  0x05889, 0x0588a, 0x0588c, 0x0588d, 0x0588f, 0x05890, 0x05894, 0x05896,
  0x0589d, 0x058a0, 0x058a1, 0x058a2, 0x058a6, 0x058a9, 0x058b1, 0x058b2,
  0x058c4, 0x058bc, 0x058c2, 0x058c8, 0x058cd, 0x058ce, 0x058d0, 0x058d2,
  0x058d4, 0x058d6, 0x058da, 0x058dd, 0x058e1, 0x058e2, 0x058e9, 0x058f3,
  0x05905, 0x05906, 0x0590b, 0x0590c, 0x05912, 0x05913, 0x05914, 0x08641,
  0x0591d, 0x05921, 0x05923, 0x05924, 0x05928, 0x0592f, 0x05930, 0x05933,
  0x05935, 0x05936, 0x0593f, 0x05943, 0x05946, 0x05952, 0x05953, 0x05959,
  0x0595b, 0x0595d, 0x0595e, 0x0595f, 0x05961, 0x05963, 0x0596b, 0x0596d,
  0x0596f, 0x05972, 0x05975, 0x05976, 0x05979, 0x0597b, 0x0597c, 0x0598b,
  0x0598c, 0x0598e, 0x05992, 0x05995, 0x05997, 0x0599f, 0x059a4, 0x059a7,
  0x059ad, 0x059ae, 0x059af, 0x059b0, 0x059b3, 0x059b7, 0x059ba, 0x059bc,
  0x059c1, 0x059c3, 0x059c4, 0x059c8, 0x059ca, 0x059cd, 0x059d2, 0x059dd,
  0x059de, 0x059df, 0x059e3, 0x059e4, 0x059e7, 0x059ee, 0x059ef, 0x059f1,
  0x059f2, 0x059f4, 0x059f7, 0x05a00, 0x05a04, 0x05a0c, 0x05a0d, 0x05a0e,
  0x05a12, 0x05a13, 0x05a1e, 0x05a23, 0x05a24, 0x05a27, 0x05a28, 0x05a2a,
  0x05a2d, 0x05a30, 0x05a44, 0x05a45, 0x05a47, 0x05a48, 0x05a4c, 0x05a50,
  0x05a55, 0x05a5e, 0x05a63, 0x05a65, 0x05a67, 0x05a6d, 0x05a77, 0x05a7a,
  0x05a7b, 0x05a7e, 0x05a8b, 0x05a90, 0x05a93, 0x05a96, 0x05a99, 0x05a9c,
  0x05a9e, 0x05a9f, 0x05aa0, 0x05aa2, 0x05aa7, 0x05aac, 0x05ab1, 0x05ab2,
  0x05ab3, 0x05ab5, 0x05ab8, 0x05aba, 0x05abb, 0x05abf, 0x05ac4, 0x05ac6,
  0x05ac8, 0x05acf, 0x05ada, 0x05adc, 0x05ae0, 0x05ae5, 0x05aea, 0x05aee,
  0x05af5, 0x05af6, 0x05afd, 0x05b00, 0x05b01, 0x05b08, 0x05b17, 0x05b34,
  0x05b19, 0x05b1b, 0x05b1d, 0x05b21, 0x05b25, 0x05b2d, 0x05b38, 0x05b41,
  0x05b4b, 0x05b4c, 0x05b52, 0x05b56, 0x05b5e, 0x05b68, 0x05b6e, 0x05b6f,
  0x05b7c, 0x05b7d, 0x05b7e, 0x05b7f, 0x05b81, 0x05b84, 0x05b86, 0x05b8a,
  0x05b8e, 0x05b90, 0x05b91, 0x05b93, 0x05b94, 0x05b96, 0x05ba8, 0x05ba9,
  0x05bac, 0x05bad, 0x05baf, 0x05bb1, 0x05bb2, 0x05bb7, 0x05bba, 0x05bbc,
  0x05bc0, 0x05bc1, 0x05bcd, 0x05bcf, 0x05bd6, 0x05bd7, 0x05bd8, 0x05bd9,
  0x05bda, 0x05be0, 0x05bef, 0x05bf1, 0x05bf4, 0x05bfd, 0x05c0c, 0x05c17,
  0x05c1e, 0x05c1f, 0x05c23, 0x05c26, 0x05c29, 0x05c2b, 0x05c2c, 0x05c2e,
  0x05c30, 0x05c32, 0x05c35, 0x05c36, 0x05c59, 0x05c5a, 0x05c5c, 0x05c62,
  0x05c63, 0x05c67, 0x05c68, 0x05c69, 0x05c6d, 0x05c70, 0x05c74, 0x05c75,
  0x05c7a, 0x05c7b, 0x05c7c, 0x05c7d, 0x05c87, 0x05c88, 0x05c8a, 0x05c8f,
  0x05c92, 0x05c9d, 0x05c9f, 0x05ca0, 0x05ca2, 0x05ca3, 0x05ca6, 0x05caa,
  0x05cb2, 0x05cb4, 0x05cb5, 0x05cba, 0x05cc9, 0x05ccb, 0x05cd2, 0x05cdd,
  0x05cd7, 0x05cee, 0x05cf1, 0x05cf2, 0x05cf4, 0x05d01, 0x05d06, 0x05d0d,
  0x05d12, 0x05d2b, 0x05d23, 0x05d24, 0x05d26, 0x05d27, 0x05d31, 0x05d34,
  0x05d39, 0x05d3d, 0x05d3f, 0x05d42, 0x05d43, 0x05d46, 0x05d48, 0x05d55,
  0x05d51, 0x05d59, 0x05d4a, 0x05d5f, 0x05d60, 0x05d61, 0x05d62, 0x05d64,
  0x05d6a, 0x05d6d, 0x05d70, 0x05d79, 0x05d7a, 0x05d7e, 0x05d7f, 0x05d81,
  0x05d83, 0x05d88, 0x05d8a, 0x05d92, 0x05d93, 0x05d94, 0x05d95, 0x05d99,
  0x05d9b, 0x05d9f, 0x05da0, 0x05da7, 0x05dab, 0x05db0, 0x05db4, 0x05db8,
  0x05db9, 0x05dc3, 0x05dc7, 0x05dcb, 0x05dd0, 0x05dce, 0x05dd8, 0x05dd9,
  0x05de0, 0x05de4, 0x05de9, 0x05df8, 0x05df9, 0x05e00, 0x05e07, 0x05e0d,
  0x05e12, 0x05e14, 0x05e15, 0x05e18, 0x05e1f, 0x05e20, 0x05e2e, 0x05e28,
  0x05e32, 0x05e35, 0x05e3e, 0x05e4b, 0x05e50, 0x05e49, 0x05e51, 0x05e56,
  0x05e58, 0x05e5b, 0x05e5c, 0x05e5e, 0x05e68, 0x05e6a, 0x05e6b, 0x05e6c,
  0x05e6d, 0x05e6e, 0x05e70, 0x05e80, 0x05e8b, 0x05e8e, 0x05ea2, 0x05ea4,
  0x05ea5, 0x05ea8, 0x05eaa, 0x05eac, 0x05eb1, 0x05eb3, 0x05ebd, 0x05ebe,
  0x05ebf, 0x05ec6, 0x05ecc, 0x05ecb, 0x05ece, 0x05ed1, 0x05ed2, 0x05ed4,
  0x05ed5, 0x05edc, 0x05ede, 0x05ee5, 0x05eeb, 0x05f02, 0x05f06, 0x05f07,
  0x05f08, 0x05f0e, 0x05f19, 0x05f1c, 0x05f1d, 0x05f21, 0x05f22, 0x05f23,
  0x05f24, 0x05f28, 0x05f2b, 0x05f2c, 0x05f2e, 0x05f30, 0x05f34, 0x05f36,
  0x05f3b, 0x05f3d, 0x05f3f, 0x05f40, 0x05f44, 0x05f45, 0x05f47, 0x05f4d,
  0x05f50, 0x05f54, 0x05f58, 0x05f5b, 0x05f60, 0x05f63, 0x05f64, 0x05f67,
  0x05f6f, 0x05f72, 0x05f74, 0x05f75, 0x05f78, 0x05f7a, 0x05f7d, 0x05f7e,
  0x05f89, 0x05f8d, 0x05f8f, 0x05f96, 0x05f9c, 0x05f9d, 0x05fa2, 0x05fa7,
  0x05fab, 0x05fa4, 0x05fac, 0x05faf, 0x05fb0, 0x05fb1, 0x05fb8, 0x05fc4,
  0x05fc7, 0x05fc8, 0x05fc9, 0x05fcb, 0x05fd0, 0x05fd1, 0x05fd2, 0x05fd3,
  0x05fd4, 0x05fde, 0x05fe1, 0x05fe2, 0x05fe8, 0x05fe9, 0x05fea, 0x05fec,
  0x05fed, 0x05fee, 0x05fef, 0x05ff2, 0x05ff3, 0x05ff6, 0x05ffa, 0x05ffc,
  0x06007, 0x0600a, 0x0600d, 0x06013, 0x06014, 0x06017, 0x06018, 0x0601a,
  0x0601f, 0x06024, 0x0602d, 0x06033, 0x06035, 0x06040, 0x06047, 0x06048,
  0x06049, 0x0604c, 0x06051, 0x06054, 0x06056, 0x06057, 0x0605d, 0x06061,
  0x06067, 0x06071, 0x0607e, 0x0607f, 0x06082, 0x06086, 0x06088, 0x0608a,
  0x0608e, 0x06091, 0x06093, 0x06095, 0x06098, 0x0609d, 0x0609e, 0x060a2,
  0x060a4, 0x060a5, 0x060a8, 0x060b0, 0x060b1, 0x060b7, 0x060bb, 0x060be,
  0x060c2, 0x060c4, 0x060c8, 0x060c9, 0x060ca, 0x060cb, 0x060ce, 0x060cf,
  0x060d4, 0x060d5, 0x060d9, 0x060db, 0x060dd, 0x060de, 0x060e2, 0x060e5,
  0x060f2, 0x060f5, 0x060f8, 0x060fc, 0x060fd, 0x06102, 0x06107, 0x0610a,
  0x0610c, 0x06110, 0x06111, 0x06112, 0x06113, 0x06114, 0x06116, 0x06117,
  0x06119, 0x0611c, 0x0611e, 0x06122, 0x0612a, 0x0612b, 0x06130, 0x06131,
  0x06135, 0x06136, 0x06137, 0x06139, 0x06141, 0x06145, 0x06146, 0x06149,
  0x0615e, 0x06160, 0x0616c, 0x06172, 0x06178, 0x0617b, 0x0617c, 0x0617f,
  0x06180, 0x06181, 0x06183, 0x06184, 0x0618b, 0x0618d, 0x06192, 0x06193,
  0x06197, 0x06198, 0x0619c, 0x0619d, 0x0619f, 0x061a0, 0x061a5, 0x061a8,
  0x061aa, 0x061ad, 0x061b8, 0x061b9, 0x061bc, 0x061c0, 0x061c1, 0x061c2,
  0x061ce, 0x061cf, 0x061d5, 0x061dc, 0x061dd, 0x061de, 0x061df, 0x061e1,
  0x061e2, 0x061e7, 0x061e9, 0x061e5, 0x061ec, 0x061ed, 0x061ef, 0x06201,
  0x06203, 0x06204, 0x06207, 0x06213, 0x06215, 0x0621c, 0x06220, 0x06222,
  0x06223, 0x06227, 0x06229, 0x0622b, 0x06239, 0x0623d, 0x06242, 0x06243,
  0x06244, 0x06246, 0x0624c, 0x06250, 0x06251, 0x06252, 0x06254, 0x06256,
  0x0625a, 0x0625c, 0x06264, 0x0626d, 0x0626f, 0x06273, 0x0627a, 0x0627d,
  0x0628d, 0x0628e, 0x0628f, 0x06290, 0x062a6, 0x062a8, 0x062b3, 0x062b6,
  0x062b7, 0x062ba, 0x062be, 0x062bf, 0x062c4, 0x062ce, 0x062d5, 0x062d6,
  0x062da, 0x062ea, 0x062f2, 0x062f4, 0x062fc, 0x062fd, 0x06303, 0x06304,
  0x0630a, 0x0630b, 0x0630d, 0x06310, 0x06313, 0x06316, 0x06318, 0x06329,
  0x0632a, 0x0632d, 0x06335, 0x06336, 0x06339, 0x0633c, 0x06341, 0x06342,
  0x06343, 0x06344, 0x06346, 0x0634a, 0x0634b, 0x0634e, 0x06352, 0x06353,
  0x06354, 0x06358, 0x0635b, 0x06365, 0x06366, 0x0636c, 0x0636d, 0x06371,
  0x06374, 0x06375, 0x06378, 0x0637c, 0x0637d, 0x0637f, 0x06382, 0x06384,
  0x06387, 0x0638a, 0x06390, 0x06394, 0x06395, 0x06399, 0x0639a, 0x0639e,
  0x063a4, 0x063a6, 0x063ad, 0x063ae, 0x063af, 0x063bd, 0x063c1, 0x063c5,
  0x063c8, 0x063ce, 0x063d1, 0x063d3, 0x063d4, 0x063d5, 0x063dc, 0x063e0,
  0x063e5, 0x063ea, 0x063ec, 0x063f2, 0x063f3, 0x063f5, 0x063f8, 0x063f9,
  0x06409, 0x0640a, 0x06410, 0x06412, 0x06414, 0x06418, 0x0641e, 0x06420,
  0x06422, 0x06424, 0x06425, 0x06429, 0x0642a, 0x0642f, 0x06430, 0x06435,
  0x0643d, 0x0643f, 0x0644b, 0x0644f, 0x06451, 0x06452, 0x06453, 0x06454,
  0x0645a, 0x0645b, 0x0645c, 0x0645d, 0x0645f, 0x06460, 0x06461, 0x06463,
  0x0646d, 0x06473, 0x06474, 0x0647b, 0x0647d, 0x06485, 0x06487, 0x0648f,
  0x06490, 0x06491, 0x06498, 0x06499, 0x0649b, 0x0649d, 0x0649f, 0x064a1,
  0x064a3, 0x064a6, 0x064a8, 0x064ac, 0x064b3, 0x064bd, 0x064be, 0x064bf,
  0x064c4, 0x064c9, 0x064ca, 0x064cb, 0x064cc, 0x064ce, 0x064d0, 0x064d1,
  0x064d5, 0x064d7, 0x064e4, 0x064e5, 0x064e9, 0x064ea, 0x064ed, 0x064f0,
  0x064f5, 0x064f7, 0x064fb, 0x064ff, 0x06501, 0x06504, 0x06508, 0x06509,
  0x0650a, 0x0650f, 0x06513, 0x06514, 0x06516, 0x06519, 0x0651b, 0x0651e,
  0x0651f, 0x06522, 0x06526, 0x06529, 0x0652e, 0x06531, 0x0653a, 0x0653c,
  0x0653d, 0x06543, 0x06547, 0x06549, 0x06550, 0x06552, 0x06554, 0x0655f,
  0x06560, 0x06567, 0x0656b, 0x0657a, 0x0657d, 0x06581, 0x06585, 0x0658a,
  0x06592, 0x06595, 0x06598, 0x0659d, 0x065a0, 0x065a3, 0x065a6, 0x065ae,
  0x065b2, 0x065b3, 0x065b4, 0x065bf, 0x065c2, 0x065c8, 0x065c9, 0x065ce,
  0x065d0, 0x065d4, 0x065d6, 0x065d8, 0x065df, 0x065f0, 0x065f2, 0x065f4,
  0x065f5, 0x065f9, 0x065fe, 0x065ff, 0x06600, 0x06604, 0x06608, 0x06609,
  0x0660d, 0x06611, 0x06612, 0x06615, 0x06616, 0x0661d, 0x0661e, 0x06621,
  0x06622, 0x06623, 0x06624, 0x06626, 0x06629, 0x0662a, 0x0662b, 0x0662c,
  0x0662e, 0x06630, 0x06631, 0x06633, 0x06639, 0x06637, 0x06640, 0x06645,
  0x06646, 0x0664a, 0x0664c, 0x06651, 0x0664e, 0x06657, 0x06658, 0x06659,
  0x0665b, 0x0665c, 0x06660, 0x06661, 0x066fb, 0x0666a, 0x0666b, 0x0666c,
  0x0667e, 0x06673, 0x06675, 0x0667f, 0x06677, 0x06678, 0x06679, 0x0667b,
  0x06680, 0x0667c, 0x0668b, 0x0668c, 0x0668d, 0x06690, 0x06692, 0x06699,
  0x0669a, 0x0669b, 0x0669c, 0x0669f, 0x066a0, 0x066a4, 0x066ad, 0x066b1,
  0x066b2, 0x066b5, 0x066bb, 0x066bf, 0x066c0, 0x066c2, 0x066c3, 0x066c8,
  0x066cc, 0x066ce, 0x066cf, 0x066d4, 0x066db, 0x066df, 0x066e8, 0x066eb,
  0x066ec, 0x066ee, 0x066fa, 0x06705, 0x06707, 0x0670e, 0x06713, 0x06719,
  0x0671c, 0x06720, 0x06722, 0x06733, 0x0673e, 0x06745, 0x06747, 0x06748,
  0x0674c, 0x06754, 0x06755, 0x0675d, 0x06766, 0x0676c, 0x0676e, 0x06774,
  0x06776, 0x0677b, 0x06781, 0x06784, 0x0678e, 0x0678f, 0x06791, 0x06793,
  0x06796, 0x06798, 0x06799, 0x0679b, 0x067b0, 0x067b1, 0x067b2, 0x067b5,
  0x067bb, 0x067bc, 0x067bd, 0x067f9, 0x067c0, 0x067c2, 0x067c3, 0x067c5,
  0x067c8, 0x067c9, 0x067d2, 0x067d7, 0x067d9, 0x067dc, 0x067e1, 0x067e6,
  0x067f0, 0x067f2, 0x067f6, 0x067f7, 0x06852, 0x06814, 0x06819, 0x0681d,
  0x0681f, 0x06828, 0x06827, 0x0682c, 0x0682d, 0x0682f, 0x06830, 0x06831,
  0x06833, 0x0683b, 0x0683f, 0x06844, 0x06845, 0x0684a, 0x0684c, 0x06855,
  0x06857, 0x06858, 0x0685b, 0x0686b, 0x0686e, 0x0686f, 0x06870, 0x06871,
  0x06872, 0x06875, 0x06879, 0x0687a, 0x0687b, 0x0687c, 0x06882, 0x06884,
  0x06886, 0x06888, 0x06896, 0x06898, 0x0689a, 0x0689c, 0x068a1, 0x068a3,
  0x068a5, 0x068a9, 0x068aa, 0x068ae, 0x068b2, 0x068bb, 0x068c5, 0x068c8,
  0x068cc, 0x068cf, 0x068d0, 0x068d1, 0x068d3, 0x068d6, 0x068d9, 0x068dc,
  0x068dd, 0x068e5, 0x068e8, 0x068ea, 0x068eb, 0x068ec, 0x068ed, 0x068f0,
  0x068f1, 0x068f5, 0x068f6, 0x068fb, 0x068fc, 0x068fd, 0x06906, 0x06909,
  0x0690a, 0x06910, 0x06911, 0x06913, 0x06916, 0x06917, 0x06931, 0x06933,
  0x06935, 0x06938, 0x0693b, 0x06942, 0x06945, 0x06949, 0x0694e, 0x06957,
  0x0695b, 0x06963, 0x06964, 0x06965, 0x06966, 0x06968, 0x06969, 0x0696c,
  0x06970, 0x06971, 0x06972, 0x0697a, 0x0697b, 0x0697f, 0x06980, 0x0698d,
  0x06992, 0x06996, 0x06998, 0x069a1, 0x069a5, 0x069a6, 0x069a8, 0x069ab,
  0x069ad, 0x069af, 0x069b7, 0x069b8, 0x069ba, 0x069bc, 0x069c5, 0x069c8,
  0x069d1, 0x069d6, 0x069d7, 0x069e2, 0x069e5, 0x069ee, 0x069ef, 0x069f1,
  0x069f3, 0x069f5, 0x069fe, 0x06a00, 0x06a01, 0x06a03, 0x06a0f, 0x06a11,
  0x06a15, 0x06a1a, 0x06a1d, 0x06a20, 0x06a24, 0x06a28, 0x06a30, 0x06a32,
  0x06a34, 0x06a37, 0x06a3b, 0x06a3e, 0x06a3f, 0x06a45, 0x06a46, 0x06a49,
  0x06a4a, 0x06a4e, 0x06a50, 0x06a51, 0x06a52, 0x06a55, 0x06a56, 0x06a5b,
  0x06a64, 0x06a67, 0x06a6a, 0x06a71, 0x06a73, 0x06a7e, 0x06a81, 0x06a83,
  0x06a86, 0x06a87, 0x06a89, 0x06a8b, 0x06a91, 0x06a9b, 0x06a9d, 0x06a9e,
  0x06a9f, 0x06aa5, 0x06aab, 0x06aaf, 0x06ab0, 0x06ab1, 0x06ab4, 0x06abd,
  0x06abe, 0x06abf, 0x06ac6, 0x06ac9, 0x06ac8, 0x06acc, 0x06ad0, 0x06ad4,
  0x06ad5, 0x06ad6, 0x06adc, 0x06add, 0x06ae4, 0x06ae7, 0x06aec, 0x06af0,
  0x06af1, 0x06af2, 0x06afc, 0x06afd, 0x06b02, 0x06b03, 0x06b06, 0x06b07,
  0x06b09, 0x06b0f, 0x06b10, 0x06b11, 0x06b17, 0x06b1b, 0x06b1e, 0x06b24,
  0x06b28, 0x06b2b, 0x06b2c, 0x06b2f, 0x06b35, 0x06b36, 0x06b3b, 0x06b3f,
  0x06b46, 0x06b4a, 0x06b4d, 0x06b52, 0x06b56, 0x06b58, 0x06b5d, 0x06b60,
  0x06b67, 0x06b6b, 0x06b6e, 0x06b70, 0x06b75, 0x06b7d, 0x06b7e, 0x06b82,
  0x06b85, 0x06b97, 0x06b9b, 0x06b9f, 0x06ba0, 0x06ba2, 0x06ba3, 0x06ba8,
  0x06ba9, 0x06bac, 0x06bad, 0x06bae, 0x06bb0, 0x06bb8, 0x06bb9, 0x06bbd,
  0x06bbe, 0x06bc3, 0x06bc4, 0x06bc9, 0x06bcc, 0x06bd6, 0x06bda, 0x06be1,
  0x06be3, 0x06be6, 0x06be7, 0x06bee, 0x06bf1, 0x06bf7, 0x06bf9, 0x06bff,
  0x06c02, 0x06c04, 0x06c05, 0x06c09, 0x06c0d, 0x06c0e, 0x06c10, 0x06c12,
  0x06c19, 0x06c1f, 0x06c26, 0x06c27, 0x06c28, 0x06c2c, 0x06c2e, 0x06c33,
  0x06c35, 0x06c36, 0x06c3a, 0x06c3b, 0x06c3f, 0x06c4a, 0x06c4b, 0x06c4d,
  0x06c4f, 0x06c52, 0x06c54, 0x06c59, 0x06c5b, 0x06c5c, 0x06c6b, 0x06c6d,
  0x06c6f, 0x06c74, 0x06c76, 0x06c78, 0x06c79, 0x06c7b, 0x06c85, 0x06c86,
  0x06c87, 0x06c89, 0x06c94, 0x06c95, 0x06c97, 0x06c98, 0x06c9c, 0x06c9f,
  0x06cb0, 0x06cb2, 0x06cb4, 0x06cc2, 0x06cc6, 0x06ccd, 0x06ccf, 0x06cd0,
  0x06cd1, 0x06cd2, 0x06cd4, 0x06cd6, 0x06cda, 0x06cdc, 0x06ce0, 0x06ce7,
  0x06ce9, 0x06ceb, 0x06cec, 0x06cee, 0x06cf2, 0x06cf4, 0x06d04, 0x06d07,
  0x06d0a, 0x06d0e, 0x06d0f, 0x06d11, 0x06d13, 0x06d1a, 0x06d26, 0x06d27,
  0x06d28, 0x06c67, 0x06d2e, 0x06d2f, 0x06d31, 0x06d39, 0x06d3c, 0x06d3f,
  0x06d57, 0x06d5e, 0x06d5f, 0x06d61, 0x06d65, 0x06d67, 0x06d6f, 0x06d70,
  0x06d7c, 0x06d82, 0x06d87, 0x06d91, 0x06d92, 0x06d94, 0x06d96, 0x06d97,
  0x06d98, 0x06daa, 0x06dac, 0x06db4, 0x06db7, 0x06db9, 0x06dbd, 0x06dbf,
  0x06dc4, 0x06dc8, 0x06dca, 0x06dce, 0x06dcf, 0x06dd6, 0x06ddb, 0x06ddd,
  0x06ddf, 0x06de0, 0x06de2, 0x06de5, 0x06de9, 0x06def, 0x06df0, 0x06df4,
  0x06df6, 0x06dfc, 0x06e00, 0x06e04, 0x06e1e, 0x06e22, 0x06e27, 0x06e32,
  0x06e36, 0x06e39, 0x06e3b, 0x06e3c, 0x06e44, 0x06e45, 0x06e48, 0x06e49,
  0x06e4b, 0x06e4f, 0x06e51, 0x06e52, 0x06e53, 0x06e54, 0x06e57, 0x06e5c,
  0x06e5d, 0x06e5e, 0x06e62, 0x06e63, 0x06e68, 0x06e73, 0x06e7b, 0x06e7d,
  0x06e8d, 0x06e93, 0x06e99, 0x06ea0, 0x06ea7, 0x06ead, 0x06eae, 0x06eb1,
  0x06eb3, 0x06ebb, 0x06ebf, 0x06ec0, 0x06ec1, 0x06ec3, 0x06ec7, 0x06ec8,
  0x06eca, 0x06ecd, 0x06ece, 0x06ecf, 0x06eeb, 0x06eed, 0x06eee, 0x06ef9,
  0x06efb, 0x06efd, 0x06f04, 0x06f08, 0x06f0a, 0x06f0c, 0x06f0d, 0x06f16,
  0x06f18, 0x06f1a, 0x06f1b, 0x06f26, 0x06f29, 0x06f2a, 0x06f2f, 0x06f30,
  0x06f33, 0x06f36, 0x06f3b, 0x06f3c, 0x06f2d, 0x06f4f, 0x06f51, 0x06f52,
  0x06f53, 0x06f57, 0x06f59, 0x06f5a, 0x06f5d, 0x06f5e, 0x06f61, 0x06f62,
  0x06f68, 0x06f6c, 0x06f7d, 0x06f7e, 0x06f83, 0x06f87, 0x06f88, 0x06f8b,
  0x06f8c, 0x06f8d, 0x06f90, 0x06f92, 0x06f93, 0x06f94, 0x06f96, 0x06f9a,
  0x06f9f, 0x06fa0, 0x06fa5, 0x06fa6, 0x06fa7, 0x06fa8, 0x06fae, 0x06faf,
  0x06fb0, 0x06fb5, 0x06fb6, 0x06fbc, 0x06fc5, 0x06fc7, 0x06fc8, 0x06fca,
  0x06fda, 0x06fde, 0x06fe8, 0x06fe9, 0x06ff0, 0x06ff5, 0x06ff9, 0x06ffc,
  0x06ffd, 0x07000, 0x07005, 0x07006, 0x07007, 0x0700d, 0x07017, 0x07020,
  0x07023, 0x0702f, 0x07034, 0x07037, 0x07039, 0x0703c, 0x07043, 0x07044,
  0x07048, 0x07049, 0x0704a, 0x0704b, 0x07054, 0x07055, 0x0705d, 0x0705e,
  0x0704e, 0x07064, 0x07065, 0x0706c, 0x0706e, 0x07075, 0x07076, 0x0707e,
  0x07081, 0x07085, 0x07086, 0x07094, 0x07095, 0x07096, 0x07097, 0x07098,
  0x0709b, 0x070a4, 0x070ab, 0x070b0, 0x070b1, 0x070b4, 0x070b7, 0x070ca,
  0x070d1, 0x070d3, 0x070d4, 0x070d5, 0x070d6, 0x070d8, 0x070dc, 0x070e4,
  0x070fa, 0x07103, 0x07104, 0x07105, 0x07106, 0x07107, 0x0710b, 0x0710c,
  0x0710f, 0x0711e, 0x07120, 0x0712b, 0x0712d, 0x0712f, 0x07130, 0x07131,
  0x07138, 0x07141, 0x07145, 0x07146, 0x07147, 0x0714a, 0x0714b, 0x07150,
  0x07152, 0x07157, 0x0715a, 0x0715c, 0x0715e, 0x07160, 0x07168, 0x07179,
  0x07180, 0x07185, 0x07187, 0x0718c, 0x07192, 0x0719a, 0x0719b, 0x071a0,
  0x071a2, 0x071af, 0x071b0, 0x071b2, 0x071b3, 0x071ba, 0x071bf, 0x071c0,
  0x071c1, 0x071c4, 0x071cb, 0x071cc, 0x071d3, 0x071d6, 0x071d9, 0x071da,
  0x071dc, 0x071f8, 0x071fe, 0x07200, 0x07207, 0x07208, 0x07209, 0x07213,
  0x07217, 0x0721a, 0x0721d, 0x0721f, 0x07224, 0x0722b, 0x0722f, 0x07234,
  0x07238, 0x07239, 0x07241, 0x07242, 0x07243, 0x07245, 0x0724e, 0x0724f,
  0x07250, 0x07253, 0x07255, 0x07256, 0x0725a, 0x0725c, 0x0725e, 0x07260,
  0x07263, 0x07268, 0x0726b, 0x0726e, 0x0726f, 0x07271, 0x07277, 0x07278,
  0x0727b, 0x0727c, 0x0727f, 0x07284, 0x07289, 0x0728d, 0x0728e, 0x07293,
  0x0729b, 0x072a8, 0x072ad, 0x072ae, 0x072b1, 0x072b4, 0x072be, 0x072c1,
  0x072c7, 0x072c9, 0x072cc, 0x072d5, 0x072d6, 0x072d8, 0x072df, 0x072e5,
  0x072f3, 0x072f4, 0x072fa, 0x072fb, 0x072fe, 0x07302, 0x07304, 0x07305,
  0x07307, 0x0730b, 0x0730d, 0x07312, 0x07313, 0x07318, 0x07319, 0x0731e,
  0x07322, 0x07324, 0x07327, 0x07328, 0x0732c, 0x07331, 0x07332, 0x07335,
  0x0733a, 0x0733b, 0x0733d, 0x07343, 0x0734d, 0x07350, 0x07352, 0x07356,
  0x07358, 0x0735d, 0x0735e, 0x0735f, 0x07360, 0x07366, 0x07367, 0x07369,
  0x0736b, 0x0736c, 0x0736e, 0x0736f, 0x07371, 0x07377, 0x07379, 0x0737c,
  0x07380, 0x07381, 0x07383, 0x07385, 0x07386, 0x0738e, 0x07390, 0x07393,
  0x07395, 0x07397, 0x07398, 0x0739c, 0x0739e, 0x0739f, 0x073a0, 0x073a2,
  0x073a5, 0x073a6, 0x073aa, 0x073ab, 0x073ad, 0x073b5, 0x073b7, 0x073b9,
  0x073bc, 0x073bd, 0x073bf, 0x073c5, 0x073c6, 0x073c9, 0x073cb, 0x073cc,
  0x073cf, 0x073d2, 0x073d3, 0x073d6, 0x073d9, 0x073dd, 0x073e1, 0x073e3,
  0x073e6, 0x073e7, 0x073e9, 0x073f4, 0x073f5, 0x073f7, 0x073f9, 0x073fa,
  0x073fb, 0x073fd, 0x073ff, 0x07400, 0x07401, 0x07404, 0x07407, 0x0740a,
  0x07411, 0x0741a, 0x0741b, 0x07424, 0x07426, 0x07428, 0x07429, 0x0742a,
  0x0742b, 0x0742c, 0x0742d, 0x0742e, 0x0742f, 0x07430, 0x07431, 0x07439,
  0x07440, 0x07443, 0x07444, 0x07446, 0x07447, 0x0744b, 0x0744d, 0x07451,
  0x07452, 0x07457, 0x0745d, 0x07462, 0x07466, 0x07467, 0x07468, 0x0746b,
  0x0746d, 0x0746e, 0x07471, 0x07472, 0x07480, 0x07481, 0x07485, 0x07486,
  0x07487, 0x07489, 0x0748f, 0x07490, 0x07491, 0x07492, 0x07498, 0x07499,
  0x0749a, 0x0749c, 0x0749f, 0x074a0, 0x074a1, 0x074a3, 0x074a6, 0x074a8,
  0x074a9, 0x074aa, 0x074ab, 0x074ae, 0x074af, 0x074b1, 0x074b2, 0x074b5,
  0x074b9, 0x074bb, 0x074bf, 0x074c8, 0x074c9, 0x074cc, 0x074d0, 0x074d3,
  0x074d8, 0x074da, 0x074db, 0x074de, 0x074df, 0x074e4, 0x074e8, 0x074ea,
  0x074eb, 0x074ef, 0x074f4, 0x074fa, 0x074fb, 0x074fc, 0x074ff, 0x07506,
  0x07512, 0x07516, 0x07517, 0x07520, 0x07521, 0x07524, 0x07527, 0x07529,
  0x0752a, 0x0752f, 0x07536, 0x07539, 0x0753d, 0x0753e, 0x0753f, 0x07540,
  0x07543, 0x07547, 0x07548, 0x0754e, 0x07550, 0x07552, 0x07557, 0x0755e,
  0x0755f, 0x07561, 0x0756f, 0x07571, 0x07579, 0x0757a, 0x0757b, 0x0757c,
  0x0757d, 0x0757e, 0x07581, 0x07585, 0x07590, 0x07592, 0x07593, 0x07595,
  0x07599, 0x0759c, 0x075a2, 0x075a4, 0x075b4, 0x075ba, 0x075bf, 0x075c0,
  0x075c1, 0x075c4, 0x075c6, 0x075cc, 0x075ce, 0x075cf, 0x075d7, 0x075dc,
  0x075df, 0x075e0, 0x075e1, 0x075e4, 0x075e7, 0x075ec, 0x075ee, 0x075ef,
  0x075f1, 0x075f9, 0x07600, 0x07602, 0x07603, 0x07604, 0x07607, 0x07608,
  0x0760a, 0x0760c, 0x0760f, 0x07612, 0x07613, 0x07615, 0x07616, 0x07619,
  0x0761b, 0x0761c, 0x0761d, 0x0761e, 0x07623, 0x07625, 0x07626, 0x07629,
  0x0762d, 0x07632, 0x07633, 0x07635, 0x07638, 0x07639, 0x0763a, 0x0763c,
  0x0764a, 0x07640, 0x07641, 0x07643, 0x07644, 0x07645, 0x07649, 0x0764b,
  0x07655, 0x07659, 0x0765f, 0x07664, 0x07665, 0x0766d, 0x0766e, 0x0766f,
  0x07671, 0x07674, 0x07681, 0x07685, 0x0768c, 0x0768d, 0x07695, 0x0769b,
  0x0769c, 0x0769d, 0x0769f, 0x076a0, 0x076a2, 0x076a3, 0x076a4, 0x076a5,
  0x076a6, 0x076a7, 0x076a8, 0x076aa, 0x076ad, 0x076bd, 0x076c1, 0x076c5,
  0x076c9, 0x076cb, 0x076cc, 0x076ce, 0x076d4, 0x076d9, 0x076e0, 0x076e6,
  0x076e8, 0x076ec, 0x076f0, 0x076f1, 0x076f6, 0x076f9, 0x076fc, 0x07700,
  0x07706, 0x0770a, 0x0770e, 0x07712, 0x07714, 0x07715, 0x07717, 0x07719,
  0x0771a, 0x0771c, 0x07722, 0x07728, 0x0772d, 0x0772e, 0x0772f, 0x07734,
  0x07735, 0x07736, 0x07739, 0x0773d, 0x0773e, 0x07742, 0x07745, 0x07746,
  0x0774a, 0x0774d, 0x0774e, 0x0774f, 0x07752, 0x07756, 0x07757, 0x0775c,
  0x0775e, 0x0775f, 0x07760, 0x07762, 0x07764, 0x07767, 0x0776a, 0x0776c,
  0x07770, 0x07772, 0x07773, 0x07774, 0x0777a, 0x0777d, 0x07780, 0x07784,
  0x0778c, 0x0778d, 0x07794, 0x07795, 0x07796, 0x0779a, 0x0779f, 0x077a2,
  0x077a7, 0x077aa, 0x077ae, 0x077af, 0x077b1, 0x077b5, 0x077be, 0x077c3,
  0x077c9, 0x077d1, 0x077d2, 0x077d5, 0x077d9, 0x077de, 0x077df, 0x077e0,
  0x077e4, 0x077e6, 0x077ea, 0x077ec, 0x077f0, 0x077f1, 0x077f4, 0x077f8,
  0x077fb, 0x07805, 0x07806, 0x07809, 0x0780d, 0x0780e, 0x07811, 0x0781d,
  0x07821, 0x07822, 0x07823, 0x0782d, 0x0782e, 0x07830, 0x07835, 0x07837,
  0x07843, 0x07844, 0x07847, 0x07848, 0x0784c, 0x0784e, 0x07852, 0x0785c,
  0x0785e, 0x07860, 0x07861, 0x07863, 0x07864, 0x07868, 0x0786a, 0x0786e,
  0x0787a, 0x0787e, 0x0788a, 0x0788f, 0x07894, 0x07898, 0x078a1, 0x0789d,
  0x0789e, 0x0789f, 0x078a4, 0x078a8, 0x078ac, 0x078ad, 0x078b0, 0x078b1,
  0x078b2, 0x078b3, 0x078bb, 0x078bd, 0x078bf, 0x078c7, 0x078c8, 0x078c9,
  0x078cc, 0x078ce, 0x078d2, 0x078d3, 0x078d5, 0x078d6, 0x078e4, 0x078db,
  0x078df, 0x078e0, 0x078e1, 0x078e6, 0x078ea, 0x078f2, 0x078f3, 0x07900,
  0x078f6, 0x078f7, 0x078fa, 0x078fb, 0x078ff, 0x07906, 0x0790c, 0x07910,
  0x0791a, 0x0791c, 0x0791e, 0x0791f, 0x07920, 0x07925, 0x07927, 0x07929,
  0x0792d, 0x07931, 0x07934, 0x07935, 0x0793b, 0x0793d, 0x0793f, 0x07944,
  0x07945, 0x07946, 0x0794a, 0x0794b, 0x0794f, 0x07951, 0x07954, 0x07958,
  0x0795b, 0x0795c, 0x07967, 0x07969, 0x0796b, 0x07972, 0x07979, 0x0797b,
  0x0797c, 0x0797e, 0x0798b, 0x0798c, 0x07991, 0x07993, 0x07994, 0x07995,
  0x07996, 0x07998, 0x0799b, 0x0799c, 0x079a1, 0x079a8, 0x079a9, 0x079ab,
  0x079af, 0x079b1, 0x079b4, 0x079b8, 0x079bb, 0x079c2, 0x079c4, 0x079c7,
  0x079c8, 0x079ca, 0x079cf, 0x079d4, 0x079d6, 0x079da, 0x079dd, 0x079de,
  0x079e0, 0x079e2, 0x079e5, 0x079ea, 0x079eb, 0x079ed, 0x079f1, 0x079f8,
  0x079fc, 0x07a02, 0x07a03, 0x07a07, 0x07a09, 0x07a0a, 0x07a0c, 0x07a11,
  0x07a15, 0x07a1b, 0x07a1e, 0x07a21, 0x07a27, 0x07a2b, 0x07a2d, 0x07a2f,
  0x07a30, 0x07a34, 0x07a35, 0x07a38, 0x07a39, 0x07a3a, 0x07a44, 0x07a45,
  0x07a47, 0x07a48, 0x07a4c, 0x07a55, 0x07a56, 0x07a59, 0x07a5c, 0x07a5d,
  0x07a5f, 0x07a60, 0x07a65, 0x07a67, 0x07a6a, 0x07a6d, 0x07a75, 0x07a78,
  0x07a7e, 0x07a80, 0x07a82, 0x07a85, 0x07a86, 0x07a8a, 0x07a8b, 0x07a90,
  0x07a91, 0x07a94, 0x07a9e, 0x07aa0, 0x07aa3, 0x07aac, 0x07ab3, 0x07ab5,
  0x07ab9, 0x07abb, 0x07abc, 0x07ac6, 0x07ac9, 0x07acc, 0x07ace, 0x07ad1,
  0x07adb, 0x07ae8, 0x07ae9, 0x07aeb, 0x07aec, 0x07af1, 0x07af4, 0x07afb,
  0x07afd, 0x07afe, 0x07b07, 0x07b14, 0x07b1f, 0x07b23, 0x07b27, 0x07b29,
  0x07b2a, 0x07b2b, 0x07b2d, 0x07b2e, 0x07b2f, 0x07b30, 0x07b31, 0x07b34,
  0x07b3d, 0x07b3f, 0x07b40, 0x07b41, 0x07b47, 0x07b4e, 0x07b55, 0x07b60,
  0x07b64, 0x07b66, 0x07b69, 0x07b6a, 0x07b6d, 0x07b6f, 0x07b72, 0x07b73,
  0x07b77, 0x07b84, 0x07b89, 0x07b8e, 0x07b90, 0x07b91, 0x07b96, 0x07b9b,
  0x07b9e, 0x07ba0, 0x07ba5, 0x07bac, 0x07baf, 0x07bb0, 0x07bb2, 0x07bb5,
  0x07bb6, 0x07bba, 0x07bbb, 0x07bbc, 0x07bbd, 0x07bc2, 0x07bc5, 0x07bc8,
  0x07bca, 0x07bd4, 0x07bd6, 0x07bd7, 0x07bd9, 0x07bda, 0x07bdb, 0x07be8,
  0x07bea, 0x07bf2, 0x07bf4, 0x07bf5, 0x07bf8, 0x07bf9, 0x07bfa, 0x07bfc,
  0x07bfe, 0x07c01, 0x07c02, 0x07c03, 0x07c04, 0x07c06, 0x07c09, 0x07c0b,
  0x07c0c, 0x07c0e, 0x07c0f, 0x07c19, 0x07c1b, 0x07c20, 0x07c25, 0x07c26,
  0x07c28, 0x07c2c, 0x07c31, 0x07c33, 0x07c34, 0x07c36, 0x07c39, 0x07c3a,
  0x07c46, 0x07c4a, 0x07c55, 0x07c51, 0x07c52, 0x07c53, 0x07c59, 0x07c5a,
  0x07c5b, 0x07c5c, 0x07c5d, 0x07c5e, 0x07c61, 0x07c63, 0x07c67, 0x07c69,
  0x07c6d, 0x07c6e, 0x07c70, 0x07c72, 0x07c79, 0x07c7c, 0x07c7d, 0x07c86,
  0x07c87, 0x07c8f, 0x07c94, 0x07c9e, 0x07ca0, 0x07ca6, 0x07cb0, 0x07cb6,
  0x07cb7, 0x07cba, 0x07cbb, 0x07cbc, 0x07cbf, 0x07cc4, 0x07cc7, 0x07cc8,
  0x07cc9, 0x07ccd, 0x07ccf, 0x07cd3, 0x07cd4, 0x07cd5, 0x07cd7, 0x07cd9,
  0x07cda, 0x07cdd, 0x07ce6, 0x07ce9, 0x07ceb, 0x07cf5, 0x07d03, 0x07d07,
  0x07d08, 0x07d09, 0x07d0f, 0x07d11, 0x07d12, 0x07d13, 0x07d16, 0x07d1d,
  0x07d1e, 0x07d23, 0x07d26, 0x07d2a, 0x07d2d, 0x07d31, 0x07d3c, 0x07d3d,
  0x07d3e, 0x07d40, 0x07d41, 0x07d47, 0x07d48, 0x07d4d, 0x07d51, 0x07d53,
  0x07d57, 0x07d59, 0x07d5a, 0x07d5c, 0x07d5d, 0x07d65, 0x07d67, 0x07d6a,
  0x07d70, 0x07d78, 0x07d7a, 0x07d7b, 0x07d7f, 0x07d81, 0x07d82, 0x07d83,
  0x07d85, 0x07d86, 0x07d88, 0x07d8b, 0x07d8c, 0x07d8d, 0x07d91, 0x07d96,
  0x07d97, 0x07d9d, 0x07d9e, 0x07da6, 0x07da7, 0x07daa, 0x07db3, 0x07db6,
  0x07db7, 0x07db9, 0x07dc2, 0x07dc3, 0x07dc4, 0x07dc5, 0x07dc6, 0x07dcc,
  0x07dcd, 0x07dce, 0x07dd7, 0x07dd9, 0x07e00, 0x07de2, 0x07de5, 0x07de6,
  0x07dea, 0x07deb, 0x07ded, 0x07df1, 0x07df5, 0x07df6, 0x07df9, 0x07dfa,
  0x07e08, 0x07e10, 0x07e11, 0x07e15, 0x07e17, 0x07e1c, 0x07e1d, 0x07e20,
  0x07e27, 0x07e28, 0x07e2c, 0x07e2d, 0x07e2f, 0x07e33, 0x07e36, 0x07e3f,
  0x07e44, 0x07e45, 0x07e47, 0x07e4e, 0x07e50, 0x07e52, 0x07e58, 0x07e5f,
  0x07e61, 0x07e62, 0x07e65, 0x07e6b, 0x07e6e, 0x07e6f, 0x07e73, 0x07e78,
  0x07e7e, 0x07e81, 0x07e86, 0x07e87, 0x07e8a, 0x07e8d, 0x07e91, 0x07e95,
  0x07e98, 0x07e9a, 0x07e9d, 0x07e9e, 0x07f3c, 0x07f3b, 0x07f3d, 0x07f3e,
  0x07f3f, 0x07f43, 0x07f44, 0x07f47, 0x07f4f, 0x07f52, 0x07f53, 0x07f5b,
  0x07f5c, 0x07f5d, 0x07f61, 0x07f63, 0x07f64, 0x07f65, 0x07f66, 0x07f6d,
  0x07f71, 0x07f7d, 0x07f7e, 0x07f7f, 0x07f80, 0x07f8b, 0x07f8d, 0x07f8f,
  0x07f90, 0x07f91, 0x07f96, 0x07f97, 0x07f9c, 0x07fa1, 0x07fa2, 0x07fa6,
  0x07faa, 0x07fad, 0x07fb4, 0x07fbc, 0x07fbf, 0x07fc0, 0x07fc3, 0x07fc8,
  0x07fce, 0x07fcf, 0x07fdb, 0x07fdf, 0x07fe3, 0x07fe5, 0x07fe8, 0x07fec,
  0x07fee, 0x07fef, 0x07ff2, 0x07ffa, 0x07ffd, 0x07ffe, 0x07fff, 0x08007,
  0x08008, 0x0800a, 0x0800d, 0x0800e, 0x0800f, 0x08011, 0x08013, 0x08014,
  0x08016, 0x0801d, 0x0801e, 0x0801f, 0x08020, 0x08024, 0x08026, 0x0802c,
  0x0802e, 0x08030, 0x08034, 0x08035, 0x08037, 0x08039, 0x0803a, 0x0803c,
  0x0803e, 0x08040, 0x08044, 0x08060, 0x08064, 0x08066, 0x0806d, 0x08071,
  0x08075, 0x08081, 0x08088, 0x0808e, 0x0809c, 0x0809e, 0x080a6, 0x080a7,
  0x080ab, 0x080b8, 0x080b9, 0x080c8, 0x080cd, 0x080cf, 0x080d2, 0x080d4,
  0x080d5, 0x080d7, 0x080d8, 0x080e0, 0x080ed, 0x080ee, 0x080f0, 0x080f2,
  0x080f3, 0x080f6, 0x080f9, 0x080fa, 0x080fe, 0x08103, 0x0810b, 0x08116,
  0x08117, 0x08118, 0x0811c, 0x0811e, 0x08120, 0x08124, 0x08127, 0x0812c,
  0x08130, 0x08135, 0x0813a, 0x0813c, 0x08145, 0x08147, 0x0814a, 0x0814c,
  0x08152, 0x08157, 0x08160, 0x08161, 0x08167, 0x08168, 0x08169, 0x0816d,
  0x0816f, 0x08177, 0x08181, 0x08190, 0x08184, 0x08185, 0x08186, 0x0818b,
  0x0818e, 0x08196, 0x08198, 0x0819b, 0x0819e, 0x081a2, 0x081ae, 0x081b2,
  0x081b4, 0x081bb, 0x081cb, 0x081c3, 0x081c5, 0x081ca, 0x081ce, 0x081cf,
  0x081d5, 0x081d7, 0x081db, 0x081dd, 0x081de, 0x081e1, 0x081e4, 0x081eb,
  0x081ec, 0x081f0, 0x081f1, 0x081f2, 0x081f5, 0x081f6, 0x081f8, 0x081f9,
  0x081fd, 0x081ff, 0x08200, 0x08203, 0x0820f, 0x08213, 0x08214, 0x08219,
  0x0821a, 0x0821d, 0x08221, 0x08222, 0x08228, 0x08232, 0x08234, 0x0823a,
  0x08243, 0x08244, 0x08245, 0x08246, 0x0824b, 0x0824e, 0x0824f, 0x08251,
  0x08256, 0x0825c, 0x08260, 0x08263, 0x08267, 0x0826d, 0x08274, 0x0827b,
  0x0827d, 0x0827f, 0x08280, 0x08281, 0x08283, 0x08284, 0x08287, 0x08289,
  0x0828a, 0x0828e, 0x08291, 0x08294, 0x08296, 0x08298, 0x0829a, 0x0829b,
  0x082a0, 0x082a1, 0x082a3, 0x082a4, 0x082a7, 0x082a8, 0x082a9, 0x082aa,
  0x082ae, 0x082b0, 0x082b2, 0x082b4, 0x082b7, 0x082ba, 0x082bc, 0x082be,
  0x082bf, 0x082c6, 0x082d0, 0x082d5, 0x082da, 0x082e0, 0x082e2, 0x082e4,
  0x082e8, 0x082ea, 0x082ed, 0x082ef, 0x082f6, 0x082f7, 0x082fd, 0x082fe,
  0x08300, 0x08301, 0x08307, 0x08308, 0x0830a, 0x0830b, 0x08354, 0x0831b,
  0x0831d, 0x0831e, 0x0831f, 0x08321, 0x08322, 0x0832c, 0x0832d, 0x0832e,
  0x08330, 0x08333, 0x08337, 0x0833a, 0x0833c, 0x0833d, 0x08342, 0x08343,
  0x08344, 0x08347, 0x0834d, 0x0834e, 0x08351, 0x08355, 0x08356, 0x08357,
  0x08370, 0x08378, 0x0837d, 0x0837f, 0x08380, 0x08382, 0x08384, 0x08386,
  0x0838d, 0x08392, 0x08394, 0x08395, 0x08398, 0x08399, 0x0839b, 0x0839c,
  0x0839d, 0x083a6, 0x083a7, 0x083a9, 0x083ac, 0x083be, 0x083bf, 0x083c0,
  0x083c7, 0x083c9, 0x083cf, 0x083d0, 0x083d1, 0x083d4, 0x083dd, 0x08353,
  0x083e8, 0x083ea, 0x083f6, 0x083f8, 0x083f9, 0x083fc, 0x08401, 0x08406,
  0x0840a, 0x0840f, 0x08411, 0x08415, 0x08419, 0x083ad, 0x0842f, 0x08439,
  0x08445, 0x08447, 0x08448, 0x0844a, 0x0844d, 0x0844f, 0x08451, 0x08452,
  0x08456, 0x08458, 0x08459, 0x0845a, 0x0845c, 0x08460, 0x08464, 0x08465,
  0x08467, 0x0846a, 0x08470, 0x08473, 0x08474, 0x08476, 0x08478, 0x0847c,
  0x0847d, 0x08481, 0x08485, 0x08492, 0x08493, 0x08495, 0x0849e, 0x084a6,
  0x084a8, 0x084a9, 0x084aa, 0x084af, 0x084b1, 0x084b4, 0x084ba, 0x084bd,
  0x084be, 0x084c0, 0x084c2, 0x084c7, 0x084c8, 0x084cc, 0x084cf, 0x084d3,
  0x084dc, 0x084e7, 0x084ea, 0x084ef, 0x084f0, 0x084f1, 0x084f2, 0x084f7,
  0x08532, 0x084fa, 0x084fb, 0x084fd, 0x08502, 0x08503, 0x08507, 0x0850c,
  0x0850e, 0x08510, 0x0851c, 0x0851e, 0x08522, 0x08523, 0x08524, 0x08525,
  0x08527, 0x0852a, 0x0852b, 0x0852f, 0x08533, 0x08534, 0x08536, 0x0853f,
  0x08546, 0x0854f, 0x08550, 0x08551, 0x08552, 0x08553, 0x08556, 0x08559,
  0x0855c, 0x0855d, 0x0855e, 0x0855f, 0x08560, 0x08561, 0x08562, 0x08564,
  0x0856b, 0x0856f, 0x08579, 0x0857a, 0x0857b, 0x0857d, 0x0857f, 0x08581,
  0x08585, 0x08586, 0x08589, 0x0858b, 0x0858c, 0x0858f, 0x08593, 0x08598,
  0x0859d, 0x0859f, 0x085a0, 0x085a2, 0x085a5, 0x085a7, 0x085b4, 0x085b6,
  0x085b7, 0x085b8, 0x085bc, 0x085bd, 0x085be, 0x085bf, 0x085c2, 0x085c7,
  0x085ca, 0x085cb, 0x085ce, 0x085ad, 0x085d8, 0x085da, 0x085df, 0x085e0,
  0x085e6, 0x085e8, 0x085ed, 0x085f3, 0x085f6, 0x085fc, 0x085ff, 0x08600,
  0x08604, 0x08605, 0x0860d, 0x0860e, 0x08610, 0x08611, 0x08612, 0x08618,
  0x08619, 0x0861b, 0x0861e, 0x08621, 0x08627, 0x08629, 0x08636, 0x08638,
  0x0863a, 0x0863c, 0x0863d, 0x08640, 0x08642, 0x08646, 0x08652, 0x08653,
  0x08656, 0x08657, 0x08658, 0x08659, 0x0865d, 0x08660, 0x08661, 0x08662,
  0x08663, 0x08664, 0x08669, 0x0866c, 0x0866f, 0x08675, 0x08676, 0x08677,
  0x0867a, 0x0868d, 0x08691, 0x08696, 0x08698, 0x0869a, 0x0869c, 0x086a1,
  0x086a6, 0x086a7, 0x086a8, 0x086ad, 0x086b1, 0x086b3, 0x086b4, 0x086b5,
  0x086b7, 0x086b8, 0x086b9, 0x086bf, 0x086c0, 0x086c1, 0x086c3, 0x086c5,
  0x086d1, 0x086d2, 0x086d5, 0x086d7, 0x086da, 0x086dc, 0x086e0, 0x086e3,
  0x086e5, 0x086e7, 0x08688, 0x086fa, 0x086fc, 0x086fd, 0x08704, 0x08705,
  0x08707, 0x0870b, 0x0870e, 0x0870f, 0x08710, 0x08713, 0x08714, 0x08719,
  0x0871e, 0x0871f, 0x08721, 0x08723, 0x08728, 0x0872e, 0x0872f, 0x08731,
  0x08732, 0x08739, 0x0873a, 0x0873c, 0x0873d, 0x0873e, 0x08740, 0x08743,
  0x08745, 0x0874d, 0x08758, 0x0875d, 0x08761, 0x08764, 0x08765, 0x0876f,
  0x08771, 0x08772, 0x0877b, 0x08783, 0x08784, 0x08785, 0x08786, 0x08787,
  0x08788, 0x08789, 0x0878b, 0x0878c, 0x08790, 0x08793, 0x08795, 0x08797,
  0x08798, 0x08799, 0x0879e, 0x087a0, 0x087a3, 0x087a7, 0x087ac, 0x087ad,
  0x087ae, 0x087b1, 0x087b5, 0x087be, 0x087bf, 0x087c1, 0x087c8, 0x087c9,
  0x087ca, 0x087ce, 0x087d5, 0x087d6, 0x087d9, 0x087da, 0x087dc, 0x087df,
  0x087e2, 0x087e3, 0x087e4, 0x087ea, 0x087eb, 0x087ed, 0x087f1, 0x087f3,
  0x087f8, 0x087fa, 0x087ff, 0x08801, 0x08803, 0x08806, 0x08809, 0x0880a,
  0x0880b, 0x08810, 0x08819, 0x08812, 0x08813, 0x08814, 0x08818, 0x0881a,
  0x0881b, 0x0881c, 0x0881e, 0x0881f, 0x08828, 0x0882d, 0x0882e, 0x08830,
  0x08832, 0x08835, 0x0883a, 0x0883c, 0x08841, 0x08843, 0x08845, 0x08848,
  0x08849, 0x0884a, 0x0884b, 0x0884e, 0x08851, 0x08855, 0x08856, 0x08858,
  0x0885a, 0x0885c, 0x0885f, 0x08860, 0x08864, 0x08869, 0x08871, 0x08879,
  0x0887b, 0x08880, 0x08898, 0x0889a, 0x0889b, 0x0889c, 0x0889f, 0x088a0,
  0x088a8, 0x088aa, 0x088ba, 0x088bd, 0x088be, 0x088c0, 0x088ca, 0x088cb,
  0x088cc, 0x088cd, 0x088ce, 0x088d1, 0x088d2, 0x088d3, 0x088db, 0x088de,
  0x088e7, 0x088ef, 0x088f0, 0x088f1, 0x088f5, 0x088f7, 0x08901, 0x08906,
  0x0890d, 0x0890e, 0x0890f, 0x08915, 0x08916, 0x08918, 0x08919, 0x0891a,
  0x0891c, 0x08920, 0x08926, 0x08927, 0x08928, 0x08930, 0x08931, 0x08932,
  0x08935, 0x08939, 0x0893a, 0x0893e, 0x08940, 0x08942, 0x08945, 0x08946,
  0x08949, 0x0894f, 0x08952, 0x08957, 0x0895a, 0x0895b, 0x0895c, 0x08961,
  0x08962, 0x08963, 0x0896b, 0x0896e, 0x08970, 0x08973, 0x08975, 0x0897a,
  0x0897b, 0x0897c, 0x0897d, 0x08989, 0x0898d, 0x08990, 0x08994, 0x08995,
  0x0899b, 0x0899c, 0x0899f, 0x089a0, 0x089a5, 0x089b0, 0x089b4, 0x089b5,
  0x089b6, 0x089b7, 0x089bc, 0x089d4, 0x089d5, 0x089d6, 0x089d7, 0x089d8,
  0x089e5, 0x089e9, 0x089eb, 0x089ed, 0x089f1, 0x089f3, 0x089f6, 0x089f9,
  0x089fd, 0x089ff, 0x08a04, 0x08a05, 0x08a07, 0x08a0f, 0x08a11, 0x08a12,
  0x08a14, 0x08a15, 0x08a1e, 0x08a20, 0x08a22, 0x08a24, 0x08a26, 0x08a2b,
  0x08a2c, 0x08a2f, 0x08a35, 0x08a37, 0x08a3d, 0x08a3e, 0x08a40, 0x08a43,
  0x08a45, 0x08a47, 0x08a49, 0x08a4d, 0x08a4e, 0x08a53, 0x08a56, 0x08a57,
  0x08a58, 0x08a5c, 0x08a5d, 0x08a61, 0x08a65, 0x08a67, 0x08a75, 0x08a76,
  0x08a77, 0x08a79, 0x08a7a, 0x08a7b, 0x08a7e, 0x08a7f, 0x08a80, 0x08a83,
  0x08a86, 0x08a8b, 0x08a8f, 0x08a90, 0x08a92, 0x08a96, 0x08a97, 0x08a99,
  0x08a9f, 0x08aa7, 0x08aa9, 0x08aae, 0x08aaf, 0x08ab3, 0x08ab6, 0x08ab7,
  0x08abb, 0x08abe, 0x08ac3, 0x08ac6, 0x08ac8, 0x08ac9, 0x08aca, 0x08ad1,
  0x08ad3, 0x08ad4, 0x08ad5, 0x08ad7, 0x08add, 0x08adf, 0x08aec, 0x08af0,
  0x08af4, 0x08af5, 0x08af6, 0x08afc, 0x08aff, 0x08b05, 0x08b06, 0x08b0b,
  0x08b11, 0x08b1c, 0x08b1e, 0x08b1f, 0x08b0a, 0x08b2d, 0x08b30, 0x08b37,
  0x08b3c, 0x08b42, 0x08b43, 0x08b44, 0x08b45, 0x08b46, 0x08b48, 0x08b52,
  0x08b53, 0x08b54, 0x08b59, 0x08b4d, 0x08b5e, 0x08b63, 0x08b6d, 0x08b76,
  0x08b78, 0x08b79, 0x08b7c, 0x08b7e, 0x08b81, 0x08b84, 0x08b85, 0x08b8b,
  0x08b8d, 0x08b8f, 0x08b94, 0x08b95, 0x08b9c, 0x08b9e, 0x08b9f, 0x08c38,
  0x08c39, 0x08c3d, 0x08c3e, 0x08c45, 0x08c47, 0x08c49, 0x08c4b, 0x08c4f,
  0x08c51, 0x08c53, 0x08c54, 0x08c57, 0x08c58, 0x08c5b, 0x08c5d, 0x08c59,
  0x08c63, 0x08c64, 0x08c66, 0x08c68, 0x08c69, 0x08c6d, 0x08c73, 0x08c75,
  0x08c76, 0x08c7b, 0x08c7e, 0x08c86, 0x08c87, 0x08c8b, 0x08c90, 0x08c92,
  0x08c93, 0x08c99, 0x08c9b, 0x08c9c, 0x08ca4, 0x08cb9, 0x08cba, 0x08cc5,
  0x08cc6, 0x08cc9, 0x08ccb, 0x08ccf, 0x08cd6, 0x08cd5, 0x08cd9, 0x08cdd,
  0x08ce1, 0x08ce8, 0x08cec, 0x08cef, 0x08cf0, 0x08cf2, 0x08cf5, 0x08cf7,
  0x08cf8, 0x08cfe, 0x08cff, 0x08d01, 0x08d03, 0x08d09, 0x08d12, 0x08d17,
  0x08d1b, 0x08d65, 0x08d69, 0x08d6c, 0x08d6e, 0x08d7f, 0x08d82, 0x08d84,
  0x08d88, 0x08d8d, 0x08d90, 0x08d91, 0x08d95, 0x08d9e, 0x08d9f, 0x08da0,
  0x08da6, 0x08dab, 0x08dac, 0x08daf, 0x08db2, 0x08db5, 0x08db7, 0x08db9,
  0x08dbb, 0x08dc0, 0x08dc5, 0x08dc6, 0x08dc7, 0x08dc8, 0x08dca, 0x08dce,
  0x08dd1, 0x08dd4, 0x08dd5, 0x08dd7, 0x08dd9, 0x08de4, 0x08de5, 0x08de7,
  0x08dec, 0x08df0, 0x08dbc, 0x08df1, 0x08df2, 0x08df4, 0x08dfd, 0x08e01,
  0x08e04, 0x08e05, 0x08e06, 0x08e0b, 0x08e11, 0x08e14, 0x08e16, 0x08e20,
  0x08e21, 0x08e22, 0x08e23, 0x08e26, 0x08e27, 0x08e31, 0x08e33, 0x08e36,
  0x08e37, 0x08e38, 0x08e39, 0x08e3d, 0x08e40, 0x08e41, 0x08e4b, 0x08e4d,
  0x08e4e, 0x08e4f, 0x08e54, 0x08e5b, 0x08e5c, 0x08e5d, 0x08e5e, 0x08e61,
  0x08e62, 0x08e69, 0x08e6c, 0x08e6d, 0x08e6f, 0x08e70, 0x08e71, 0x08e79,
  0x08e7a, 0x08e7b, 0x08e82, 0x08e83, 0x08e89, 0x08e90, 0x08e92, 0x08e95,
  0x08e9a, 0x08e9b, 0x08e9d, 0x08e9e, 0x08ea2, 0x08ea7, 0x08ea9, 0x08ead,
  0x08eae, 0x08eb3, 0x08eb5, 0x08eba, 0x08ebb, 0x08ec0, 0x08ec1, 0x08ec3,
  0x08ec4, 0x08ec7, 0x08ecf, 0x08ed1, 0x08ed4, 0x08edc, 0x08ee8, 0x08eee,
  0x08ef0, 0x08ef1, 0x08ef7, 0x08ef9, 0x08efa, 0x08eed, 0x08f00, 0x08f02,
  0x08f07, 0x08f08, 0x08f0f, 0x08f10, 0x08f16, 0x08f17, 0x08f18, 0x08f1e,
  0x08f20, 0x08f21, 0x08f23, 0x08f25, 0x08f27, 0x08f28, 0x08f2c, 0x08f2d,
  0x08f2e, 0x08f34, 0x08f35, 0x08f36, 0x08f37, 0x08f3a, 0x08f40, 0x08f41,
  0x08f43, 0x08f47, 0x08f4f, 0x08f51, 0x08f52, 0x08f53, 0x08f54, 0x08f55,
  0x08f58, 0x08f5d, 0x08f5e, 0x08f65, 0x08f9d, 0x08fa0, 0x08fa1, 0x08fa4,
  0x08fa5, 0x08fa6, 0x08fb5, 0x08fb6, 0x08fb8, 0x08fbe, 0x08fc0, 0x08fc1,
  0x08fc6, 0x08fca, 0x08fcb, 0x08fcd, 0x08fd0, 0x08fd2, 0x08fd3, 0x08fd5,
  0x08fe0, 0x08fe3, 0x08fe4, 0x08fe8, 0x08fee, 0x08ff1, 0x08ff5, 0x08ff6,
  0x08ffb, 0x08ffe, 0x09002, 0x09004, 0x09008, 0x0900c, 0x09018, 0x0901b,
  0x09028, 0x09029, 0x0902f, 0x0902a, 0x0902c, 0x0902d, 0x09033, 0x09034,
  0x09037, 0x0903f, 0x09043, 0x09044, 0x0904c, 0x0905b, 0x0905d, 0x09062,
  0x09066, 0x09067, 0x0906c, 0x09070, 0x09074, 0x09079, 0x09085, 0x09088,
  0x0908b, 0x0908c, 0x0908e, 0x09090, 0x09095, 0x09097, 0x09098, 0x09099,
  0x0909b, 0x090a0, 0x090a1, 0x090a2, 0x090a5, 0x090b0, 0x090b2, 0x090b3,
  0x090b4, 0x090b6, 0x090bd, 0x090cc, 0x090be, 0x090c3, 0x090c4, 0x090c5,
  0x090c7, 0x090c8, 0x090d5, 0x090d7, 0x090d8, 0x090d9, 0x090dc, 0x090dd,
  0x090df, 0x090e5, 0x090d2, 0x090f6, 0x090eb, 0x090ef, 0x090f0, 0x090f4,
  0x090fe, 0x090ff, 0x09100, 0x09104, 0x09105, 0x09106, 0x09108, 0x0910d,
  0x09110, 0x09114, 0x09116, 0x09117, 0x09118, 0x0911a, 0x0911c, 0x0911e,
  0x09120, 0x09125, 0x09122, 0x09123, 0x09127, 0x09129, 0x0912e, 0x0912f,
  0x09131, 0x09134, 0x09136, 0x09137, 0x09139, 0x0913a, 0x0913c, 0x0913d,
  0x09143, 0x09147, 0x09148, 0x0914f, 0x09153, 0x09157, 0x09159, 0x0915a,
  0x0915b, 0x09161, 0x09164, 0x09167, 0x0916d, 0x09174, 0x09179, 0x0917a,
  0x0917b, 0x09181, 0x09183, 0x09185, 0x09186, 0x0918a, 0x0918e, 0x09191,
  0x09193, 0x09194, 0x09195, 0x09198, 0x0919e, 0x091a1, 0x091a6, 0x091a8,
  0x091ac, 0x091ad, 0x091ae, 0x091b0, 0x091b1, 0x091b2, 0x091b3, 0x091b6,
  0x091bb, 0x091bc, 0x091bd, 0x091bf, 0x091c2, 0x091c3, 0x091c5, 0x091d3,
  0x091d4, 0x091d7, 0x091d9, 0x091da, 0x091de, 0x091e4, 0x091e5, 0x091e9,
  0x091ea, 0x091ec, 0x091ed, 0x091ee, 0x091ef, 0x091f0, 0x091f1, 0x091f7,
  0x091f9, 0x091fb, 0x091fd, 0x09200, 0x09201, 0x09204, 0x09205, 0x09206,
  0x09207, 0x09209, 0x0920a, 0x0920c, 0x09210, 0x09212, 0x09213, 0x09216,
  0x09218, 0x0921c, 0x0921d, 0x09223, 0x09224, 0x09225, 0x09226, 0x09228,
  0x0922e, 0x0922f, 0x09230, 0x09233, 0x09235, 0x09236, 0x09238, 0x09239,
  0x0923a, 0x0923c, 0x0923e, 0x09240, 0x09242, 0x09243, 0x09246, 0x09247,
  0x0924a, 0x0924d, 0x0924e, 0x0924f, 0x09251, 0x09258, 0x09259, 0x0925c,
  0x0925d, 0x09260, 0x09261, 0x09265, 0x09267, 0x09268, 0x09269, 0x0926e,
  0x0926f, 0x09270, 0x09275, 0x09276, 0x09277, 0x09278, 0x09279, 0x0927b,
  0x0927c, 0x0927d, 0x0927f, 0x09288, 0x09289, 0x0928a, 0x0928d, 0x0928e,
  0x09292, 0x09297, 0x09299, 0x0929f, 0x092a0, 0x092a4, 0x092a5, 0x092a7,
  0x092a8, 0x092ab, 0x092af, 0x092b2, 0x092b6, 0x092b8, 0x092ba, 0x092bb,
  0x092bc, 0x092bd, 0x092bf, 0x092c0, 0x092c1, 0x092c2, 0x092c3, 0x092c5,
  0x092c6, 0x092c7, 0x092c8, 0x092cb, 0x092cc, 0x092cd, 0x092ce, 0x092d0,
  0x092d3, 0x092d5, 0x092d7, 0x092d8, 0x092d9, 0x092dc, 0x092dd, 0x092df,
  0x092e0, 0x092e1, 0x092e3, 0x092e5, 0x092e7, 0x092e8, 0x092ec, 0x092ee,
  0x092f0, 0x092f9, 0x092fb, 0x092ff, 0x09300, 0x09302, 0x09308, 0x0930d,
  0x09311, 0x09314, 0x09315, 0x0931c, 0x0931d, 0x0931e, 0x0931f, 0x09321,
  0x09324, 0x09325, 0x09327, 0x09329, 0x0932a, 0x09333, 0x09334, 0x09336,
  0x09337, 0x09347, 0x09348, 0x09349, 0x09350, 0x09351, 0x09352, 0x09355,
  0x09357, 0x09358, 0x0935a, 0x0935e, 0x09364, 0x09365, 0x09367, 0x09369,
  0x0936a, 0x0936d, 0x0936f, 0x09370, 0x09371, 0x09373, 0x09374, 0x09376,
  0x0937a, 0x0937d, 0x0937f, 0x09380, 0x09381, 0x09382, 0x09388, 0x0938a,
  0x0938b, 0x0938d, 0x0938f, 0x09392, 0x09395, 0x09398, 0x0939b, 0x0939e,
  0x093a1, 0x093a3, 0x093a4, 0x093a6, 0x093a8, 0x093ab, 0x093b4, 0x093b5,
  0x093b6, 0x093ba, 0x093a9, 0x093c1, 0x093c4, 0x093c5, 0x093c6, 0x093c7,
  0x093c9, 0x093ca, 0x093cb, 0x093cc, 0x093cd, 0x093d3, 0x093d9, 0x093dc,
  0x093de, 0x093df, 0x093e2, 0x093e6, 0x093e7, 0x093f9, 0x093f7, 0x093f8,
  0x093fa, 0x093fb, 0x093fd, 0x09401, 0x09402, 0x09404, 0x09408, 0x09409,
  0x0940d, 0x0940e, 0x0940f, 0x09415, 0x09416, 0x09417, 0x0941f, 0x0942e,
  0x0942f, 0x09431, 0x09432, 0x09433, 0x09434, 0x0943b, 0x0943f, 0x0943d,
  0x09443, 0x09445, 0x09448, 0x0944a, 0x0944c, 0x09455, 0x09459, 0x0945c,
  0x0945f, 0x09461, 0x09463, 0x09468, 0x0946b, 0x0946d, 0x0946e, 0x0946f,
  0x09471, 0x09472, 0x09484, 0x09483, 0x09578, 0x09579, 0x0957e, 0x09584,
  0x09588, 0x0958c, 0x0958d, 0x0958e, 0x0959d, 0x0959e, 0x0959f, 0x095a1,
  0x095a6, 0x095a9, 0x095ab, 0x095ac, 0x095b4, 0x095b6, 0x095ba, 0x095bd,
  0x095bf, 0x095c6, 0x095c8, 0x095c9, 0x095cb, 0x095d0, 0x095d1, 0x095d2,
  0x095d3, 0x095d9, 0x095da, 0x095dd, 0x095de, 0x095df, 0x095e0, 0x095e4,
  0x095e6, 0x0961d, 0x0961e, 0x09622, 0x09624, 0x09625, 0x09626, 0x0962c,
  0x09631, 0x09633, 0x09637, 0x09638, 0x09639, 0x0963a, 0x0963c, 0x0963d,
  0x09641, 0x09652, 0x09654, 0x09656, 0x09657, 0x09658, 0x09661, 0x0966e,
  0x09674, 0x0967b, 0x0967c, 0x0967e, 0x0967f, 0x09681, 0x09682, 0x09683,
  0x09684, 0x09689, 0x09691, 0x09696, 0x0969a, 0x0969d, 0x0969f, 0x096a4,
  0x096a5, 0x096a6, 0x096a9, 0x096ae, 0x096af, 0x096b3, 0x096ba, 0x096ca,
  0x096d2, 0x05db2, 0x096d8, 0x096da, 0x096dd, 0x096de, 0x096df, 0x096e9,
  0x096ef, 0x096f1, 0x096fa, 0x09702, 0x09703, 0x09705, 0x09709, 0x0971a,
  0x0971b, 0x0971d, 0x09721, 0x09722, 0x09723, 0x09728, 0x09731, 0x09733,
  0x09741, 0x09743, 0x0974a, 0x0974e, 0x0974f, 0x09755, 0x09757, 0x09758,
  0x0975a, 0x0975b, 0x09763, 0x09767, 0x0976a, 0x0976e, 0x09773, 0x09776,
  0x09777, 0x09778, 0x0977b, 0x0977d, 0x0977f, 0x09780, 0x09789, 0x09795,
  0x09796, 0x09797, 0x09799, 0x0979a, 0x0979e, 0x0979f, 0x097a2, 0x097ac,
  0x097ae, 0x097b1, 0x097b2, 0x097b5, 0x097b6, 0x097b8, 0x097b9, 0x097ba,
  0x097bc, 0x097be, 0x097bf, 0x097c1, 0x097c4, 0x097c5, 0x097c7, 0x097c9,
  0x097ca, 0x097cc, 0x097cd, 0x097ce, 0x097d0, 0x097d1, 0x097d4, 0x097d7,
  0x097d8, 0x097d9, 0x097dd, 0x097de, 0x097e0, 0x097db, 0x097e1, 0x097e4,
  0x097ef, 0x097f1, 0x097f4, 0x097f7, 0x097f8, 0x097fa, 0x09807, 0x0980a,
  0x09819, 0x0980d, 0x0980e, 0x09814, 0x09816, 0x0981c, 0x0981e, 0x09820,
  0x09823, 0x09826, 0x0982b, 0x0982e, 0x0982f, 0x09830, 0x09832, 0x09833,
  0x09835, 0x09825, 0x0983e, 0x09844, 0x09847, 0x0984a, 0x09851, 0x09852,
  0x09853, 0x09856, 0x09857, 0x09859, 0x0985a, 0x09862, 0x09863, 0x09865,
  0x09866, 0x0986a, 0x0986c, 0x098ab, 0x098ad, 0x098ae, 0x098b0, 0x098b4,
  0x098b7, 0x098b8, 0x098ba, 0x098bb, 0x098bf, 0x098c2, 0x098c5, 0x098c8,
  0x098cc, 0x098e1, 0x098e3, 0x098e5, 0x098e6, 0x098e7, 0x098ea, 0x098f3,
  0x098f6, 0x09902, 0x09907, 0x09908, 0x09911, 0x09915, 0x09916, 0x09917,
  0x0991a, 0x0991b, 0x0991c, 0x0991f, 0x09922, 0x09926, 0x09927, 0x0992b,
  0x09931, 0x09932, 0x09933, 0x09934, 0x09935, 0x09939, 0x0993a, 0x0993b,
  0x0993c, 0x09940, 0x09941, 0x09946, 0x09947, 0x09948, 0x0994d, 0x0994e,
  0x09954, 0x09958, 0x09959, 0x0995b, 0x0995c, 0x0995e, 0x0995f, 0x09960,
  0x0999b, 0x0999d, 0x0999f, 0x099a6, 0x099b0, 0x099b1, 0x099b2, 0x099b5,
  0x099b9, 0x099ba, 0x099bd, 0x099bf, 0x099c3, 0x099c9, 0x099d3, 0x099d4,
  0x099d9, 0x099da, 0x099dc, 0x099de, 0x099e7, 0x099ea, 0x099eb, 0x099ec,
  0x099f0, 0x099f4, 0x099f5, 0x099f9, 0x099fd, 0x099fe, 0x09a02, 0x09a03,
  0x09a04, 0x09a0b, 0x09a0c, 0x09a10, 0x09a11, 0x09a16, 0x09a1e, 0x09a20,
  0x09a22, 0x09a23, 0x09a24, 0x09a27, 0x09a2d, 0x09a2e, 0x09a33, 0x09a35,
  0x09a36, 0x09a38, 0x09a47, 0x09a41, 0x09a44, 0x09a4a, 0x09a4b, 0x09a4c,
  0x09a4e, 0x09a51, 0x09a54, 0x09a56, 0x09a5d, 0x09aaa, 0x09aac, 0x09aae,
  0x09aaf, 0x09ab2, 0x09ab4, 0x09ab5, 0x09ab6, 0x09ab9, 0x09abb, 0x09abe,
  0x09abf, 0x09ac1, 0x09ac3, 0x09ac6, 0x09ac8, 0x09ace, 0x09ad0, 0x09ad2,
  0x09ad5, 0x09ad6, 0x09ad7, 0x09adb, 0x09adc, 0x09ae0, 0x09ae4, 0x09ae5,
  0x09ae7, 0x09ae9, 0x09aec, 0x09af2, 0x09af3, 0x09af5, 0x09af9, 0x09afa,
  0x09afd, 0x09aff, 0x09b00, 0x09b01, 0x09b02, 0x09b03, 0x09b04, 0x09b05,
  0x09b08, 0x09b09, 0x09b0b, 0x09b0c, 0x09b0d, 0x09b0e, 0x09b10, 0x09b12,
  0x09b16, 0x09b19, 0x09b1b, 0x09b1c, 0x09b20, 0x09b26, 0x09b2b, 0x09b2d,
  0x09b33, 0x09b34, 0x09b35, 0x09b37, 0x09b39, 0x09b3a, 0x09b3d, 0x09b48,
  0x09b4b, 0x09b4c, 0x09b55, 0x09b56, 0x09b57, 0x09b5b, 0x09b5e, 0x09b61,
  0x09b63, 0x09b65, 0x09b66, 0x09b68, 0x09b6a, 0x09b6b, 0x09b6c, 0x09b6d,
  0x09b6e, 0x09b73, 0x09b75, 0x09b77, 0x09b78, 0x09b79, 0x09b7f, 0x09b80,
  0x09b84, 0x09b85, 0x09b86, 0x09b87, 0x09b89, 0x09b8a, 0x09b8b, 0x09b8d,
  0x09b8f, 0x09b90, 0x09b94, 0x09b9a, 0x09b9d, 0x09b9e, 0x09ba6, 0x09ba7,
  0x09ba9, 0x09bac, 0x09bb0, 0x09bb1, 0x09bb2, 0x09bb7, 0x09bb8, 0x09bbb,
  0x09bbc, 0x09bbe, 0x09bbf, 0x09bc1, 0x09bc7, 0x09bc8, 0x09bce, 0x09bd0,
  0x09bd7, 0x09bd8, 0x09bdd, 0x09bdf, 0x09be5, 0x09be7, 0x09bea, 0x09beb,
  0x09bef, 0x09bf3, 0x09bf7, 0x09bf8, 0x09bf9, 0x09bfa, 0x09bfd, 0x09bff,
  0x09c00, 0x09c02, 0x09c0b, 0x09c0f, 0x09c11, 0x09c16, 0x09c18, 0x09c19,
  0x09c1a, 0x09c1c, 0x09c1e, 0x09c22, 0x09c23, 0x09c26, 0x09c27, 0x09c28,
  0x09c29, 0x09c2a, 0x09c31, 0x09c35, 0x09c36, 0x09c37, 0x09c3d, 0x09c41,
  0x09c43, 0x09c44, 0x09c45, 0x09c49, 0x09c4a, 0x09c4e, 0x09c4f, 0x09c50,
  0x09c53, 0x09c54, 0x09c56, 0x09c58, 0x09c5b, 0x09c5d, 0x09c5e, 0x09c5f,
  0x09c63, 0x09c69, 0x09c6a, 0x09c5c, 0x09c6b, 0x09c68, 0x09c6e, 0x09c70,
  0x09c72, 0x09c75, 0x09c77, 0x09c7b, 0x09ce6, 0x09cf2, 0x09cf7, 0x09cf9,
  0x09d0b, 0x09d02, 0x09d11, 0x09d17, 0x09d18, 0x09d1c, 0x09d1d, 0x09d1e,
  0x09d2f, 0x09d30, 0x09d32, 0x09d33, 0x09d34, 0x09d3a, 0x09d3c, 0x09d45,
  0x09d3d, 0x09d42, 0x09d43, 0x09d47, 0x09d4a, 0x09d53, 0x09d54, 0x09d5f,
  0x09d63, 0x09d62, 0x09d65, 0x09d69, 0x09d6a, 0x09d6b, 0x09d70, 0x09d76,
  0x09d77, 0x09d7b, 0x09d7c, 0x09d7e, 0x09d83, 0x09d84, 0x09d86, 0x09d8a,
  0x09d8d, 0x09d8e, 0x09d92, 0x09d93, 0x09d95, 0x09d96, 0x09d97, 0x09d98,
  0x09da1, 0x09daa, 0x09dac, 0x09dae, 0x09db1, 0x09db5, 0x09db9, 0x09dbc,
  0x09dbf, 0x09dc3, 0x09dc7, 0x09dc9, 0x09dca, 0x09dd4, 0x09dd5, 0x09dd6,
  0x09dd7, 0x09dda, 0x09dde, 0x09ddf, 0x09de0, 0x09de5, 0x09de7, 0x09de9,
  0x09deb, 0x09dee, 0x09df0, 0x09df3, 0x09df4, 0x09dfe, 0x09e0a, 0x09e02,
  0x09e07, 0x09e0e, 0x09e10, 0x09e11, 0x09e12, 0x09e15, 0x09e16, 0x09e19,
  0x09e1c, 0x09e1d, 0x09e7a, 0x09e7b, 0x09e7c, 0x09e80, 0x09e82, 0x09e83,
  0x09e84, 0x09e85, 0x09e87, 0x09e8e, 0x09e8f, 0x09e96, 0x09e98, 0x09e9b,
  0x09e9e, 0x09ea4, 0x09ea8, 0x09eac, 0x09eae, 0x09eaf, 0x09eb0, 0x09eb3,
  0x09eb4, 0x09eb5, 0x09ec6, 0x09ec8, 0x09ecb, 0x09ed5, 0x09edf, 0x09ee4,
  0x09ee7, 0x09eec, 0x09eed, 0x09eee, 0x09ef0, 0x09ef1, 0x09ef2, 0x09ef5,
  0x09ef8, 0x09eff, 0x09f02, 0x09f03, 0x09f09, 0x09f0f, 0x09f10, 0x09f11,
  0x09f12, 0x09f14, 0x09f16, 0x09f17, 0x09f19, 0x09f1a, 0x09f1b, 0x09f1f,
  0x09f22, 0x09f26, 0x09f2a, 0x09f2b, 0x09f2f, 0x09f31, 0x09f32, 0x09f34,
  0x09f37, 0x09f39, 0x09f3a, 0x09f3c, 0x09f3d, 0x09f3f, 0x09f41, 0x09f43,
  0x09f44, 0x09f45, 0x09f46, 0x09f47, 0x09f53, 0x09f55, 0x09f56, 0x09f57,
  0x09f58, 0x09f5a, 0x09f5d, 0x09f5e, 0x09f68, 0x09f69, 0x09f6d, 0x09f6e,
  0x09f6f, 0x09f70, 0x09f71, 0x09f73, 0x09f75, 0x09f7a, 0x09f7d, 0x09f8f,
  0x09f90, 0x09f91, 0x09f92, 0x09f94, 0x09f96, 0x09f97, 0x09f9e, 0x09fa1,
  0x09fa2, 0x09fa3, 0x09fa5, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x06b9b, 0x06bae,
  0x23a98, 0x06bbd, 0x06bbe, 0x06bc7, 0x06bc8, 0x06bc9, 0x06bda, 0x06be6,
  0x06be7, 0x06bee, 0x06bf1, 0x06c02, 0x06c0a, 0x06c0e, 0x06c35, 0x06c36,
  0x06c3a, 0x23c7f, 0x06c3f, 0x06c4d, 0x06c5b, 0x06c6d, 0x06c84, 0x06c89,
  0x03cc3, 0x06c94, 0x06c95, 0x06c97, 0x06cad, 0x06cc2, 0x06cd0, 0x03cd2,
  0x06cd6, 0x06cda, 0x06cdc, 0x06ce9, 0x06cec, 0x06ced, 0x23d00, 0x06d00,
  0x06d0a, 0x06d24, 0x06d26, 0x06d27, 0x06c67, 0x06d2f, 0x06d3c, 0x06d5b,
  0x06d5e, 0x06d60, 0x06d70, 0x06d80, 0x06d81, 0x06d8a, 0x06d8d, 0x06d91,
  0x06d98, 0x23d40, 0x06e17, 0x23dfa, 0x23df9, 0x23dd3, 0x06dab, 0x06dae,
  0x06db4, 0x06dc2, 0x06d34, 0x06dc8, 0x06dce, 0x06dcf, 0x06dd0, 0x06ddf,
  0x06de9, 0x06df6, 0x06e36, 0x06e1e, 0x06e22, 0x06e27, 0x03d11, 0x06e32,
  0x06e3c, 0x06e48, 0x06e49, 0x06e4b, 0x06e4c, 0x06e4f, 0x06e51, 0x06e53,
  0x06e54, 0x06e57, 0x06e63, 0x03d1e, 0x06e93, 0x06ea7, 0x06eb4, 0x06ebf,
  0x06ec3, 0x06eca, 0x06ed9, 0x06f35, 0x06eeb, 0x06ef9, 0x06efb, 0x06f0a,
  0x06f0c, 0x06f18, 0x06f25, 0x06f36, 0x06f3c, 0x23f7e, 0x06f52, 0x06f57,
  0x06f5a, 0x06f60, 0x06f68, 0x06f98, 0x06f7d, 0x06f90, 0x06f96, 0x06fbe,
  0x06f9f, 0x06fa5, 0x06faf, 0x03d64, 0x06fb5, 0x06fc8, 0x06fc9, 0x06fda,
  0x06fde, 0x06fe9, 0x24096, 0x06ffc, 0x07000, 0x07007, 0x0700a, 0x07023,
  0x24103, 0x07039, 0x0703a, 0x0703c, 0x07043, 0x07047, 0x0704b, 0x03d9a,
  0x07054, 0x07065, 0x07069, 0x0706c, 0x0706e, 0x07076, 0x0707e, 0x07081,
  0x07086, 0x07095, 0x07097, 0x070bb, 0x241c6, 0x0709f, 0x070b1, 0x241fe,
  0x070ec, 0x070ca, 0x070d1, 0x070d3, 0x070dc, 0x07103, 0x07104, 0x07106,
  0x07107, 0x07108, 0x0710c, 0x03dc0, 0x0712f, 0x07131, 0x07150, 0x0714a,
  0x07153, 0x0715e, 0x03dd4, 0x07196, 0x07180, 0x0719b, 0x071a0, 0x071a2,
  0x071ae, 0x071af, 0x071b3, 0x243bc, 0x071cb, 0x071d3, 0x071d9, 0x071dc,
  0x07207, 0x03e05, 0x0fa49, 0x0722b, 0x07234, 0x07238, 0x07239, 0x04e2c,
  0x07242, 0x07253, 0x07257, 0x07263, 0x24629, 0x0726e, 0x0726f, 0x07278,
  0x0727f, 0x0728e, 0x246a5, 0x072ad, 0x072ae, 0x072b0, 0x072b1, 0x072c1,
  0x03e60, 0x072cc, 0x03e66, 0x03e68, 0x072f3, 0x072fa, 0x07307, 0x07312,
  0x07318, 0x07319, 0x03e83, 0x07339, 0x0732c, 0x07331, 0x07333, 0x0733d,
  0x07352, 0x03e94, 0x0736b, 0x0736c, 0x24896, 0x0736e, 0x0736f, 0x07371,
  0x07377, 0x07381, 0x07385, 0x0738a, 0x07394, 0x07398, 0x0739c, 0x0739e,
  0x073a5, 0x073a8, 0x073b5, 0x073b7, 0x073b9, 0x073bc, 0x073bf, 0x073c5,
  0x073cb, 0x073e1, 0x073e7, 0x073f9, 0x07413, 0x073fa, 0x07401, 0x07424,
  0x07431, 0x07439, 0x07453, 0x07440, 0x07443, 0x0744d, 0x07452, 0x0745d,
  0x07471, 0x07481, 0x07485, 0x07488, 0x24a4d, 0x07492, 0x07497, 0x07499,
  0x074a0, 0x074a1, 0x074a5, 0x074aa, 0x074ab, 0x074b9, 0x074bb, 0x074ba,
  0x074d6, 0x074d8, 0x074de, 0x074ef, 0x074eb, 0x24b56, 0x074fa, 0x24b6f,
  0x07520, 0x07524, 0x0752a, 0x03f57, 0x24c16, 0x0753d, 0x0753e, 0x07540,
  0x07548, 0x0754e, 0x07550, 0x07552, 0x0756c, 0x07572, 0x07571, 0x0757a,
  0x0757d, 0x0757e, 0x07581, 0x24d14, 0x0758c, 0x03f75, 0x075a2, 0x03f77,
  0x075b0, 0x075b7, 0x075bf, 0x075c0, 0x075c6, 0x075cf, 0x075d3, 0x075dd,
  0x075df, 0x075e0, 0x075e7, 0x075ec, 0x075ee, 0x075f1, 0x075f9, 0x07603,
  0x07618, 0x07607, 0x0760f, 0x03fae, 0x24e0e, 0x07613, 0x0761b, 0x0761c,
  0x24e37, 0x07625, 0x07628, 0x0763c, 0x07633, 0x24e6a, 0x03fc9, 0x07641,
  0x24e8b, 0x07649, 0x07655, 0x03fd7, 0x0766e, 0x07695, 0x0769c, 0x076a1,
  0x076a0, 0x076a7, 0x076a8, 0x076af, 0x2504a, 0x076c9, 0x25055, 0x076e8,
  0x076ec, 0x25122, 0x07717, 0x0771a, 0x0772d, 0x07735, 0x251a9, 0x04039,
  0x251e5, 0x251cd, 0x07758, 0x07760, 0x0776a, 0x2521e, 0x07772, 0x0777c,
  0x0777d, 0x2524c, 0x04058, 0x0779a, 0x0779f, 0x077a2, 0x077a4, 0x077a9,
  0x077de, 0x077df, 0x077e4, 0x077e6, 0x077ea, 0x077ec, 0x04093, 0x077f0,
  0x077f4, 0x077fb, 0x2542e, 0x07805, 0x07806, 0x07809, 0x0780d, 0x07819,
  0x07821, 0x0782c, 0x07847, 0x07864, 0x0786a, 0x254d9, 0x0788a, 0x07894,
  0x078a4, 0x0789d, 0x0789e, 0x0789f, 0x078bb, 0x078c8, 0x078cc, 0x078ce,
  0x078d5, 0x078e0, 0x078e1, 0x078e6, 0x078f9, 0x078fa, 0x078fb, 0x078fe,
  0x255a7, 0x07910, 0x0791b, 0x07930, 0x07925, 0x0793b, 0x0794a, 0x07958,
  0x0795b, 0x04105, 0x07967, 0x07972, 0x07994, 0x07995, 0x07996, 0x0799b,
  0x079a1, 0x079a9, 0x079b4, 0x079bb, 0x079c2, 0x079c7, 0x079cc, 0x079cd,
  0x079d6, 0x04148, 0x257a9, 0x257b4, 0x0414f, 0x07a0a, 0x07a11, 0x07a15,
  0x07a1b, 0x07a1e, 0x04163, 0x07a2d, 0x07a38, 0x07a47, 0x07a4c, 0x07a56,
  0x07a59, 0x07a5c, 0x07a5f, 0x07a60, 0x07a67, 0x07a6a, 0x07a75, 0x07a78,
  0x07a82, 0x07a8a, 0x07a90, 0x07aa3, 0x07aac, 0x259d4, 0x041b4, 0x07ab9,
  0x07abc, 0x07abe, 0x041bf, 0x07acc, 0x07ad1, 0x07ae7, 0x07ae8, 0x07af4,
  0x25ae4, 0x25ae3, 0x07b07, 0x25af1, 0x07b3d, 0x07b27, 0x07b2a, 0x07b2e,
  0x07b2f, 0x07b31, 0x041e6, 0x041f3, 0x07b7f, 0x07b41, 0x041ee, 0x07b55,
  0x07b79, 0x07b64, 0x07b66, 0x07b69, 0x07b73, 0x25bb2, 0x04207, 0x07b90,
  0x07b91, 0x07b9b, 0x0420e, 0x07baf, 0x07bb5, 0x07bbc, 0x07bc5, 0x07bca,
  0x25c4b, 0x25c64, 0x07bd4, 0x07bd6, 0x07bda, 0x07bea, 0x07bf0, 0x07c03,
  0x07c0b, 0x07c0e, 0x07c0f, 0x07c26, 0x07c45, 0x07c4a, 0x07c51, 0x07c57,
  0x07c5e, 0x07c61, 0x07c69, 0x07c6e, 0x07c6f, 0x07c70, 0x25e2e, 0x25e56,
  0x25e65, 0x07ca6, 0x25e62, 0x07cb6, 0x07cb7, 0x07cbf, 0x25ed8, 0x07cc4,
  0x25ec2, 0x07cc8, 0x07ccd, 0x25ee8, 0x07cd7, 0x25f23, 0x07ce6, 0x07ceb,
  0x25f5c, 0x07cf5, 0x07d03, 0x07d09, 0x042c6, 0x07d12, 0x07d1e, 0x25fe0,
  0x25fd4, 0x07d3d, 0x07d3e, 0x07d40, 0x07d47, 0x2600c, 0x25ffb, 0x042d6,
  0x07d59, 0x07d5a, 0x07d6a, 0x07d70, 0x042dd, 0x07d7f, 0x26017, 0x07d86,
  0x07d88, 0x07d8c, 0x07d97, 0x26060, 0x07d9d, 0x07da7, 0x07daa, 0x07db6,
  0x07db7, 0x07dc0, 0x07dd7, 0x07dd9, 0x07de6, 0x07df1, 0x07df9, 0x04302,
  0x260ed, 0x0fa58, 0x07e10, 0x07e17, 0x07e1d, 0x07e20, 0x07e27, 0x07e2c,
  0x07e45, 0x07e73, 0x07e75, 0x07e7e, 0x07e86, 0x07e87, 0x0432b, 0x07e91,
  0x07e98, 0x07e9a, 0x04343, 0x07f3c, 0x07f3b, 0x07f3e, 0x07f43, 0x07f44,
  0x07f4f, 0x034c1, 0x26270, 0x07f52, 0x26286, 0x07f61, 0x07f63, 0x07f64,
  0x07f6d, 0x07f7d, 0x07f7e, 0x2634c, 0x07f90, 0x0517b, 0x23d0e, 0x07f96,
  0x07f9c, 0x07fad, 0x26402, 0x07fc3, 0x07fcf, 0x07fe3, 0x07fe5, 0x07fef,
  0x07ff2, 0x08002, 0x0800a, 0x08008, 0x0800e, 0x08011, 0x08016, 0x08024,
  0x0802c, 0x08030, 0x08043, 0x08066, 0x08071, 0x08075, 0x0807b, 0x08099,
  0x0809c, 0x080a4, 0x080a7, 0x080b8, 0x2667e, 0x080c5, 0x080d5, 0x080d8,
  0x080e6, 0x266b0, 0x0810d, 0x080f5, 0x080fb, 0x043ee, 0x08135, 0x08116,
  0x0811e, 0x043f0, 0x08124, 0x08127, 0x0812c, 0x2671d, 0x0813d, 0x04408,
  0x08169, 0x04417, 0x08181, 0x0441c, 0x08184, 0x08185, 0x04422, 0x08198,
  0x081b2, 0x081c1, 0x081c3, 0x081d6, 0x081db, 0x268dd, 0x081e4, 0x268ea,
  0x081ec, 0x26951, 0x081fd, 0x081ff, 0x2696f, 0x08204, 0x269dd, 0x08219,
  0x08221, 0x08222, 0x26a1e, 0x08232, 0x08234, 0x0823c, 0x08246, 0x08249,
  0x08245, 0x26a58, 0x0824b, 0x04476, 0x0824f, 0x0447a, 0x08257, 0x26a8c,
  0x0825c, 0x08263, 0x26ab7, 0x0fa5d, 0x0fa5e, 0x08279, 0x04491, 0x0827d,
  0x0827f, 0x08283, 0x0828a, 0x08293, 0x082a7, 0x082a8, 0x082b2, 0x082b4,
  0x082ba, 0x082bc, 0x082e2, 0x082e8, 0x082f7, 0x08307, 0x08308, 0x0830c,
  0x08354, 0x0831b, 0x0831d, 0x08330, 0x0833c, 0x08344, 0x08357, 0x044be,
  0x0837f, 0x044d4, 0x044b3, 0x0838d, 0x08394, 0x08395, 0x0839b, 0x0839d,
  0x083c9, 0x083d0, 0x083d4, 0x083dd, 0x083e5, 0x083f9, 0x0840f, 0x08411,
  0x08415, 0x26c73, 0x08417, 0x08439, 0x0844a, 0x0844f, 0x08451, 0x08452,
  0x08459, 0x0845a, 0x0845c, 0x26cdd, 0x08465, 0x08476, 0x08478, 0x0847c,
  0x08481, 0x0450d, 0x084dc, 0x08497, 0x084a6, 0x084be, 0x04508, 0x084ce,
  0x084cf, 0x084d3, 0x26e65, 0x084e7, 0x084ea, 0x084ef, 0x084f0, 0x084f1,
  0x084fa, 0x084fd, 0x0850c, 0x0851b, 0x08524, 0x08525, 0x0852b, 0x08534,
  0x0854f, 0x0856f, 0x04525, 0x04543, 0x0853e, 0x08551, 0x08553, 0x0855e,
  0x08561, 0x08562, 0x26f94, 0x0857b, 0x0857d, 0x0857f, 0x08581, 0x08586,
  0x08593, 0x0859d, 0x0859f, 0x26ff8, 0x26ff6, 0x26ff7, 0x085b7, 0x085bc,
  0x085c7, 0x085ca, 0x085d8, 0x085d9, 0x085df, 0x085e1, 0x085e6, 0x085f6,
  0x08600, 0x08611, 0x0861e, 0x08621, 0x08624, 0x08627, 0x2710d, 0x08639,
  0x0863c, 0x27139, 0x08640, 0x0fa20, 0x08653, 0x08656, 0x0866f, 0x08677,
  0x0867a, 0x08687, 0x08689, 0x0868d, 0x08691, 0x0869c, 0x0869d, 0x086a8,
  0x0fa21, 0x086b1, 0x086b3, 0x086c1, 0x086c3, 0x086d1, 0x086d5, 0x086d7,
  0x086e3, 0x086e6, 0x045b8, 0x08705, 0x08707, 0x0870e, 0x08710, 0x08713,
  0x08719, 0x0871f, 0x08721, 0x08723, 0x08731, 0x0873a, 0x0873e, 0x08740,
  0x08743, 0x08751, 0x08758, 0x08764, 0x08765, 0x08772, 0x0877c, 0x273db,
  0x273da, 0x087a7, 0x08789, 0x0878b, 0x08793, 0x087a0, 0x273fe, 0x045e5,
  0x087be, 0x27410, 0x087c1, 0x087ce, 0x087f5, 0x087df, 0x27449, 0x087e3,
  0x087e5, 0x087e6, 0x087ea, 0x087eb, 0x087ed, 0x08801, 0x08803, 0x0880b,
  0x08813, 0x08828, 0x0882e, 0x08832, 0x0883c, 0x0460f, 0x0884a, 0x08858,
  0x0885f, 0x08864, 0x27615, 0x27614, 0x08869, 0x27631, 0x0886f, 0x088a0,
  0x088bc, 0x088bd, 0x088be, 0x088c0, 0x088d2, 0x27693, 0x088d1, 0x088d3,
  0x088db, 0x088f0, 0x088f1, 0x04641, 0x08901, 0x2770e, 0x08937, 0x27723,
  0x08942, 0x08945, 0x08949, 0x27752, 0x04665, 0x08962, 0x08980, 0x08989,
  0x08990, 0x0899f, 0x089b0, 0x089b7, 0x089d6, 0x089d8, 0x089eb, 0x046a1,
  0x089f1, 0x089f3, 0x089fd, 0x089ff, 0x046af, 0x08a11, 0x08a14, 0x27985,
  0x08a21, 0x08a35, 0x08a3e, 0x08a45, 0x08a4d, 0x08a58, 0x08aae, 0x08a90,
  0x08ab7, 0x08abe, 0x08ad7, 0x08afc, 0x27a84, 0x08b0a, 0x08b05, 0x08b0d,
  0x08b1c, 0x08b1f, 0x08b2d, 0x08b43, 0x0470c, 0x08b51, 0x08b5e, 0x08b76,
  0x08b7f, 0x08b81, 0x08b8b, 0x08b94, 0x08b95, 0x08b9c, 0x08b9e, 0x08c39,
  0x27bb3, 0x08c3d, 0x27bbe, 0x27bc7, 0x08c45, 0x08c47, 0x08c4f, 0x08c54,
  0x08c57, 0x08c69, 0x08c6d, 0x08c73, 0x27cb8, 0x08c93, 0x08c92, 0x08c99,
  0x04764, 0x08c9b, 0x08ca4, 0x08cd6, 0x08cd5, 0x08cd9, 0x27da0, 0x08cf0,
  0x08cf1, 0x27e10, 0x08d09, 0x08d0e, 0x08d6c, 0x08d84, 0x08d95, 0x08da6,
  0x27fb7, 0x08dc6, 0x08dc8, 0x08dd9, 0x08dec, 0x08e0c, 0x047fd, 0x08dfd,
  0x08e06, 0x2808a, 0x08e14, 0x08e16, 0x08e21, 0x08e22, 0x08e27, 0x280bb,
  0x04816, 0x08e36, 0x08e39, 0x08e4b, 0x08e54, 0x08e62, 0x08e6c, 0x08e6d,
  0x08e6f, 0x08e98, 0x08e9e, 0x08eae, 0x08eb3, 0x08eb5, 0x08eb6, 0x08ebb,
  0x28282, 0x08ed1, 0x08ed4, 0x0484e, 0x08ef9, 0x282f3, 0x08f00, 0x08f08,
  0x08f17, 0x08f2b, 0x08f40, 0x08f4a, 0x08f58, 0x2840c, 0x08fa4, 0x08fb4,
  0x0fa66, 0x08fb6, 0x28455, 0x08fc1, 0x08fc6, 0x0fa24, 0x08fca, 0x08fcd,
  0x08fd3, 0x08fd5, 0x08fe0, 0x08ff1, 0x08ff5, 0x08ffb, 0x09002, 0x0900c,
  0x09037, 0x2856b, 0x09043, 0x09044, 0x0905d, 0x285c8, 0x285c9, 0x09085,
  0x0908c, 0x09090, 0x0961d, 0x090a1, 0x048b5, 0x090b0, 0x090b6, 0x090c3,
  0x090c8, 0x286d7, 0x090dc, 0x090df, 0x286fa, 0x090f6, 0x090f2, 0x09100,
  0x090eb, 0x090fe, 0x090ff, 0x09104, 0x09106, 0x09118, 0x0911c, 0x0911e,
  0x09137, 0x09139, 0x0913a, 0x09146, 0x09147, 0x09157, 0x09159, 0x09161,
  0x09164, 0x09174, 0x09179, 0x09185, 0x0918e, 0x091a8, 0x091ae, 0x091b3,
  0x091b6, 0x091c3, 0x091c4, 0x091da, 0x28949, 0x28946, 0x091ec, 0x091ee,
  0x09201, 0x0920a, 0x09216, 0x09217, 0x2896b, 0x09233, 0x09242, 0x09247,
  0x0924a, 0x0924e, 0x09251, 0x09256, 0x09259, 0x09260, 0x09261, 0x09265,
  0x09267, 0x09268, 0x28987, 0x28988, 0x0927c, 0x0927d, 0x0927f, 0x09289,
  0x0928d, 0x09297, 0x09299, 0x0929f, 0x092a7, 0x092ab, 0x289ba, 0x289bb,
  0x092b2, 0x092bf, 0x092c0, 0x092c6, 0x092ce, 0x092d0, 0x092d7, 0x092d9,
  0x092e5, 0x092e7, 0x09311, 0x28a1e, 0x28a29, 0x092f7, 0x092f9, 0x092fb,
  0x09302, 0x0930d, 0x09315, 0x0931d, 0x0931e, 0x09327, 0x09329, 0x28a71,
  0x28a43, 0x09347, 0x09351, 0x09357, 0x0935a, 0x0936b, 0x09371, 0x09373,
  0x093a1, 0x28a99, 0x28acd, 0x09388, 0x0938b, 0x0938f, 0x0939e, 0x093f5,
  0x28ae4, 0x28add, 0x093f1, 0x093c1, 0x093c7, 0x093dc, 0x093e2, 0x093e7,
  0x09409, 0x0940f, 0x09416, 0x09417, 0x093fb, 0x09432, 0x09434, 0x0943b,
  0x09445, 0x28bc1, 0x28bef, 0x0946d, 0x0946f, 0x09578, 0x09579, 0x09586,
  0x0958c, 0x0958d, 0x28d10, 0x095ab, 0x095b4, 0x28d71, 0x095c8, 0x28dfb,
  0x28e1f, 0x0962c, 0x09633, 0x09634, 0x28e36, 0x0963c, 0x09641, 0x09661,
  0x28e89, 0x09682, 0x28eeb, 0x0969a, 0x28f32, 0x049e7, 0x096a9, 0x096af,
  0x096b3, 0x096ba, 0x096bd, 0x049fa, 0x28ff8, 0x096d8, 0x096da, 0x096dd,
  0x04a04, 0x09714, 0x09723, 0x04a29, 0x09736, 0x09741, 0x09747, 0x09755,
  0x09757, 0x0975b, 0x0976a, 0x292a0, 0x292b1, 0x09796, 0x0979a, 0x0979e,
  0x097a2, 0x097b1, 0x097b2, 0x097be, 0x097cc, 0x097d1, 0x097d4, 0x097d8,
  0x097d9, 0x097e1, 0x097f1, 0x09804, 0x0980d, 0x0980e, 0x09814, 0x09816,
  0x04abc, 0x29490, 0x09823, 0x09832, 0x09833, 0x09825, 0x09847, 0x09866,
  0x098ab, 0x098ad, 0x098b0, 0x295cf, 0x098b7, 0x098b8, 0x098bb, 0x098bc,
  0x098bf, 0x098c2, 0x098c7, 0x098cb, 0x098e0, 0x2967f, 0x098e1, 0x098e3,
  0x098e5, 0x098ea, 0x098f0, 0x098f1, 0x098f3, 0x09908, 0x04b3b, 0x296f0,
  0x09916, 0x09917, 0x29719, 0x0991a, 0x0991b, 0x0991c, 0x29750, 0x09931,
  0x09932, 0x09933, 0x0993a, 0x0993b, 0x0993c, 0x09940, 0x09941, 0x09946,
  0x0994d, 0x0994e, 0x0995c, 0x0995f, 0x09960, 0x099a3, 0x099a6, 0x099b9,
  0x099bd, 0x099bf, 0x099c3, 0x099c9, 0x099d4, 0x099d9, 0x099de, 0x298c6,
  0x099f0, 0x099f9, 0x099fc, 0x09a0a, 0x09a11, 0x09a16, 0x09a1a, 0x09a20,
  0x09a31, 0x09a36, 0x09a44, 0x09a4c, 0x09a58, 0x04bc2, 0x09aaf, 0x04bca,
  0x09ab7, 0x04bd2, 0x09ab9, 0x29a72, 0x09ac6, 0x09ad0, 0x09ad2, 0x09ad5,
  0x04be8, 0x09adc, 0x09ae0, 0x09ae5, 0x09ae9, 0x09b03, 0x09b0c, 0x09b10,
  0x09b12, 0x09b16, 0x09b1c, 0x09b2b, 0x09b33, 0x09b3d, 0x04c20, 0x09b4b,
  0x09b63, 0x09b65, 0x09b6b, 0x09b6c, 0x09b73, 0x09b76, 0x09b77, 0x09ba6,
  0x09bac, 0x09bb1, 0x29ddb, 0x29e3d, 0x09bb2, 0x09bb8, 0x09bbe, 0x09bc7,
  0x09bf3, 0x09bd8, 0x09bdd, 0x09be7, 0x09bea, 0x09beb, 0x09bef, 0x09bee,
  0x29e15, 0x09bfa, 0x29e8a, 0x09bf7, 0x29e49, 0x09c16, 0x09c18, 0x09c19,
  0x09c1a, 0x09c1d, 0x09c22, 0x09c27, 0x09c29, 0x09c2a, 0x29ec4, 0x09c31,
  0x09c36, 0x09c37, 0x09c45, 0x09c5c, 0x29ee9, 0x09c49, 0x09c4a, 0x29edb,
  0x09c54, 0x09c58, 0x09c5b, 0x09c5d, 0x09c5f, 0x09c69, 0x09c6a, 0x09c6b,
  0x09c6d, 0x09c6e, 0x09c70, 0x09c72, 0x09c75, 0x09c7a, 0x09ce6, 0x09cf2,
  0x09d0b, 0x09d02, 0x29fce, 0x09d11, 0x09d17, 0x09d18, 0x2a02f, 0x04cc4,
  0x2a01a, 0x09d32, 0x04cd1, 0x09d42, 0x09d4a, 0x09d5f, 0x09d62, 0x2a0f9,
  0x09d69, 0x09d6b, 0x2a082, 0x09d73, 0x09d76, 0x09d77, 0x09d7e, 0x09d84,
  0x09d8d, 0x09d99, 0x09da1, 0x09dbf, 0x09db5, 0x09db9, 0x09dbd, 0x09dc3,
  0x09dc7, 0x09dc9, 0x09dd6, 0x09dda, 0x09ddf, 0x09de0, 0x09de3, 0x09df4,
  0x04d07, 0x09e0a, 0x09e02, 0x09e0d, 0x09e19, 0x09e1c, 0x09e1d, 0x09e7b,
  0x22218, 0x09e80, 0x09e85, 0x09e9b, 0x09ea8, 0x2a38c, 0x09ebd, 0x2a437,
  0x09edf, 0x09ee7, 0x09eee, 0x09eff, 0x09f02, 0x04d77, 0x09f03, 0x09f17,
  0x09f19, 0x09f2f, 0x09f37, 0x09f3a, 0x09f3d, 0x09f41, 0x09f45, 0x09f46,
  0x09f53, 0x09f55, 0x09f58, 0x2a5f1, 0x09f5d, 0x2a602, 0x09f69, 0x2a61a,
  0x09f6d, 0x09f70, 0x09f75, 0x2a6b2, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, 
};

int32_t ucode_cp932[U_kanji] = {
  0x03000, 0x03001, 0x03002, 0x0ff0c, 0x0ff0e, 0x030fb, 0x0ff1a, 0x0ff1b,
  0x0ff1f, 0x0ff01, 0x0309b, 0x0309c, 0x000b4, 0x0ff40, 0x000a8, 0x0ff3e,
  0x0ffe3, 0x0ff3f, 0x030fd, 0x030fe, 0x0309d, 0x0309e, 0x03003, 0x04edd,
  0x03005, 0x03006, 0x03007, 0x030fc, 0x02015, 0x02010, 0x0ff0f, 0x0ff3c,
  0x0ff5e, 0x02225, 0x0ff5c, 0x02026, 0x02025, 0x02018, 0x02019, 0x0201c,
  0x0201d, 0x0ff08, 0x0ff09, 0x03014, 0x03015, 0x0ff3b, 0x0ff3d, 0x0ff5b,
  0x0ff5d, 0x03008, 0x03009, 0x0300a, 0x0300b, 0x0300c, 0x0300d, 0x0300e,
  0x0300f, 0x03010, 0x03011, 0x0ff0b, 0x0ff0d, 0x000b1, 0x000d7, 0x000f7,
  0x0ff1d, 0x02260, 0x0ff1c, 0x0ff1e, 0x02266, 0x02267, 0x0221e, 0x02234,
  0x02642, 0x02640, 0x000b0, 0x02032, 0x02033, 0x02103, 0x0ffe5, 0x0ff04,
  0x0ffe0, 0x0ffe1, 0x0ff05, 0x0ff03, 0x0ff06, 0x0ff0a, 0x0ff20, 0x000a7,
  0x02606, 0x02605, 0x025cb, 0x025cf, 0x025ce, 0x025c7, 0x025c6, 0x025a1,
  0x025a0, 0x025b3, 0x025b2, 0x025bd, 0x025bc, 0x0203b, 0x03012, 0x02192,
  0x02190, 0x02191, 0x02193, 0x03013, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, 0x02208,
  0x0220b, 0x02286, 0x02287, 0x02282, 0x02283, 0x0222a, 0x02229, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, 0x02227,
  0x02228, 0x0ffe2, 0x021d2, 0x021d4, 0x02200, 0x02203, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, 0x02220, 0x022a5, 0x02312, 0x02202, 0x02207, 0x02261, 0x02252,
  0x0226a, 0x0226b, 0x0221a, 0x0223d, 0x0221d, 0x02235, 0x0222b, 0x0222c,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, 0x0212b,
  0x02030, 0x0266f, 0x0266d, 0x0266a, 0x02020, 0x02021, 0x000b6, U_error,
  U_error, U_error, U_error, 0x025ef, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, 0x0ff10, 0x0ff11, 0x0ff12, 0x0ff13, 0x0ff14,
  0x0ff15, 0x0ff16, 0x0ff17, 0x0ff18, 0x0ff19, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, 0x0ff21, 0x0ff22, 0x0ff23, 0x0ff24,
  0x0ff25, 0x0ff26, 0x0ff27, 0x0ff28, 0x0ff29, 0x0ff2a, 0x0ff2b, 0x0ff2c,
  0x0ff2d, 0x0ff2e, 0x0ff2f, 0x0ff30, 0x0ff31, 0x0ff32, 0x0ff33, 0x0ff34,
  0x0ff35, 0x0ff36, 0x0ff37, 0x0ff38, 0x0ff39, 0x0ff3a, U_error, U_error,
  U_error, U_error, U_error, U_error, 0x0ff41, 0x0ff42, 0x0ff43, 0x0ff44,
  0x0ff45, 0x0ff46, 0x0ff47, 0x0ff48, 0x0ff49, 0x0ff4a, 0x0ff4b, 0x0ff4c,
  0x0ff4d, 0x0ff4e, 0x0ff4f, 0x0ff50, 0x0ff51, 0x0ff52, 0x0ff53, 0x0ff54,
  0x0ff55, 0x0ff56, 0x0ff57, 0x0ff58, 0x0ff59, 0x0ff5a, U_error, U_error,
  U_error, U_error, 0x03041, 0x03042, 0x03043, 0x03044, 0x03045, 0x03046,
  0x03047, 0x03048, 0x03049, 0x0304a, 0x0304b, 0x0304c, 0x0304d, 0x0304e,
  0x0304f, 0x03050, 0x03051, 0x03052, 0x03053, 0x03054, 0x03055, 0x03056,
  0x03057, 0x03058, 0x03059, 0x0305a, 0x0305b, 0x0305c, 0x0305d, 0x0305e,
  0x0305f, 0x03060, 0x03061, 0x03062, 0x03063, 0x03064, 0x03065, 0x03066,
  0x03067, 0x03068, 0x03069, 0x0306a, 0x0306b, 0x0306c, 0x0306d, 0x0306e,
  0x0306f, 0x03070, 0x03071, 0x03072, 0x03073, 0x03074, 0x03075, 0x03076,
  0x03077, 0x03078, 0x03079, 0x0307a, 0x0307b, 0x0307c, 0x0307d, 0x0307e,
  0x0307f, 0x03080, 0x03081, 0x03082, 0x03083, 0x03084, 0x03085, 0x03086,
  0x03087, 0x03088, 0x03089, 0x0308a, 0x0308b, 0x0308c, 0x0308d, 0x0308e,
  0x0308f, 0x03090, 0x03091, 0x03092, 0x03093, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  0x030a1, 0x030a2, 0x030a3, 0x030a4, 0x030a5, 0x030a6, 0x030a7, 0x030a8,
  0x030a9, 0x030aa, 0x030ab, 0x030ac, 0x030ad, 0x030ae, 0x030af, 0x030b0,
  0x030b1, 0x030b2, 0x030b3, 0x030b4, 0x030b5, 0x030b6, 0x030b7, 0x030b8,
  0x030b9, 0x030ba, 0x030bb, 0x030bc, 0x030bd, 0x030be, 0x030bf, 0x030c0,
  0x030c1, 0x030c2, 0x030c3, 0x030c4, 0x030c5, 0x030c6, 0x030c7, 0x030c8,
  0x030c9, 0x030ca, 0x030cb, 0x030cc, 0x030cd, 0x030ce, 0x030cf, 0x030d0,
  0x030d1, 0x030d2, 0x030d3, 0x030d4, 0x030d5, 0x030d6, 0x030d7, 0x030d8,
  0x030d9, 0x030da, 0x030db, 0x030dc, 0x030dd, 0x030de, 0x030df, 0x030e0,
  0x030e1, 0x030e2, 0x030e3, 0x030e4, 0x030e5, 0x030e6, 0x030e7, 0x030e8,
  0x030e9, 0x030ea, 0x030eb, 0x030ec, 0x030ed, 0x030ee, 0x030ef, 0x030f0,
  0x030f1, 0x030f2, 0x030f3, 0x030f4, 0x030f5, 0x030f6, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x00391, 0x00392,
  0x00393, 0x00394, 0x00395, 0x00396, 0x00397, 0x00398, 0x00399, 0x0039a,
  0x0039b, 0x0039c, 0x0039d, 0x0039e, 0x0039f, 0x003a0, 0x003a1, 0x003a3,
  0x003a4, 0x003a5, 0x003a6, 0x003a7, 0x003a8, 0x003a9, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x003b1, 0x003b2,
  0x003b3, 0x003b4, 0x003b5, 0x003b6, 0x003b7, 0x003b8, 0x003b9, 0x003ba,
  0x003bb, 0x003bc, 0x003bd, 0x003be, 0x003bf, 0x003c0, 0x003c1, 0x003c3,
  0x003c4, 0x003c5, 0x003c6, 0x003c7, 0x003c8, 0x003c9, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, 0x00410, 0x00411, 0x00412, 0x00413,
  0x00414, 0x00415, 0x00401, 0x00416, 0x00417, 0x00418, 0x00419, 0x0041a,
  0x0041b, 0x0041c, 0x0041d, 0x0041e, 0x0041f, 0x00420, 0x00421, 0x00422,
  0x00423, 0x00424, 0x00425, 0x00426, 0x00427, 0x00428, 0x00429, 0x0042a,
  0x0042b, 0x0042c, 0x0042d, 0x0042e, 0x0042f, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, 0x00430, 0x00431, 0x00432, 0x00433,
  0x00434, 0x00435, 0x00451, 0x00436, 0x00437, 0x00438, 0x00439, 0x0043a,
  0x0043b, 0x0043c, 0x0043d, 0x0043e, 0x0043f, 0x00440, 0x00441, 0x00442,
  0x00443, 0x00444, 0x00445, 0x00446, 0x00447, 0x00448, 0x00449, 0x0044a,
  0x0044b, 0x0044c, 0x0044d, 0x0044e, 0x0044f, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, 0x02500, 0x02502, 0x0250c, 0x02510, 0x02518, 0x02514,
  0x0251c, 0x0252c, 0x02524, 0x02534, 0x0253c, 0x02501, 0x02503, 0x0250f,
  0x02513, 0x0251b, 0x02517, 0x02523, 0x02533, 0x0252b, 0x0253b, 0x0254b,
  0x02520, 0x0252f, 0x02528, 0x02537, 0x0253f, 0x0251d, 0x02530, 0x02525,
  0x02538, 0x02542, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  0x02460, 0x02461, 0x02462, 0x02463, 0x02464, 0x02465, 0x02466, 0x02467,
  0x02468, 0x02469, 0x0246a, 0x0246b, 0x0246c, 0x0246d, 0x0246e, 0x0246f,
  0x02470, 0x02471, 0x02472, 0x02473, 0x02160, 0x02161, 0x02162, 0x02163,
  0x02164, 0x02165, 0x02166, 0x02167, 0x02168, 0x02169, U_error, 0x03349,
  0x03314, 0x03322, 0x0334d, 0x03318, 0x03327, 0x03303, 0x03336, 0x03351,
  0x03357, 0x0330d, 0x03326, 0x03323, 0x0332b, 0x0334a, 0x0333b, 0x0339c,
  0x0339d, 0x0339e, 0x0338e, 0x0338f, 0x033c4, 0x033a1, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, 0x0337b, 0x0301d,
  0x0301f, 0x02116, 0x033cd, 0x02121, 0x032a4, 0x032a5, 0x032a6, 0x032a7,
  0x032a8, 0x03231, 0x03232, 0x03239, 0x0337e, 0x0337d, 0x0337c, 0x02252,
  0x02261, 0x0222b, 0x0222e, 0x02211, 0x0221a, 0x022a5, 0x02220, 0x0221f,
  0x022bf, 0x02235, 0x02229, 0x0222a, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, 0x02170, 0x02171, 0x02172, 0x02173, 0x02174, 0x02175,
  0x02176, 0x02177, 0x02178, 0x02179, 0x02160, 0x02161, 0x02162, 0x02163,
  0x02164, 0x02165, 0x02166, 0x02167, 0x02168, 0x02169, 0x0ffe2, 0x0ffe4,
  0x0ff07, 0x0ff02, 0x03231, 0x02116, 0x02121, 0x02235, 0x07e8a, 0x0891c,
  0x09348, 0x09288, 0x084dc, 0x04fc9, 0x070bb, 0x06631, 0x068c8, 0x092f9,
  0x066fb, 0x05f45, 0x04e28, 0x04ee1, 0x04efc, 0x04f00, 0x04f03, 0x04f39,
  0x04f56, 0x04f92, 0x04f8a, 0x04f9a, 0x04f94, 0x04fcd, 0x05040, 0x05022,
  0x04fff, 0x0501e, 0x05046, 0x05070, 0x05042, 0x05094, 0x050f4, 0x050d8,
  0x0514a, 0x05164, 0x0519d, 0x051be, 0x051ec, 0x05215, 0x0529c, 0x052a6,
  0x052c0, 0x052db, 0x05300, 0x05307, 0x05324, 0x05372, 0x05393, 0x053b2,
  0x053dd, 0x0fa0e, 0x0549c, 0x0548a, 0x054a9, 0x054ff, 0x05586, 0x05759,
  0x05765, 0x057ac, 0x057c8, 0x057c7, 0x0fa0f, 0x0fa10, 0x0589e, 0x058b2,
  0x0590b, 0x05953, 0x0595b, 0x0595d, 0x05963, 0x059a4, 0x059ba, 0x05b56,
  0x05bc0, 0x0752f, 0x05bd8, 0x05bec, 0x05c1e, 0x05ca6, 0x05cba, 0x05cf5,
  0x05d27, 0x05d53, 0x0fa11, 0x05d42, 0x05d6d, 0x05db8, 0x05db9, 0x05dd0,
  0x05f21, 0x05f34, 0x05f67, 0x05fb7, 0x05fde, 0x0605d, 0x06085, 0x0608a,
  0x060de, 0x060d5, 0x06120, 0x060f2, 0x06111, 0x06137, 0x06130, 0x06198,
  0x06213, 0x062a6, 0x063f5, 0x06460, 0x0649d, 0x064ce, 0x0654e, 0x06600,
  0x06615, 0x0663b, 0x06609, 0x0662e, 0x0661e, 0x06624, 0x06665, 0x06657,
  0x06659, 0x0fa12, 0x06673, 0x06699, 0x066a0, 0x066b2, 0x066bf, 0x066fa,
  0x0670e, 0x0f929, 0x06766, 0x067bb, 0x06852, 0x067c0, 0x06801, 0x06844,
  0x068cf, 0x0fa13, 0x06968, 0x0fa14, 0x06998, 0x069e2, 0x06a30, 0x06a6b,
  0x06a46, 0x06a73, 0x06a7e, 0x06ae2, 0x06ae4, 0x06bd6, 0x06c3f, 0x06c5c,
  0x06c86, 0x06c6f, 0x06cda, 0x06d04, 0x06d87, 0x06d6f, 0x06d96, 0x06dac,
  0x06dcf, 0x06df8, 0x06df2, 0x06dfc, 0x06e39, 0x06e5c, 0x06e27, 0x06e3c,
  0x06ebf, 0x06f88, 0x06fb5, 0x06ff5, 0x07005, 0x07007, 0x07028, 0x07085,
  0x070ab, 0x0710f, 0x07104, 0x0715c, 0x07146, 0x07147, 0x0fa15, 0x071c1,
  0x071fe, 0x072b1, 0x072be, 0x07324, 0x0fa16, 0x07377, 0x073bd, 0x073c9,
  0x073d6, 0x073e3, 0x073d2, 0x07407, 0x073f5, 0x07426, 0x0742a, 0x07429,
  0x0742e, 0x07462, 0x07489, 0x0749f, 0x07501, 0x0756f, 0x07682, 0x0769c,
  0x0769e, 0x0769b, 0x076a6, 0x0fa17, 0x07746, 0x052af, 0x07821, 0x0784e,
  0x07864, 0x0787a, 0x07930, 0x0fa18, 0x0fa19, 0x0fa1a, 0x07994, 0x0fa1b,
  0x0799b, 0x07ad1, 0x07ae7, 0x0fa1c, 0x07aeb, 0x07b9e, 0x0fa1d, 0x07d48,
  0x07d5c, 0x07db7, 0x07da0, 0x07dd6, 0x07e52, 0x07f47, 0x07fa1, 0x0fa1e,
  0x08301, 0x08362, 0x0837f, 0x083c7, 0x083f6, 0x08448, 0x084b4, 0x08553,
  0x08559, 0x0856b, 0x0fa1f, 0x085b0, 0x0fa20, 0x0fa21, 0x08807, 0x088f5,
  0x08a12, 0x08a37, 0x08a79, 0x08aa7, 0x08abe, 0x08adf, 0x0fa22, 0x08af6,
  0x08b53, 0x08b7f, 0x08cf0, 0x08cf4, 0x08d12, 0x08d76, 0x0fa23, 0x08ecf,
  0x0fa24, 0x0fa25, 0x09067, 0x090de, 0x0fa26, 0x09115, 0x09127, 0x091da,
  0x091d7, 0x091de, 0x091ed, 0x091ee, 0x091e4, 0x091e5, 0x09206, 0x09210,
  0x0920a, 0x0923a, 0x09240, 0x0923c, 0x0924e, 0x09259, 0x09251, 0x09239,
  0x09267, 0x092a7, 0x09277, 0x09278, 0x092e7, 0x092d7, 0x092d9, 0x092d0,
  0x0fa27, 0x092d5, 0x092e0, 0x092d3, 0x09325, 0x09321, 0x092fb, 0x0fa28,
  0x0931e, 0x092ff, 0x0931d, 0x09302, 0x09370, 0x09357, 0x093a4, 0x093c6,
  0x093de, 0x093f8, 0x09431, 0x09445, 0x09448, 0x09592, 0x0f9dc, 0x0fa29,
  0x0969d, 0x096af, 0x09733, 0x0973b, 0x09743, 0x0974d, 0x0974f, 0x09751,
  0x09755, 0x09857, 0x09865, 0x0fa2a, 0x0fa2b, 0x09927, 0x0fa2c, 0x0999e,
  0x09a4e, 0x09ad9, 0x09adc, 0x09b75, 0x09b72, 0x09b8f, 0x09bb1, 0x09bbb,
  0x09c00, 0x09d70, 0x09d6b, 0x0fa2d, 0x09e19, 0x09ed1, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  0x07e8a, 0x0891c, 0x09348, 0x09288, 0x084dc, 0x04fc9, 0x070bb, 0x06631,
  0x068c8, 0x092f9, 0x066fb, 0x05f45, 0x04e28, 0x04ee1, 0x04efc, 0x04f00,
  0x04f03, 0x04f39, 0x04f56, 0x04f92, 0x04f8a, 0x04f9a, 0x04f94, 0x04fcd,
  0x05040, 0x05022, 0x04fff, 0x0501e, 0x05046, 0x05070, 0x05042, 0x05094,
  0x050f4, 0x050d8, 0x0514a, 0x05164, 0x0519d, 0x051be, 0x051ec, 0x05215,
  0x0529c, 0x052a6, 0x052c0, 0x052db, 0x05300, 0x05307, 0x05324, 0x05372,
  0x05393, 0x053b2, 0x053dd, 0x0fa0e, 0x0549c, 0x0548a, 0x054a9, 0x054ff,
  0x05586, 0x05759, 0x05765, 0x057ac, 0x057c8, 0x057c7, 0x0fa0f, 0x0fa10,
  0x0589e, 0x058b2, 0x0590b, 0x05953, 0x0595b, 0x0595d, 0x05963, 0x059a4,
  0x059ba, 0x05b56, 0x05bc0, 0x0752f, 0x05bd8, 0x05bec, 0x05c1e, 0x05ca6,
  0x05cba, 0x05cf5, 0x05d27, 0x05d53, 0x0fa11, 0x05d42, 0x05d6d, 0x05db8,
  0x05db9, 0x05dd0, 0x05f21, 0x05f34, 0x05f67, 0x05fb7, 0x05fde, 0x0605d,
  0x06085, 0x0608a, 0x060de, 0x060d5, 0x06120, 0x060f2, 0x06111, 0x06137,
  0x06130, 0x06198, 0x06213, 0x062a6, 0x063f5, 0x06460, 0x0649d, 0x064ce,
  0x0654e, 0x06600, 0x06615, 0x0663b, 0x06609, 0x0662e, 0x0661e, 0x06624,
  0x06665, 0x06657, 0x06659, 0x0fa12, 0x06673, 0x06699, 0x066a0, 0x066b2,
  0x066bf, 0x066fa, 0x0670e, 0x0f929, 0x06766, 0x067bb, 0x06852, 0x067c0,
  0x06801, 0x06844, 0x068cf, 0x0fa13, 0x06968, 0x0fa14, 0x06998, 0x069e2,
  0x06a30, 0x06a6b, 0x06a46, 0x06a73, 0x06a7e, 0x06ae2, 0x06ae4, 0x06bd6,
  0x06c3f, 0x06c5c, 0x06c86, 0x06c6f, 0x06cda, 0x06d04, 0x06d87, 0x06d6f,
  0x06d96, 0x06dac, 0x06dcf, 0x06df8, 0x06df2, 0x06dfc, 0x06e39, 0x06e5c,
  0x06e27, 0x06e3c, 0x06ebf, 0x06f88, 0x06fb5, 0x06ff5, 0x07005, 0x07007,
  0x07028, 0x07085, 0x070ab, 0x0710f, 0x07104, 0x0715c, 0x07146, 0x07147,
  0x0fa15, 0x071c1, 0x071fe, 0x072b1, 0x072be, 0x07324, 0x0fa16, 0x07377,
  0x073bd, 0x073c9, 0x073d6, 0x073e3, 0x073d2, 0x07407, 0x073f5, 0x07426,
  0x0742a, 0x07429, 0x0742e, 0x07462, 0x07489, 0x0749f, 0x07501, 0x0756f,
  0x07682, 0x0769c, 0x0769e, 0x0769b, 0x076a6, 0x0fa17, 0x07746, 0x052af,
  0x07821, 0x0784e, 0x07864, 0x0787a, 0x07930, 0x0fa18, 0x0fa19, 0x0fa1a,
  0x07994, 0x0fa1b, 0x0799b, 0x07ad1, 0x07ae7, 0x0fa1c, 0x07aeb, 0x07b9e,
  0x0fa1d, 0x07d48, 0x07d5c, 0x07db7, 0x07da0, 0x07dd6, 0x07e52, 0x07f47,
  0x07fa1, 0x0fa1e, 0x08301, 0x08362, 0x0837f, 0x083c7, 0x083f6, 0x08448,
  0x084b4, 0x08553, 0x08559, 0x0856b, 0x0fa1f, 0x085b0, 0x0fa20, 0x0fa21,
  0x08807, 0x088f5, 0x08a12, 0x08a37, 0x08a79, 0x08aa7, 0x08abe, 0x08adf,
  0x0fa22, 0x08af6, 0x08b53, 0x08b7f, 0x08cf0, 0x08cf4, 0x08d12, 0x08d76,
  0x0fa23, 0x08ecf, 0x0fa24, 0x0fa25, 0x09067, 0x090de, 0x0fa26, 0x09115,
  0x09127, 0x091da, 0x091d7, 0x091de, 0x091ed, 0x091ee, 0x091e4, 0x091e5,
  0x09206, 0x09210, 0x0920a, 0x0923a, 0x09240, 0x0923c, 0x0924e, 0x09259,
  0x09251, 0x09239, 0x09267, 0x092a7, 0x09277, 0x09278, 0x092e7, 0x092d7,
  0x092d9, 0x092d0, 0x0fa27, 0x092d5, 0x092e0, 0x092d3, 0x09325, 0x09321,
  0x092fb, 0x0fa28, 0x0931e, 0x092ff, 0x0931d, 0x09302, 0x09370, 0x09357,
  0x093a4, 0x093c6, 0x093de, 0x093f8, 0x09431, 0x09445, 0x09448, 0x09592,
  0x0f9dc, 0x0fa29, 0x0969d, 0x096af, 0x09733, 0x0973b, 0x09743, 0x0974d,
  0x0974f, 0x09751, 0x09755, 0x09857, 0x09865, 0x0fa2a, 0x0fa2b, 0x09927,
  0x0fa2c, 0x0999e, 0x09a4e, 0x09ad9, 0x09adc, 0x09b75, 0x09b72, 0x09b8f,
  0x09bb1, 0x09bbb, 0x09c00, 0x09d70, 0x09d6b, 0x0fa2d, 0x09e19, 0x09ed1,
  U_error, U_error, 0x02170, 0x02171, 0x02172, 0x02173, 0x02174, 0x02175,
  0x02176, 0x02177, 0x02178, 0x02179, 0x0ffe2, 0x0ffe4, 0x0ff07, 0x0ff02,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error, U_error, U_error, U_error, U_error,
  U_error, U_error, U_error, U_error,
};
